// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *pdb_name = NULL;
static char *chd_name = NULL;
static char *grid_name = NULL;
static char *asa_name = NULL;
static char *consurf_name = NULL;
static char *ligand_name = NULL;
static RNScalar min_distance = 0;
static RNScalar max_distance = 20;
static RNScalar nearby_threshold = 1;
static RNScalar bin_spacing = 1;
static int nelements = 1;
static RNBoolean splat_type = 0; // 0=add_probability, 2=mult_probability, 3=add_count
static int max_resolution = 256;
static RNScalar grid_spacing = 0.5; // angstroms
static RNScalar world_sigma = 0; // angstroms
static RNScalar world_radius = PDB_UNKNOWN;
static R3Point world_centroid(PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN);
static RNBoolean weight_spans_by_conservation = FALSE;
static RNBoolean weight_spans_by_surface_area = FALSE;
static RNBoolean normalization_type = 0; 
static RNBoolean mask_to_ligand = FALSE;
static RNBoolean mask_to_target_volume = FALSE;
static RNBoolean target_volume_from_ligand = FALSE;
static RNBoolean target_volume_from_ligand_elements = FALSE;
static RNBoolean largest_connected_component_only = FALSE;
static RNScalar target_volume_default = 100;
static RNScalar min_offset_distance = 0;
static RNScalar max_offset_distance = 6;
static RNScalar min_ligand_volume = 100;
static int print_verbose = 0;
static int print_debug = 0;



// Histogram data

static const int max_distance_bins = 1024;
static const int max_parameter_bins = max_distance_bins;
static const int max_element_bins = 6;
static RNScalar bins[max_distance_bins][max_parameter_bins][max_element_bins] = { 0 };
static RNScalar observations[max_distance_bins] = { 0 };



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static int
ReadASAFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read ASA file
  int natoms = file->ReadASAFile(filename);
  if (natoms == 0) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read accessible surface area file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Surface Atoms = %d\n", natoms);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ReadConsurfFiles(PDBFile *file, char *hssp_basename)
{
  // Check number of models
  if (file->NModels() < 1) {
    fprintf(stderr, "File must have at least one model to read conservation scores: %s.\n", hssp_basename);
    return 0;
  }

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read conservation files
  int nresidues = file->ReadConsurfFiles(hssp_basename);

  // Print statistics
  if (print_verbose && (nresidues > 0)) {
    printf("Read consurf files ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Residues = %d\n", nresidues);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static RNArray<PDBAtom *> *
FindLigandAtoms(PDBFile *file, const char *ligand_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int residue_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);

  // Allocate array of atoms
  RNArray<PDBAtom *> *ligand_atoms = new RNArray<PDBAtom *>();
  if (!ligand_atoms) {
    fprintf(stderr, "Unable to allocate array of ligand atoms.\n");
    return NULL;
  }

  // Check ligand name
  if (ligand_name) {
    // Find ligand
    PDBResidue *ligand = file->FindResidue(ligand_name);
    if (!ligand) {
      fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, pdb_name); 
      return NULL;
    }

    // Add atoms of ligand to array
    ligand_atoms->Append(ligand->atoms);
    residue_count = 1;
  }
  else {
    // Add all hetatoms to array
    for (int i = 0; i < model->NResidues(); i++) {
      PDBResidue *residue = model->Residue(i);
      RNBoolean found = FALSE;
      for (int j = 0; j < residue->NAtoms(); j++) {
        PDBAtom *atom = residue->Atom(j);
        if (!atom->IsHetAtom()) continue;
        ligand_atoms->Insert(atom);
        found = TRUE;
      }
      if (found) residue_count++;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Found ligand ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Ligands = %d\n", residue_count);
    printf("  # Atoms = %d\n", ligand_atoms->NEntries());
    fflush(stdout);
  }

  // Return ligand atoms
  return ligand_atoms;
}



static int
ReadChordFile(const char *chd_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(chd_name, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open chord histogram file: %s\n", chd_name);
    return 0;
  }

  // Read parameters into global variables
  int max_bin, bin;
  if (fscanf(fp, "%d%lf%lf%lf%lf%d", &max_bin, &min_distance, &max_distance, &bin_spacing, &nearby_threshold, &nelements) != 6) {
    fprintf(stderr, "Error reading chord file header\n");
    return 0;
  }

  // Read histogram into global variables
  int nbins = 0;
  RNScalar sum = 0;
  RNScalar max_value = 0;
  assert(max_bin < max_distance_bins);
  assert(max_bin == (int) (max_distance / bin_spacing));
  for (int i = 0; i <= max_bin; i++) {
    // Read per-distance information
    if (fscanf(fp, "%d%lf", &bin, &(observations[i])) != 2) {
      fprintf(stderr, "Error reading chord file leader for %d\n", i);
      return 0;
    }

    // Check bin number
    if (bin != i) {
      fprintf(stderr, "Bin number does not match for %d in %s\n", i, chd_name);
      return 0;
    }

    // Read ":" 
    char buffer[64];
    if ((fscanf(fp, "%s", buffer) != 1) || (strcmp(buffer, ":"))) {
      fprintf(stderr, "Error reading chord file colon for %d\n", i);
      return 0;
    }


    // Read bins
    for (int j = 0; j <= i; j++) {
      for (int k = 0; k < nelements; k++) {
        // Read bin
        if (fscanf(fp, "%lf", &(bins[i][j][k])) != 1) {
          fprintf(stderr, "Error reading chord file bin (%d %d %d)\n", i, j, k);
          return 0;
        }

        // Gather statistics
        if (bins[i][j][k] > max_value) max_value = bins[i][j][k];
        sum += bins[i][j][k];
        nbins++;
      }
    }
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Read chord histogram ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bins = %d\n", nbins);
    printf("  # Elements = %d\n", nelements);
    printf("  Max bin = %d\n", max_bin);
    printf("  Min distance = %g\n", min_distance);
    printf("  Max distance = %g\n", max_distance);
    printf("  Bin spacing = %g\n", bin_spacing);
    printf("  Nearby threshold = %g\n", nearby_threshold);
    printf("  Max value = %g\n", max_value);
    printf("  Sum = %g\n", sum);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
SplatChords(PDBFile *file, R3Grid **grids, int ngrids)
{
  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);

  // Compute mask (1 outside protein, and 0 inside)
  RNScalar grid_spacing = 0.5;
  const R3Box& model_box = model->BBox();
  int xres = (int) (model_box.XLength() / grid_spacing + 1); if (xres > 256) xres = 256;
  int yres = (int) (model_box.YLength() / grid_spacing + 1); if (yres > 256) yres = 256;
  int zres = (int) (model_box.ZLength() / grid_spacing + 1); if (zres > 256) zres = 256;
  R3Grid mask(xres, yres, zres);
  mask.SetWorldToGridTransformation(model_box);
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (atom->IsHetAtom()) continue;
    mask.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1);
  }
  mask.Threshold(0.5, 1, 0);

  // Get world box for grid
  R3Box world_box = grids[0]->WorldBox();

  // Consider span between all pairs of atoms on protein surface
  int span_count = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    // Get atom1
    PDBAtom *atom1 = model->Atom(i);
    if (atom1->IsHetAtom()) continue;
    if (atom1->accessible_surface_area == 0) continue;
    for (int j = i+1; j < model->NAtoms(); j++) {
      // Get atom2
      PDBAtom *atom2 = model->Atom(j);
      if (atom2->IsHetAtom()) continue;
      if (atom2->accessible_surface_area == 0) continue;

      // Get span and check its length
      R3Span span(atom1->position, atom2->position);
      RNLength distance = span.Length();
      if (distance <= min_distance) continue;
      if (distance >= max_distance) continue;

      // Check if span intersects world box
      if (!R3Intersects(span, world_box)) continue;

      // Compute visibility of atom1 and atom2
      RNBoolean visible = FALSE;
      RNScalar t_step = mask.GridToWorldScaleFactor();
      RNScalar t_start = atom1->Radius() + t_step;
      RNScalar t_stop = span.Length() - atom2->Radius() - t_step; 
      for (RNScalar t = t_start; t <= t_stop; t += t_step) {
        R3Point world_position = span.Point(t);
        R3Point grid_position = mask.GridPosition(world_position);
        int x = (int) (grid_position[0] + 0.5);
        int y = (int) (grid_position[1] + 0.5);
        int z = (int) (grid_position[2] + 0.5);
        if (mask.GridValue(x, y, z) == 0) { visible = FALSE; break; }
        else visible = TRUE;
      }

      // Check visibility
      if (!visible) continue;

      // Determine weight of the observation
      RNScalar weight = 1;
      if (weight_spans_by_surface_area) {
        if (atom1->accessible_surface_area != PDB_UNKNOWN) weight *= atom1->accessible_surface_area;
        if (atom2->accessible_surface_area != PDB_UNKNOWN) weight *= atom2->accessible_surface_area;
      }
      if (weight_spans_by_conservation) {
        if (atom1->conservation == PDB_UNKNOWN) weight *= 0.1;
        else weight *= atom1->conservation;
        if (atom2->conservation == PDB_UNKNOWN) weight *= 0.1;
        else weight *= atom2->conservation;
      }

      // Splat span 
      int distance_bin = (int) (distance / bin_spacing);
      for (int parameter_bin = 0; parameter_bin <= distance_bin; parameter_bin++) {
        for (int element_bin = 0; element_bin < ngrids; element_bin++) {
          RNScalar parameter_value = (parameter_bin + 0.5) * bin_spacing;
          R3Point world_position = span.Point(parameter_value);
          RNScalar value = bins[distance_bin][parameter_bin][element_bin];
          value *= weight;
          if (splat_type == 1) {
             if (RNIsPositive(value, 1.0E-20)) value = log(value);
             else value = -100;
          }             
          else if (splat_type == 2) {
            value *= observations[distance_bin];
          }
          grids[element_bin]->RasterizeWorldPoint(world_position, value);
        }
      }

      // Update statistics
      span_count++;
    }
  }

  // Return success
  return span_count;
}



static R3Grid **
CreateGrids(PDBFile *file, RNArray<PDBAtom *> *ligand_atoms)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get PDB model from file
  if (file->NModels() < 1) return NULL;
  PDBModel *model = file->Model(0);
  if (model->NAtoms() == 0) return 0;

  // Construct world box for grids (must be square for shape matching)
  R3Point c = world_centroid;
  RNScalar r = world_radius;
  if (r == PDB_UNKNOWN) r = (ligand_name) ? 15 : 0.5 * model->BBox().LongestAxisLength();
  if (c[0] == PDB_UNKNOWN) c = (ligand_name) ? PDBCentroid(*ligand_atoms) : model->BBox().Centroid();
  R3Box world_box(c[0]-r, c[1]-r, c[2]-r, c[0]+r, c[1]+r, c[2]+r);

  // Determine grid resolution (must be power of 2 for shape matching)
  int res = 4;
  while (((world_box.XLength() / res) > grid_spacing) && (res < max_resolution))
    res *= 2;

  // Allocate array of grids
  int ngrids = nelements;
  R3Grid **grids = new R3Grid * [ ngrids ];
  if (!grids) {
    fprintf(stderr, "Unable to allocate array of grids\n");
    return NULL;
  }

  // Initialize grids
  for (int i = 0; i < ngrids; i++) {
    // Allocate grid
    grids[i] = new R3Grid(res, res, res);
    if (!grids[i]) {
      fprintf(stderr, "Unable to allocate grid\n");
      return NULL;
    }

    // Set transformation
    grids[i]->SetWorldToGridTransformation(world_box);
  }

  // Splat chords into grids
  int splat_count = SplatChords(file, grids, ngrids);

  // Print statistics
  if (print_verbose) {
    RNScalar L1 = 0;
    RNScalar L2_squared = 0;
    for (int i = 0; i < ngrids; i++) { 
      L1 += grids[i]->L1Norm(); 
      L2_squared += grids[i]->L2NormSquared(); 
    }
    RNScalar world_radius = world_box.XRadius();
    R3Point world_centroid = world_box.Centroid();
    printf("Created grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Grids = %d\n", ngrids);
    printf("  Grid Resolution = %d %d %d\n", res, res, res);
    printf("  World Spacing = %g\n", grids[0]->GridToWorldScaleFactor());
    printf("  World Centroid = %g %g %g\n", world_centroid[0], world_centroid[1], world_centroid[2]);
    printf("  World Radius = %g\n", world_radius);
    printf("  # Splats = %d\n", splat_count);
    printf("  L1Norm = %g\n", L1);
    printf("  L2Norm = %g\n", sqrt(L2_squared));
    fflush(stdout);
  }

  // Return grids
  return grids;
}



static int 
BlurGrids(R3Grid **grids, int ngrids, RNScalar world_sigma)
{
  // Check sigma
  if (world_sigma == 0) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Blur the grid
  for (int i = 0; i < ngrids; i++) {
    grids[i]->Blur(world_sigma * grids[i]->WorldToGridScaleFactor());
  }

  // Print statistics
  if (print_verbose) {
    RNScalar L1 = 0;
    RNScalar L2_squared = 0;
    for (int i = 0; i < ngrids; i++) { 
      L1 += grids[i]->L1Norm(); 
      L2_squared += grids[i]->L2NormSquared(); 
    }
    printf("Blurred grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  World Sigma = %g\n", world_sigma);
    printf("  L1Norm = %g\n", L1);
    printf("  L2Norm = %g\n", sqrt(L2_squared));
    fflush(stdout);
  }

  // Return success
  return 1;
}



static RNScalar 
GridVolume(R3Grid *grid, RNScalar threshold, RNBoolean largest_connected_component_only)
{
  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Count number of grid samples at or above threshold
  int grid_count = 0;
  if (largest_connected_component_only) {
    const int max_components = 1024;
    int component_sizes[max_components];
    int ncomponents = grid->ConnectedComponents(threshold, max_components, NULL, component_sizes);
    for (int i = 0; i < ncomponents; i++) {
      if (component_sizes[i] > grid_count) grid_count = component_sizes[i];
    }
  }
  else {
    for (int i = 0; i < grid->NEntries(); i++) {
      if (grid->GridValue(i) >= threshold) grid_count++;
    }
  }

  // Return volume of grid samples at or above threshold
  return grid_count * volume_scale_factor;
}



static RNScalar 
GridThreshold(R3Grid *grid, RNScalar target_volume, RNBoolean largest_connected_component_only)
{
  // Determine ranges of thresholds and volumes
  RNInterval threshold_range = grid->Range();
  if (threshold_range.Min() == 0) threshold_range.SetMin(1.0E-20);
  RNScalar min_volume = GridVolume(grid, threshold_range.Max(), largest_connected_component_only);
  RNScalar max_volume = GridVolume(grid, threshold_range.Min(), largest_connected_component_only);
  RNInterval volume_range(min_volume, max_volume); 

  // Check if target volume is within range
  if (target_volume >= volume_range.Max()) return threshold_range.Min();
  if (target_volume <=volume_range.Min()) return threshold_range.Max();

  // Binary search for threshold that provides target volume
  for (int i = 0; i < 1000; i++) {
    RNScalar t = (target_volume - volume_range.Min()) / (volume_range.Max() - volume_range.Min());
    RNScalar threshold = threshold_range.Max() - t * (threshold_range.Max() - threshold_range.Min());
    RNScalar volume = GridVolume(grid, threshold, largest_connected_component_only);
    if (RNIsGreater(volume, target_volume, 1)) {
      threshold_range.SetMin(threshold);
      volume_range.SetMax(volume);
    }
    else if (RNIsLess(volume, target_volume, 1)) {
      threshold_range.SetMax(threshold);
      volume_range.SetMin(volume);
    }
    else return threshold;
  }

  // Return middle of remaining range
  return threshold_range.Mid();
}



static void
MaskGridsToLigand(PDBFile *file, RNArray<PDBAtom *> *ligand_atoms, R3Grid **grids, int ngrids)
{
  // Check the ligand atoms
  if (!ligand_atoms) return;

  // Create mask
  R3Grid mask(grids[0]->XResolution(), grids[0]->YResolution(), grids[0]->ZResolution());
  mask.SetWorldToGridTransformation(grids[0]->WorldToGridTransformation());
  for (int i = 0; i < ligand_atoms->NEntries(); i++) {
    PDBAtom *atom = ligand_atoms->Kth(i);
    mask.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1);
  }

  // Mask the grids
  for (int i = 0; i < ngrids; i++) {
    grids[i]->Mask(mask);
  }
}



static void
MaskGridsToCavities(PDBFile *file, R3Grid **grids, int ngrids, 
  RNScalar min_offset_distance, RNScalar max_offset_distance, RNScalar min_cavity_volume)
{
  // Check grids
  if (ngrids == 0) return;

  // Get PDB model
  if (file->NModels() == 0) return;
  PDBModel *model = file->Model(0);

  // Create mask
  R3Grid mask(grids[0]->XResolution(), grids[0]->YResolution(), grids[0]->ZResolution());
  mask.SetWorldToGridTransformation(grids[0]->WorldToGridTransformation());
  mask.Clear(1);

  // Create min offset mask
  R3Grid min_offset_mask(mask.XResolution(), mask.YResolution(), mask.ZResolution());
  min_offset_mask.SetWorldToGridTransformation(mask.WorldToGridTransformation());
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (atom->IsHetAtom()) continue;
    RNScalar radius = atom->Radius();
    if (min_offset_distance > radius) radius = min_offset_distance;
    min_offset_mask.RasterizeWorldSphere(atom->Position(), radius, 1);
  }

  // Invert and apply min_offset_mask 
  min_offset_mask.Threshold(0.5, 1, 0);
  mask.Mask(min_offset_mask);

  // Create max offset mask
  R3Grid max_offset_mask(mask.XResolution(), mask.YResolution(), mask.ZResolution());
  max_offset_mask.SetWorldToGridTransformation(mask.WorldToGridTransformation());
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (atom->IsHetAtom()) continue;
    RNScalar radius = atom->Radius();
    if (max_offset_distance > radius) radius = max_offset_distance;
    max_offset_mask.RasterizeWorldSphere(atom->Position(), radius, 1);
  }

  // Apply max_offset_mask
  mask.Mask(max_offset_mask);

  // Keep only cavities large enough to hold a certain number of atoms
  RNScalar world_grid_cell_volume = mask.GridToWorldScaleFactor();
  world_grid_cell_volume = world_grid_cell_volume * world_grid_cell_volume * world_grid_cell_volume;
  int min_cavity_size = (int) (min_cavity_volume / world_grid_cell_volume);
  int *cavity_size = new int [ mask.NEntries() ];
  int *cavity_grid = new int [ mask.NEntries() ];
  int ncavities = mask.ConnectedComponents(0.5, mask.NEntries(), NULL, cavity_size, cavity_grid);
  RNScalar *mask_valuesp = (RNScalar *) mask.GridValues();
  for (int i = 0; i < mask.NEntries(); i++) {
    int cavity = cavity_grid[i];
    assert(cavity < ncavities);
    if (cavity < 0) continue;
    if (cavity_size[cavity] < min_cavity_size) *mask_valuesp = 0;
    mask_valuesp++;
  }
  delete [] cavity_size;
  delete [] cavity_grid;

  // Mask the grids
  for (int i = 0; i < ngrids; i++) {
    grids[i]->Mask(mask);
  }
}



static int 
MaskGrids(PDBFile *file, RNArray<PDBAtom *> *ligand_atoms, R3Grid **grids, int ngrids)
{
  // Check number of grids
  if (ngrids == 0) return 0;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Mask the grids
  if (mask_to_ligand && ligand_atoms) {
    // Mask the grids to cover only the ligand volume
    MaskGridsToLigand(file, ligand_atoms, grids, ngrids);
  }
  else {
    // Mask the grids to cover only cavities exterior to the protein
    MaskGridsToCavities(file, grids, ngrids, min_offset_distance, max_offset_distance, min_ligand_volume);
  }

  // Mask the grid to adhere to a target volume
  if (mask_to_target_volume) {
    for (int i = 0; i < ngrids; i++) {
      // Get grid
      R3Grid *grid = grids[i];

      // Determine scale factor to take number of grid cells to a volume
      RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
      RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

      // Determine target volume
      RNScalar target_volume = target_volume_default;
      if (ligand_atoms && (target_volume_from_ligand || target_volume_from_ligand_elements)) {
        // Create grid with ligand atoms of element type rasterized
        R3Grid ligand_grid(grid->XResolution(), grid->YResolution(), grid->ZResolution());
        ligand_grid.SetWorldToGridTransformation(grid->WorldToGridTransformation());
        for (int j = 0; j < ligand_atoms->NEntries(); j++) {
          PDBAtom *atom = ligand_atoms->Kth(j);
          PDBElement *element = atom->Element();
          if (!element) continue;
          int element_index = element - PDBelements;
          assert((element_index >= 0) && (element_index < PDBnelements));
          if ((element_index < 0) || (element_index >= ngrids)) element_index = 0;
          if ((target_volume_from_ligand_elements) && (i != element_index)) continue;
          ligand_grid.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
        }

        // Determine target volume from ligand
        target_volume = volume_scale_factor * ligand_grid.Cardinality();
      }

      // Determine threshold to achieve target volume
      RNScalar threshold = GridThreshold(grid, target_volume, largest_connected_component_only);

      // Mask the grid to threshold
      if (largest_connected_component_only) {
        // Allocate arrays to store connected component information
        int *component_sizes = new int [ grid->NEntries() ];
        int *component_ids = new int [ grid->NEntries() ];
        assert(component_sizes && component_ids);

        // Find connected components 
        int ncomponents = grid->ConnectedComponents(threshold, grid->NEntries(), NULL, component_sizes, component_ids);
        assert(ncomponents <= grid->NEntries());

        // Find maximum connected component
        int biggest_component = -1;
        int biggest_component_size = -1;
        for (int j = 0; j < ncomponents; j++) {
          if (component_sizes[j] > biggest_component_size) {
            biggest_component_size = component_sizes[j];
            biggest_component = j;
          }
        }

        // Assign '0' to values not in biggest component at threshold
        if (biggest_component >= 0) {
          int *component_idsp = component_ids;
          RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
          for (int j = 0; j < grid->NEntries(); j++) {
            if (*component_idsp != biggest_component) *grid_valuesp = 0;
            component_idsp++;
            grid_valuesp++;
          }
        }

        // Delete arrays to store connected component information
        delete [] component_sizes;
        delete [] component_ids;
      }
      else {
        // Threshold the grid
        grid->Threshold(threshold, 0, R3_GRID_KEEP_VALUE);
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    RNScalar L1 = 0;
    RNScalar L2_squared = 0;
    for (int i = 0; i < ngrids; i++) { 
      L1 += grids[i]->L1Norm(); 
      L2_squared += grids[i]->L2NormSquared(); 
    }
    printf("Masked grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Mask to ligand = %d\n", (mask_to_ligand && ligand_atoms) ? 1 : 0);
    printf("  Largest connected component only = %d\n", largest_connected_component_only);
    printf("  L1Norm = %g\n", L1);
    printf("  L2Norm = %g\n", sqrt(L2_squared));
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
NormalizeGrids(R3Grid **grids, int ngrids, int normalization_type)
{
  // Check arguments
  if (ngrids == 0) return 1;
  if (normalization_type == 0) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Check normalization type
  if ((normalization_type & 0x3) == 1) {
    // Normalize each grid 
    for (int i = 0; i < ngrids; i++) {
      grids[i]->Normalize();
    }
  }
  else if ((normalization_type & 0x3) == 2) {
    // Normalize each grid cell
    for (int x = 0; x < grids[0]->XResolution(); x++) {
      for (int y = 0; y < grids[0]->YResolution(); y++) {
        for (int z = 0; z < grids[0]->ZResolution(); z++) {
          // Compute sum of all grid entries at position
          RNScalar sum = 0;
          for (int i = 0; i < ngrids; i++) {
            RNScalar value = grids[i]->GridValue(x, y, z);
            sum += value * value;
          }
  
          // Rescale so that they sum to one
          if (sum > 0) {
            RNScalar l2norm = sqrt(sum);
            for (int i = 0; i < ngrids; i++) {
               grids[i]->SetGridValue(x, y, z, grids[i]->GridValue(x, y, z) / l2norm);
            }
          }
        }
      }
    }
  }
  else if ((normalization_type & 0x3) == 3) {
    // Compute sum of all grids
    RNScalar sum = 0;
    for (int i = 0; i < ngrids; i++) {
      sum += grids[i]->L2Norm();
    }

    // Rescale so that they cumulatively sum to one
    if (sum > 0) {
      for (int i = 0; i < ngrids; i++) {
        grids[i]->Divide(sum);
      }
    }
  }
  else {
    RNAbort("Invalid normalization type");
  }

  // Print statistics
  if (print_verbose) {
    printf("Normalized grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Normalization type = %d\n", normalization_type);
    printf("  Norm by element ...\n");
    RNScalar L1_sum = 0;
    RNScalar L2_squared_sum = 0;
    for (int i = 0; i < ngrids; i++) { 
      RNScalar L1 = grids[i]->L1Norm(); L1_sum += L1;
      RNScalar L2_squared = grids[i]->L2NormSquared(); L2_squared_sum += L2_squared;
      printf("    Norms %s = %g %g\n", PDBelements[i].Name(), L1, sqrt(L2_squared));
    }
    printf("  L1Norm = %g\n", L1_sum);
    printf("  L2Norm = %g\n", sqrt(L2_squared_sum));
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
WriteGrids(R3Grid **grids, int ngrids, const char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write grids
  for (int i = 0; i < ngrids; i++) {
    char name[256];
    strcpy(name, grid_name);
    char *namep = strstr(name, ".grd");
    if (!namep) namep = &name[strlen(name)];
    if (ngrids > 1) sprintf(namep, "-%s.grd", PDBelements[i].Name());
    int status = grids[i]->WriteFile(grid_name);
    if (!status) return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Wrote grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", ngrids * grids[0]->NEntries() * sizeof(RNScalar));
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
PrintPlot(R3Grid *grid, RNScalar *ligand_values, int ligand_nvalues, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = stdout;
  if (filename) {
    fp = fopen(filename, "w");
    if (!fp) {
      fprintf(stderr, "Unable to open site plot file: %s\n", filename);
      return 0;
    }
  }

  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Make sorted array of grid values
  int grid_nvalues = grid->NEntries();
  RNScalar *grid_values = new RNScalar [ grid_nvalues ];
  assert(grid_values);
  for (int i = 0; i < grid_nvalues; i++) grid_values[i] = grid->GridValue(i);
  qsort(grid_values, grid_nvalues, sizeof(RNScalar), RNCompareScalars);

  // Print plot
  fprintf(fp, "0 0\n");
  int grid_index = grid_nvalues-1;
  for (int ligand_index = ligand_nvalues-1; ligand_index >= 0; ligand_index--) {
    // Check grid index
    if (grid_index <= 0) break;

    // Check ligand value
    if (ligand_values[ligand_index] == 0) break;

    // Find next grid value less or equal to ligand value
    for (; grid_index >= 0; grid_index--) {
      if (RNIsLess(grid_values[grid_index], ligand_values[ligand_index], 1.0E-8)) break;
    }

    // Check if did not find any of grid 
    if (grid_index < 0) break;

    // Determine fraction of ligand values considered
    RNScalar ligand_fraction = (RNScalar) (ligand_nvalues - ligand_index) / (RNScalar) ligand_nvalues;

    // Determine number of grid cells with value greater of equal to ligand value
    int grid_count = grid_nvalues - grid_index;

    // Convert number of grid cells to a volume
    RNScalar grid_volume = grid_count * volume_scale_factor;

    // Print plot numbers
    fprintf(fp, "%g %g\n", ligand_fraction, grid_volume);
  }

  // Close file
  if (filename) fclose(fp);

  // Deallocate array of grid values
  delete [] grid_values;

  // Print statistics
  if (print_verbose) {
    printf("Wrote plot ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", ligand_nvalues);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
AnalyzeLigand(PDBFile *file, RNArray<PDBAtom *> *ligand_atoms, R3Grid **grids, int ngrids)
{
  // Check grids
  if (ngrids == 0) return 0;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Print leader message
  if (print_verbose) {
    printf("Analyzing grids/ligand ...\n");
    fflush(stdout);
  }

  // Compute count and sum for each grid
  int total_nsamples = 0;
  RNScalar total_sum = 0;
  int *per_grid_nsamples = new int [ ngrids ];
  RNScalar *per_grid_sum = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) per_grid_nsamples[g] = 0;
  for (int g = 0; g < ngrids; g++) per_grid_sum[g] = 0;
  for (int g = 0; g < ngrids; g++) {
    for (int i = 0; i < grids[g]->XResolution(); i++) {
      for (int j = 0; j < grids[g]->YResolution(); j++) {
        for (int k = 0; k < grids[g]->ZResolution(); k++) {
          RNScalar value = grids[g]->GridValue(i, j, k);
          if (value == 0) continue;
          total_sum += value;
          total_nsamples++;
          per_grid_sum[g] += value;
          per_grid_nsamples[g]++;
        }
      }
    }
  }

  // Check number of samples
  if (total_nsamples == 0) return 0;

  // Compute avg for each grid
  RNScalar total_avg = total_sum / total_nsamples;
  RNScalar *per_grid_avg = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) {
    if (per_grid_nsamples[g] == 0) per_grid_avg[g] = 0;
    else per_grid_avg[g] = per_grid_sum[g] / per_grid_nsamples[g];
  }

  // Compute sum of squared errors for each grid
  RNScalar total_sse = 0;
  RNScalar *per_grid_sse = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) per_grid_sse[g] = 0;
  for (int g = 0; g < ngrids; g++) {
    for (int i = 0; i < grids[g]->XResolution(); i++) {
      for (int j = 0; j < grids[g]->YResolution(); j++) {
        for (int k = 0; k < grids[g]->ZResolution(); k++) {
          RNScalar value = grids[g]->GridValue(i, j, k);
          RNScalar total_delta = value - total_avg;
          total_sse += total_delta * total_delta;
          RNScalar per_grid_delta = value - per_grid_avg[g];
          per_grid_sse[g] += per_grid_delta * per_grid_delta;
        }
      }
    }
  }

  // Compute standard deviation for each grid
  RNScalar total_stddev = sqrt(total_sse / total_nsamples);
  RNScalar *per_grid_stddev = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) {
    if (per_grid_nsamples[g] == 0) per_grid_stddev[g] = 0;
    else per_grid_stddev[g] = sqrt(per_grid_sse[g] / per_grid_nsamples[g]);
  }

  // Print grid scores
  if (print_verbose) {
    printf("  Grid Statistics ...\n");
    printf("    Grid Statistics %s = %12g ( %12g )  %12g %12g\n", "All",  
      total_sum, 100.0, total_avg, total_stddev);
    for (int g = 0; g < ngrids; g++) {
      printf("    Grid Statistics %s = %12g ( %12g )  %12g %12g\n", PDBelements[g].Name(), 
        per_grid_sum[g], 100 * per_grid_sum[g] / total_sum, per_grid_avg[g], per_grid_stddev[g]);
    }
    fflush(stdout);
  }

  // Compute grid scores
  int total_element_natoms = 0;
  RNScalar total_element_score = 0;
  int *per_grid_element_natoms = new int [ ngrids ];
  RNScalar *per_grid_element_score = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) per_grid_element_natoms[g] = 0; 
  for (int g = 0; g < ngrids; g++) per_grid_element_score[g] = 0; 
  for (int i = 0; i < ligand_atoms->NEntries(); i++) {
    PDBAtom *atom = ligand_atoms->Kth(i);

    // Get element
    PDBElement *element = atom->Element();
    if (!element) continue;
    int element_index = element - PDBelements;
    assert((element_index >= 0) && (element_index < PDBnelements));
    if ((element_index < 0) || (element_index >= ngrids)) element_index = 0;

    // Log score for element at atom position
    if (per_grid_stddev[element_index] > 0) {
      const R3Point& position = atom->Position();
      RNScalar value = grids[element_index]->WorldValue(position);
      RNScalar score = (value - per_grid_avg[element_index]) / per_grid_stddev[element_index];
      total_element_score += score;
      total_element_natoms++;
      per_grid_element_score[element_index] += score;
      per_grid_element_natoms[element_index]++;
    }
  }

  // Divide scores by number of atoms to get average score per atom
  if (total_element_natoms > 0) total_element_score /= total_element_natoms;
  for (int g = 0; g < ngrids; g++) {
    if (per_grid_element_natoms[g] > 0) per_grid_element_score[g] /= per_grid_element_natoms[g];
  }

  // Print statistics
  if (print_verbose) {
    printf("  Element Scores ...\n");
    printf("    Element Score %s = %12g\n", "All",  total_element_score);
    for (int g = 0; g < ngrids; g++) 
      printf("    Element Score %s = %12g\n", PDBelements[g].Name(), per_grid_element_score[g]);
    fflush(stdout);
  }

  // Compute per cell statistics
  int total_cell_natoms = 0;
  RNScalar total_cell_score = 0;
  int *per_grid_cell_natoms = new int [ ngrids ];
  RNScalar *per_grid_cell_score = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) per_grid_cell_natoms[g] = 0; 
  for (int g = 0; g < ngrids; g++) per_grid_cell_score[g] = 0; 
  for (int i = 0; i < ligand_atoms->NEntries(); i++) {
    PDBAtom *atom = ligand_atoms->Kth(i);

    // Get element
    PDBElement *element = atom->Element();
    if (!element) continue;
    int element_index = element - PDBelements;
    assert((element_index >= 0) && (element_index < PDBnelements));
    if ((element_index < 0) || (element_index >= ngrids)) element_index = 0;

    // Compute statistics for cells at atom position
    RNScalar cell_sum = 0;
    for (int g = 0; g < ngrids; g++) cell_sum += grids[g]->WorldValue(atom->Position());
    RNScalar cell_avg = cell_sum / ngrids;
    RNScalar cell_sse = 0;
    for (int g = 0; g < ngrids; g++) { RNScalar delta = grids[g]->WorldValue(atom->Position()) - cell_avg; cell_sse += delta * delta; }
    RNScalar cell_stddev = sqrt(cell_sse / ngrids);
    if (cell_stddev > 0) {
      RNScalar score = (grids[element_index]->WorldValue(atom->Position()) - cell_avg) / cell_stddev;
      total_cell_score += score;
      total_cell_natoms++;
      per_grid_cell_score[element_index] += score;
      per_grid_cell_natoms[element_index]++;
    }
  }

  // Divide cell score by number of atoms to get average score per atom
  if (total_cell_natoms > 0) total_cell_score /= total_cell_natoms;
  for (int g = 0; g < ngrids; g++) {
    if (per_grid_cell_natoms[g] > 0) per_grid_cell_score[g] /= per_grid_cell_natoms[g];
  }

  // Print statistics
  if (print_verbose) {
    printf("  Cell Scores ...\n");
    printf("    Cell Score %s = %12g\n", "All",  total_cell_score);
    for (int g = 0; g < ngrids; g++) 
      printf("    Cell Score %s = %12g\n", PDBelements[g].Name(), per_grid_cell_score[g]);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: chd2grd pdbfile chddir griddir [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-asa")) { 
        argc--; argv++; asa_name = *argv; 
      }
      else if (!strcmp(*argv, "-consurf")) { 
        argc--; argv++; consurf_name = *argv; 
      }
      else if (!strcmp(*argv, "-ligand")) { 
        argc--; argv++; ligand_name = *argv; 
      }
      else if (!strcmp(*argv, "-spacing")) { 
        argc--; argv++; grid_spacing = atof(*argv); 
      }
      else if (!strcmp(*argv, "-radius")) { 
        argc--; argv++; world_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-center")) { 
        argc--; argv++; world_centroid[0] = atof(*argv); 
        argc--; argv++; world_centroid[1] = atof(*argv); 
        argc--; argv++; world_centroid[2] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; max_resolution = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-sigma")) { 
        argc--; argv++; world_sigma = atof(*argv); 
      }
      else if (!strcmp(*argv, "-splat_type")) { 
        argc--; argv++; splat_type = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-weight_spans_by_surface_area")) { 
        weight_spans_by_surface_area = TRUE;
      }
      else if (!strcmp(*argv, "-weight_spans_by_conservation")) { 
        weight_spans_by_conservation = TRUE;
      }
      else if (!strcmp(*argv, "-normalization")) { 
        argc--; argv++; normalization_type = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-min_offset_distance")) { 
        argc--; argv++; min_offset_distance = atof(*argv); 
      }
      else if (!strcmp(*argv, "-max_offset_distance")) { 
        argc--; argv++; max_offset_distance = atof(*argv); 
      }
      else if (!strcmp(*argv, "-min_ligand_volume")) { 
        argc--; argv++; min_ligand_volume = atof(*argv); 
      }
      else if (!strcmp(*argv, "-mask_to_ligand")) { 
        mask_to_ligand = TRUE; 
      }
      else if (!strcmp(*argv, "-mask_to_target_volume")) { 
        mask_to_target_volume = TRUE; 
      }
      else if (!strcmp(*argv, "-target_volume")) { 
        argc--; argv++; target_volume_default = atof(*argv); 
      }
      else if (!strcmp(*argv, "-target_volume_from_ligand")) { 
        target_volume_from_ligand = TRUE; 
      }
      else if (!strcmp(*argv, "-target_volume_from_ligand_elements")) { 
        target_volume_from_ligand_elements = TRUE; 
      }
      else if (!strcmp(*argv, "-largest_connected_component_only")) { 
        largest_connected_component_only = TRUE; 
      }
      else if (!strcmp(*argv, "-debug")) { 
        print_debug = 1; 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else if (!chd_name) chd_name = *argv;
      else if (!grid_name) grid_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check chd name
  if (!chd_name) {
    fprintf(stderr, "You did not specify a grid base filename.\n");
    return 0;
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check world radius
  if (RNIsZero(world_radius)) {
    fprintf(stderr, "Invalid world radius: %g\n", world_radius);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read pdb
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Read accessible surface area file
  if (asa_name) {
    int status1 = ReadASAFile(file, asa_name);
    if (!status1) exit(-1);
  }

  // Read conservation file
  if (consurf_name) {
    int status1 = ReadConsurfFiles(file, consurf_name);
    if (!status1) exit(-1);
  }

  // Read chord histogram file 
  int status2 = ReadChordFile(chd_name);
  if (!status2) exit(-1);

  // Create array of ligand atoms
  RNArray<PDBAtom *> *ligand_atoms = FindLigandAtoms(file, ligand_name);
  if (!ligand_atoms) exit(-1);

  // Create grids
  int ngrids = nelements;
  R3Grid **grids = CreateGrids(file, ligand_atoms);
  if (!grids) exit(-1);

  // Blur grids
  int status3 = BlurGrids(grids, ngrids, world_sigma);
  if (!status3) exit(-1);

  // Mask grids
  int status4 = MaskGrids(file, ligand_atoms, grids, ngrids);
  if (!status4) exit(-1);

  // Normalize grids
  int status5 = NormalizeGrids(grids, ngrids, normalization_type);
  if (!status5) exit(-1);

  // Analyze grid
  int status6 = AnalyzeLigand(file, ligand_atoms, grids, ngrids);
  if (!status6) exit(-1);

  // Write grids
  int status7= WriteGrids(grids, ngrids, grid_name);
  if (!status7) exit(-1);

  // Return success
  return 0;
}
