// Source file for the site analyzer program



// Include files 

#include "R3Shapes/R3Shapes.h"
#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *grid_name = NULL;
static char *pdb_name = NULL;
static char *ligand_name = NULL;
static char *dart_name = NULL;
static int element = 0;
static int nelements = 1;
static RNBoolean local_maxima_only = FALSE;
static RNBoolean print_verbose = FALSE;



static R3Grid *
ReadGrid(char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    RNFail("Unable to allocated grid");
    return NULL;
  }

  // Read grid 
  if (!grid->ReadFile(grid_name)) {
    RNFail("Unable to read grid file %s", grid_name);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Dimensions = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Minimum = %g\n", grid->Minimum());
    printf("  Maximum = %g\n", grid->Maximum());
    printf("  Volume = %g\n", grid->Volume());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return success
  return grid;
}



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) {
    RNFail("Unable to read PDB file: %s", filename);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static RNArray<PDBAtom *> *
FindLigandAtoms(PDBFile *file, const char *ligand_name, int element, int nelements)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int residue_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);

  // Allocate array of atoms
  RNArray<PDBAtom *> *ligand_atoms = new RNArray<PDBAtom *>();
  if (!ligand_atoms) {
    fprintf(stderr, "Unable to allocate array of ligand atoms.\n");
    return NULL;
  }

  // Check ligand name
  if (ligand_name) {
    // Find ligand
    PDBResidue *ligand = file->FindResidue(ligand_name);
    if (!ligand) {
      fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, pdb_name); 
      return NULL;
    }

    // Add atoms of ligand to array
    residue_count = 1;
    for (int j = 0; j < ligand->NAtoms(); j++) {
      PDBAtom *atom = ligand->Atom(j);
      if (!atom->Element()) continue;
      int element_index = atom->Element() - PDBelements;
      assert((element_index >= 0) && (element_index < PDBnelements));
      if ((element_index < 0) || (element_index >= nelements)) element_index = 0;
      if (element_index != element) continue;
      ligand_atoms->Insert(atom);
    }
  }
  else {
    // Add all hetatoms to array
    for (int i = 0; i < model->NResidues(); i++) {
      PDBResidue *residue = model->Residue(i);
      RNBoolean found = FALSE;
      for (int j = 0; j < residue->NAtoms(); j++) {
        PDBAtom *atom = residue->Atom(j);
        if (!atom->IsHetAtom()) continue;
        if (!atom->Element()) continue;
        int element_index = atom->Element() - PDBelements;
        assert((element_index >= 0) && (element_index < PDBnelements));
        if ((element_index < 0) || (element_index >= nelements)) element_index = 0;
        if (element_index != element) continue;
        ligand_atoms->Insert(atom);
        found = TRUE;
      }
      if (found) residue_count++;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Found ligand atoms ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Element = %d of %d\n", element, nelements);
    printf("  # Ligands = %d\n", residue_count);
    printf("  # Atoms = %d\n", ligand_atoms->NEntries());
    fflush(stdout);
  }

  // Return ligand atoms
  return ligand_atoms;
}



int 
MaskToLocalMaxima(R3Grid *grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Keep only local maxima, make others zero
  for (int k = 0; k < grid->ZResolution(); k++) {
    for (int j = 0; j < grid->YResolution(); j++) {
      for (int i = 0; i < grid->XResolution(); i++) {
        RNScalar grid_value = grid->GridValue(i, j, k);
        for (int t = k-1; t <= k+1; t++) {
          if ((t < 0) || (t >= grid->ZResolution())) continue;
          for (int s = j-1; s <= j+1; s++) {
            if ((s < 0) || (s >= grid->YResolution())) continue;
            for (int r = i-1; r <= i+1; r++) {
              if ((r < 0) || (r >= grid->XResolution())) continue;
              if ((r == i) && (s == j) && (t == k)) continue;
              RNScalar neighbor_value = grid->GridValue(r, s, t);
              if (neighbor_value > grid_value) {
                grid->SetGridValue(i, j, k, 0);
                goto exitinteriorloops;
              }
            }
          }
        }
exitinteriorloops:;
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Masked grid to local maxima ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Maxima = %d\n", grid->Cardinality());
    fflush(stdout);
  }

  // Return success
  return 1;
}



int
PrintDarts(R3Grid *grid, PDBFile *file, RNArray<PDBAtom *> *ligand_atoms, const char *dart_name)
{
  // Variables for darts
  static const int max_darts = 100;
  static RNScalar dart_values[max_darts];
  static R3Point dart_positions[max_darts];
  static PDBResidue *dart_residues[max_darts];
  static PDBAtom *dart_atoms[max_darts];
  static RNScalar dart_residue_distances[max_darts];
  static RNScalar dart_atom_distances[max_darts];
  static RNBoolean dart_hits[max_darts];
  int ndarts = 0;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Check ligand atoms
  int hit_count = 0;
  int first_hit = -1;
  if (!ligand_atoms->IsEmpty()) {
    // Compute voronoi diagram of ligand atoms
    R3Grid voronoi(grid->XResolution(), grid->YResolution(), grid->ZResolution());
    voronoi.SetWorldToGridTransformation(grid->WorldToGridTransformation());
    for (int i = 0; i < ligand_atoms->NEntries(); i++) {
      PDBAtom *atom = ligand_atoms->Kth(i);
      if (!atom->IsHetAtom()) continue;
      R3Point grid_position = voronoi.GridPosition(atom->Position());
      int x = (int) (grid_position[0] + 0.5);
      if ((x < 0) || (x >= grid->XResolution())) continue;
      int y = (int) (grid_position[1] + 0.5);
      if ((y < 0) || (y >= grid->YResolution())) continue;
      int z = (int) (grid_position[2] + 0.5);
      if ((z < 0) || (z >= grid->ZResolution())) continue;
      voronoi.SetGridValue(x, y, z, i + 0.5);
    }
    voronoi.Voronoi();

    // Compute darts
    R3Grid dart_grid(*grid);
    for (int i = 0; i < max_darts; i++) {
      // Find maximum value
      int maximum_index = -1;
      RNScalar maximum_value = 0;
      const RNScalar *dart_grid_valuesp = dart_grid.GridValues();
      for (int j = 0; j < dart_grid.NEntries(); j++) {
        if (*dart_grid_valuesp > maximum_value) { maximum_value = *dart_grid_valuesp; maximum_index = j; }
        dart_grid_valuesp++;
      }

      // Check if found value greater than zero
      if (maximum_index < 0) break;

      // Create dart
      int x, y, z;
      dart_grid.IndexToIndices(maximum_index, x, y, z);
      R3Point dart_position = dart_grid.WorldPosition(x, y, z);
      int closest_atom_id = (int) voronoi.GridValue(maximum_index);
      PDBAtom *dart_atom = ligand_atoms->Kth(closest_atom_id);
      PDBResidue *dart_residue = dart_atom->Residue();
      RNLength dart_residue_distance = R3Distance(dart_position, dart_residue->Centroid());
      RNLength dart_atom_distance = R3Distance(dart_position, dart_atom->Position());
      RNBoolean dart_hit = (dart_atom_distance < dart_atom->Radius());
      dart_values[ndarts] = maximum_value;
      dart_positions[ndarts] = dart_position;
      dart_residues[ndarts] = dart_residue;
      dart_atoms[ndarts] = dart_atom;
      dart_residue_distances[ndarts] = dart_residue_distance;
      dart_atom_distances[ndarts] = dart_atom_distance;
      dart_hits[ndarts] = dart_hit;
      if (dart_hit && (first_hit == -1)) first_hit = i + 1;
      if (dart_hit) hit_count++;
      ndarts++;

      // Mask area around dart 
      dart_grid.RasterizeWorldSphere(dart_position, 3, -maximum_value);
    }
  }

  // Open dart file
  FILE *fp = fopen(dart_name, "w");
  if (!fp) {
    fprintf(stderr, "Unable to open dart file: %s\n", dart_name);
    return 0;
  }

  // Print darts
  fprintf(fp, "%d %d %d\n", ndarts, hit_count, first_hit);
  for (int i = 0; i < ndarts; i++) {
    fprintf(fp, "DART %d   %g    %g %g %g    %d %d    %g %g    %d %d  \n", 
            i+1, dart_values[i], dart_positions[i].X(), dart_positions[i].Y(), dart_positions[i].Z(), 
            dart_residues[i]->ID(), dart_atoms[i]->ID(), dart_residue_distances[i], dart_atom_distances[i], 
            (dart_hits[i]) ? 1 : 0, hit_count);
  }

  // Close dart file
  fclose(fp);

  // Print message
  if (print_verbose) {
    printf("Created darts ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Darts = %d\n", ndarts);
    printf("  First Hit %s = %d\n", PDBelements[element].Name(), first_hit);
    fflush(stdout);
  }

  // Return number of darts created
  return ndarts;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if ((argc == 2) && (*argv[1] == '-')) {
    printf("Usage: grd2drt gridfile pdbfile dartfile [-ligand ligandname]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = TRUE; 
      else if (!strcmp(*argv, "-ligand")) { argc--; argv++; ligand_name = *argv; }
      else if (!strcmp(*argv, "-local_maxima_only")) local_maxima_only = TRUE; 
      else if (!strcmp(*argv, "-element")) { 
        argc--; argv++; element = atoi(*argv); 
        argc--; argv++; nelements = atoi(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
    else {
      if (!grid_name) grid_name = *argv;
      else if (!pdb_name) pdb_name = *argv;
      else if (!dart_name) dart_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check dart filename
  if (!dart_name) {
    fprintf(stderr, "You did not specify a dart file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read grid
  R3Grid *grid = ReadGrid(grid_name);
  if (!grid) exit(-1);

  // Read pdb
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Find ligand atoms
  RNArray<PDBAtom *> *ligand_atoms = FindLigandAtoms(file, ligand_name, element, nelements);
  if (!ligand_atoms) exit(-1);

  // Mask grid to include only local maxima
  if (local_maxima_only) {
    int status1 = MaskToLocalMaxima(grid);
    if (!status1) exit(-1);
  }

  // Print darts
  int status2 = PrintDarts(grid, file, ligand_atoms, dart_name);
  if (!status2) exit(-1);

  // Return success 
  return 0;
}



