// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"



// Program variables

static RNArray<char *> grid_names;
static char *gdt_name = NULL;
static RNScalar sigma = 0;
static int normalization_type = 0;
static int print_verbose = 0;



static R3Grid *
ReadGrid(char *grd_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate a grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    RNFail("Unable to allocate grid");
    return NULL;
  }

  // Read grid
  int status = grid->ReadFile(grd_name);
  if (!status) {
    RNFail("Unable to read grid file %s", grd_name);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Voxels = %d %d% d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    fflush(stdout);
  }

  // Return success
  return grid;
}



static int
ComputeGEDT(R3Grid *grid, RNScalar sigma)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Check sigma
  if (RNIsPositive(sigma)) {
    // Replace each grid value with Gaussian of what was there before
    grid->SquaredDistanceTransform();
    const RNScalar sqrt_two_pi = sqrt(RN_TWO_PI);
    RNScalar a = sqrt_two_pi * sigma;
    RNScalar fac = 1.0 / (a * a * a);
    RNScalar denom = -2.0 * sigma * sigma;
    RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
    for (int i = 0; i < grid->NEntries(); i++) {
      RNScalar value = *grid_valuesp;
      *grid_valuesp = fac * exp( value / denom );
      grid_valuesp++;
    }
  }

  // Normalize descriptor
  if ((normalization_type % 2) == 1) grid->Normalize();

  // Print statistics
  if (print_verbose) {
    printf("Computed GEDT ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
NormalizeDescriptor(int nvalues, float *values, int normalization_type)
{
  // Check if normalizing
  if (normalization_type < 2) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Compute l2norm
  float l2norm = 0;
  for (int i = 0; i < nvalues; i++)
    l2norm += values[i] * values[i];
  l2norm = sqrt(l2norm);

  // Normalize descriptor
  if (l2norm > 0) {
    for (int i = 0; i < nvalues; i++)
      values[i] /= l2norm;
  }

  // Print statistics
  if (print_verbose) {
    printf("Normalized spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Normalization type = %d\n", normalization_type);
    printf("  L2Norm = %g\n", l2norm);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
WriteDescriptor(float *values, int nvalues, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(filename, "wb");
  if (!fp) {
    fprintf(stderr, "Unable to open SHD file %s\n", filename);
    return 0;
  }

  // Write number of values
  if (fwrite(&nvalues, sizeof(int), 1, fp) != 1) {
    fprintf(stderr, "Unable to write number of values to SHD file %s\n", filename);
    return 0;
  }

  // Write values
  if (fwrite(values, sizeof(float), nvalues, fp) != (unsigned long) nvalues) {
    fprintf(stderr, "Unable to write values to SHD file %s\n", filename);
    return 0;
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", sizeof(int) + nvalues * sizeof(float));
    fflush(stdout);
  }

  // Return number of values written
  return nvalues;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: grd2gdt gridfile gdtfile [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-normalize")) { argc--; argv++; normalization_type = atoi(*argv); }
      else if (!strcmp(*argv, "-sigma")) { argc--; argv++; sigma = atof(*argv); }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    else {
      grid_names.Insert(*argv);
    }
    argv++; argc--;
  }

  // Check grid filenames
  if (grid_names.NEntries() < 1) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check gdt filename
  if (grid_names.NEntries() < 2) {
    fprintf(stderr, "You did not specify a gdt file.\n");
    return 0;
  }

  // Get gdt filename
  gdt_name = grid_names.Tail();
  grid_names.RemoveTail();

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Add grids to descriptor
  int grid_size = 0;
  int descriptor_size = 0;
  float *descriptor = NULL;
  for (int i = 0; i< grid_names.NEntries(); i++) {
    char *grid_name = grid_names.Kth(i);

    // Read grid file
    R3Grid *grid = ReadGrid(grid_name);
    if (!grid) exit(-1);

    // Replace grid with GEDT 
    int status = ComputeGEDT(grid, sigma);
    if (!status) exit(-1);

    // Allocate descriptor
    if (!descriptor) {
      grid_size = grid->NEntries();
      descriptor_size = grid_names.NEntries() * grid_size * sizeof(float);
      descriptor = new float [ descriptor_size ];
      assert(descriptor);
    }
    else {
      // Check grid size
      if (grid->NEntries() != grid_size) {
        fprintf(stderr, "Grid %s has a different size (%d vs. %d)\n", grid_name, grid->NEntries(), grid_size);
        exit(-1);
      }
    }

    // Add grid to descriptor
    const RNScalar *gridp = grid->GridValues();
    float *descriptorp = &descriptor[i * grid_size];
    for (int j = 0; j < grid_size; j++) {
      *descriptorp = *gridp;
      descriptorp++;
      gridp++;
    }
  }

  // Normalize descriptor
  int status1 = NormalizeDescriptor(descriptor_size, descriptor, normalization_type);
  if (!status1) exit(-1);

  // Write descriptor
  int status2 = WriteDescriptor(descriptor, descriptor_size, gdt_name);
  if (!status2) exit(-1);

  // Return success
  return 0;
}
