// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"



// Program variables

static char *grid_name = NULL;
static char *gedt_name = NULL;
static RNScalar gedt_sigma = sqrt(8.0);
static int normalize = 0;
static int print_verbose = 0;



static R3Grid *
ReadGrid(char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate a grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    RNFail("Unable to allocate grid");
    return NULL;
  }

  // Read grid
  int status = grid->ReadFile(grid_name);
  if (!status) {
    RNFail("Unable to read grid file %s", grid_name);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Voxels = %d %d% d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    fflush(stdout);
  }

  // Return success
  return grid;
}



static R3Grid *
CreateGEDTGrid(R3Grid *grid1)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate spherical grid
  R3Grid *grid = new R3Grid(*grid1);
  if (!grid) {
    fprintf(stderr, "Unable to allocate spherical grid\n");
    exit(-1);
  }

  // Check sigma
  if (RNIsPositive(gedt_sigma)) {
    // Fill samples with Gaussian of DT
    const RNScalar sqrt_two_pi = sqrt(RN_TWO_PI);
    RNScalar a = sqrt_two_pi * gedt_sigma;
    RNScalar fac = 1.0 / (a * a * a);
    RNScalar denom = -2.0 * gedt_sigma * gedt_sigma;
    grid1->SquaredDistanceTransform();
    RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
    RNScalar *grid1_valuesp = (RNScalar *) grid1->GridValues();
    for (int i = 0; i < grid->NEntries(); i++) 
      *(grid_valuesp++) = fac * exp( *(grid1_valuesp++) / denom );
  }

  // Normalize grid
  if (normalize) grid->Normalize();

  // Print statistics
  if (print_verbose) {
    printf("Created GEDT grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Voxels = %d %d% d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return spherical grid
  return grid;
}



static int 
WriteGrid(R3Grid *grid, const char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write grid
  int status = grid->WriteFile(grid_name);

  // Print statistics
  if (print_verbose) {
    printf("Wrote grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", status * sizeof(RNScalar));
    fflush(stdout);
  }

  // Return status
  return status;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: grd2gedt gridfile gridfile [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-normalize")) normalize = 1; 
      else if (!strcmp(*argv, "-sigma")) { argc--; argv++; gedt_sigma = atof(*argv); }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    else {
      if (!grid_name) grid_name = *argv;
      else if (!gedt_name) gedt_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    argv++; argc--;
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check gedt filename
  if (!gedt_name) {
    fprintf(stderr, "You did not specify a gedt file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read grid file
  R3Grid *grid = ReadGrid(grid_name);
  if (!grid) exit(-1);

  // Create GEDT grid
  R3Grid *gedt = CreateGEDTGrid(grid);
  if (!gedt) exit(-1);

  // Write grid
  int status = WriteGrid(gedt, gedt_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}
