// Source file for the site analyzer program



// Include files 

#include "R3Shapes/R3Shapes.h"
#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *grid_name = NULL;
static char *pdb_name = NULL;
static char *plot_name = NULL;
static char *ligand_name = NULL;
static char *element_name = NULL;
static RNScalar min_cavity_radius = 0;
static RNScalar min_cavity_volume = 0;
static RNBoolean largest_cavity_only = FALSE;
static RNBoolean print_verbose = FALSE;
static RNBoolean print_debug = FALSE;



static R3Grid *
ReadGrid(char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    RNFail("Unable to allocated grid");
    return NULL;
  }

  // Read grid 
  if (!grid->ReadFile(grid_name)) {
    RNFail("Unable to read grid file %s", grid_name);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Dimensions = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Minimum = %g\n", grid->Minimum());
    printf("  Maximum = %g\n", grid->Maximum());
    printf("  Volume = %g\n", grid->Volume());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return success
  return grid;
}



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) {
    RNFail("Unable to read PDB file: %s", filename);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static RNArray<PDBAtom *> *
FindLigandAtoms(PDBFile *file, const char *ligand_name, char *element_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int residue_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);

  // Allocate array of atoms
  RNArray<PDBAtom *> *ligand_atoms = new RNArray<PDBAtom *>();
  if (!ligand_atoms) {
    fprintf(stderr, "Unable to allocate array of ligand atoms.\n");
    return NULL;
  }

  // Check ligand name
  if (ligand_name) {
    // Find ligand
    PDBResidue *ligand = file->FindResidue(ligand_name);
    if (!ligand) {
      fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, pdb_name); 
      return NULL;
    }

    // Add atoms of ligand to array
    residue_count = 1;
    for (int j = 0; j < ligand->NAtoms(); j++) {
      PDBAtom *atom = ligand->Atom(j);
      if (element_name && (!atom->Element()|| strcmp(atom->Element()->Name(), element_name))) continue;
      ligand_atoms->Insert(atom);
    }
  }
  else {
    // Add all hetatoms to array
    for (int i = 0; i < model->NResidues(); i++) {
      PDBResidue *residue = model->Residue(i);
      RNBoolean found = FALSE;
      for (int j = 0; j < residue->NAtoms(); j++) {
        PDBAtom *atom = residue->Atom(j);
        if (!atom->IsHetAtom()) continue;
        if (element_name && (!atom->Element()|| strcmp(atom->Element()->Name(), element_name))) continue;
        ligand_atoms->Insert(atom);
        found = TRUE;
      }
      if (found) residue_count++;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Found ligand atoms ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Element = %s\n", (element_name) ? element_name : "All");
    printf("  # Ligands = %d\n", residue_count);
    printf("  # Atoms = %d\n", ligand_atoms->NEntries());
    fflush(stdout);
  }

  // Return ligand atoms
  return ligand_atoms;
}



static void
MaskToCavityRadius(R3Grid *mask, RNScalar min_cavity_radius)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Create mask
  R3Grid breadth_mask(*mask);

  // Erode and then dilate
  breadth_mask.Erode(min_cavity_radius);  
  breadth_mask.Dilate(min_cavity_radius);  

  // Apply mask
  mask->Mask(breadth_mask);

  // Print statistics
  if (print_debug) {
    printf("Masked to cavity radius .. \n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Min cavity radius = %g\n", min_cavity_radius);
    printf("  Cardinality = %d\n", mask->Cardinality());
    printf("  Volume = %g\n", mask->Volume());
    RNInterval grid_range = mask->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", mask->L1Norm());
    printf("  L2Norm = %g\n", mask->L2Norm());
    fflush(stdout);
  }
}



static void
MaskToCavityVolume(R3Grid *mask, RNScalar min_cavity_volume)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Create mask
  R3Grid cavity_mask(*mask);

  // Keep only cavities large enough to hold a certain volume
  RNScalar world_grid_cell_volume = cavity_mask.GridToWorldScaleFactor();
  world_grid_cell_volume = world_grid_cell_volume * world_grid_cell_volume * world_grid_cell_volume;
  int min_cavity_size = (int) (min_cavity_volume / world_grid_cell_volume);
  int max_cavities = cavity_mask.NEntries();
  int *cavity_size = new int [ max_cavities ];
  int *cavity_grid = new int [ cavity_mask.NEntries() ];
  int ncavities = cavity_mask.ConnectedComponents(1.0E-20, max_cavities, NULL, cavity_size, cavity_grid);
  RNScalar *mask_valuesp = (RNScalar *) cavity_mask.GridValues();
  for (int i = 0; i < cavity_mask.NEntries(); i++) {
    int cavity = cavity_grid[i];
    assert(cavity < ncavities);
    if (cavity < 0) { *mask_valuesp = 0; }
    else if (cavity_size[cavity] < min_cavity_size) { *mask_valuesp = 0; }
    else *mask_valuesp = 1;
    mask_valuesp++;
  }
  delete [] cavity_size;
  delete [] cavity_grid;

  // Apply mask
  mask->Mask(cavity_mask);

  // Print statistics
  if (print_debug) {
    printf("Masked to cavity volume .. \n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Min cavity volume = %g\n", min_cavity_volume);
    printf("  # Cavities = %d\n", ncavities);
    printf("  Cardinality = %d\n", mask->Cardinality());
    printf("  Volume = %g\n", mask->Volume());
    RNInterval grid_range = mask->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", mask->L1Norm());
    printf("  L2Norm = %g\n", mask->L2Norm());
    fflush(stdout);
  }
}



static void 
MaskToLargestCavity(R3Grid *mask)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate arrays to store cavity information
  int *cavity_sizes = new int [ mask->NEntries() ];
  int *cavity_ids = new int [ mask->NEntries() ];
  assert(cavity_sizes && cavity_ids);

  // Find connected cavities 
  int ncavities = mask->ConnectedComponents(1.0E-20, mask->NEntries(), NULL, cavity_sizes, cavity_ids);
  assert(ncavities <= mask->NEntries());

  // Find maximum connected cavity
  int biggest_cavity = -1;
  int biggest_cavity_size = -1;
  for (int j = 0; j < ncavities; j++) {
    if (cavity_sizes[j] > biggest_cavity_size) {
      biggest_cavity_size = cavity_sizes[j];
      biggest_cavity = j;
    }
  }

  // Assign '0' to values not in biggest cavity at threshold
  if (biggest_cavity >= 0) {
    int *cavity_idsp = cavity_ids;
    RNScalar *grid_valuesp = (RNScalar *) mask->GridValues();
    for (int j = 0; j < mask->NEntries(); j++) {
      if (*cavity_idsp != biggest_cavity) *grid_valuesp = 0;
      cavity_idsp++;
      grid_valuesp++;
    }
  }

  // Delete arrays to store connected cavity information
  delete [] cavity_sizes;
  delete [] cavity_ids;

  // Print statistics
  if (print_debug) {
    printf("Masked to largest cavity only .. \n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Cavities = %d\n", ncavities);
    printf("  Biggest cavity size = %d\n", biggest_cavity_size);
    printf("  Cardinality = %d\n", mask->Cardinality());
    printf("  Volume = %g\n", mask->Volume());
    RNInterval grid_range = mask->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", mask->L1Norm());
    printf("  L2Norm = %g\n", mask->L2Norm());
    fflush(stdout);
  }
}



static int 
PrintPlot(R3Grid *grid, PDBFile *file, RNArray<PDBAtom *> *ligand_atoms, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = stdout;
  if (filename) {
    fp = fopen(filename, "w");
    if (!fp) {
      fprintf(stderr, "Unable to open site plot file: %s\n", filename);
      return 0;
    }
  }

  // Write first entry
  fprintf(fp, "0 0\n");

  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Create ligand grid 
  R3Grid ligand_grid(grid->XResolution(), grid->YResolution(), grid->ZResolution());
  ligand_grid.SetWorldToGridTransformation(grid->WorldToGridTransformation());
  for (int i = 0; i < ligand_atoms->NEntries(); i++) {
    PDBAtom *atom = ligand_atoms->Kth(i);
    ligand_grid.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
  }

  // Compute ligand volume
  int ligand_nvalues = ligand_grid.Cardinality();
  if (ligand_nvalues <= 0) return 0;
  RNScalar ligand_volume = ligand_nvalues * volume_scale_factor;
  if (ligand_volume <= 0) return 0;

  // Fill array of ligand values
  int count = 0;
  RNScalar *ligand_values = new RNScalar [ ligand_nvalues ];
  assert(ligand_values);
  for (int i = 0; i < grid->NEntries(); i++) {
    if (ligand_grid.GridValue(i) == 0) continue;
    assert(count < ligand_nvalues);
    RNScalar value = grid->GridValue(i);
    ligand_values[count++] = value;
  }

  // Sort array of values to check in ligand
  qsort(ligand_values, ligand_nvalues, sizeof(RNScalar), RNCompareScalars);

  // Step through ligand values printing plot values
  int nsteps = 1;
  int max_steps = 20;
  int step_size = ligand_nvalues / max_steps;
  if (step_size <= 0) step_size = 1;
  RNScalar last_threshold = FLT_MAX;
  for (int ligand_index = ligand_nvalues - step_size; ligand_index >= 0; ligand_index -= step_size) {
    // Detemine threshold
    RNScalar threshold = ligand_values[ligand_index];
    if (threshold == last_threshold) continue;
    if (print_debug) printf("Threshold = %g\n", threshold);
    last_threshold = threshold;

    // Create a copy of the grid thresholded to value
    R3Grid site_grid(*grid);

    // Mask to threshold
    site_grid.Threshold(threshold, 0, R3_GRID_KEEP_VALUE);

    // Mask by cavity radius
    if (min_cavity_radius > 0) {
      MaskToCavityRadius(&site_grid, min_cavity_radius);
    }

    // Mask by cavity volume
    if (min_cavity_volume > 0) {
      MaskToCavityVolume(&site_grid, min_cavity_volume);
    }

    // Mask to largest cavity only
    if (largest_cavity_only) {
      MaskToLargestCavity(&site_grid);
    }

    // Determine volume of site
    RNScalar site_volume = site_grid.Volume();
    RNScalar site_fraction = site_volume / ligand_volume;

    // Determine volume of site-ligand intersection
    R3Grid intersection_grid(ligand_grid);
    intersection_grid.Mask(site_grid);
    RNScalar intersection_volume = intersection_grid.Volume();
    RNScalar intersection_fraction = intersection_volume / ligand_volume;

    // Print plot numbers
    fprintf(fp, "%g %g\n", intersection_fraction, site_fraction);
    nsteps++;
  }

  // Delete array of ligand values
  delete [] ligand_values;

  // Close file
  if (filename) fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote plot ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Ligand volume = %g\n", ligand_volume);
    printf("  # Values = %d\n", nsteps);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if ((argc == 2) && (*argv[1] == '-')) {
    printf("Usage: grd2plt gridfile pdbfile plotfile [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = TRUE; 
      else if (!strcmp(*argv, "-largest_cavity_only")) { largest_cavity_only = TRUE; }
      else if (!strcmp(*argv, "-ligand")) { argc--; argv++; ligand_name = *argv; }
      else if (!strcmp(*argv, "-min_cavity_radius")) { argc--; argv++; min_cavity_radius = atof(*argv); }
      else if (!strcmp(*argv, "-min_cavity_volume")) { argc--; argv++; min_cavity_volume = atof(*argv); }
      else if (!strcmp(*argv, "-element")) { argc--; argv++; element_name = *argv; }
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
    else {
      if (!grid_name) grid_name = *argv;
      else if (!pdb_name) pdb_name = *argv;
      else if (!plot_name) plot_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check plot filename
  if (!plot_name) {
    fprintf(stderr, "You did not specify a plot file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read grid
  R3Grid *grid = ReadGrid(grid_name);
  if (!grid) exit(-1);

  // Read pdb
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Find ligand atoms
  RNArray<PDBAtom *> *ligand_atoms = FindLigandAtoms(file, ligand_name, element_name);
  if (!ligand_atoms) exit(-1);

  // Print plot
  int status = PrintPlot(grid, file, ligand_atoms, plot_name);
  if (!status) exit(-1);

  // Return success 
  return 0;
}



