// Source file for grid alignment



// Include files 

#include "R3Grids/R3Grids.h"



// Program variables

static int ngrids = 0;
static const int max_grids = 32;
static char *grid_names[max_grids] = { NULL };
static char *wigner_grid_name = NULL;
static int normalization_type = 0;
static int print_verbose = 0;
static int print_debug = 0;



static R3Grid **
ReadGrids(char **grid_names, int ngrids)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid pointers
  R3Grid **grids = new R3Grid *[ngrids];
  if (!grids) {
    RNFail("Unable to allocate grid pointers");
    return NULL;
  }

  // Allocate and read grids
  for (int i = 0; i < ngrids; i++) {
    // Allocate grid
    grids[i] = new R3Grid();
    if (!grids[i]) {
      RNFail("Unable to allocate grid");
      return NULL;
    }

    // Read grid
    int status = grids[i]->ReadFile(grid_names[i]);
    if (!status) {
      RNFail("Unable to read grid file %s", grid_names[i]);
      return NULL;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Grids = %d\n", ngrids);
    printf("  L2Norms: ");  
    for (int i = 0; i < ngrids; i++) printf("%g ", grids[i]->L2Norm()); 
    printf("\n");
    fflush(stdout);
  }

  // Return grids
  return grids;
}



static int 
NormalizeGrids(R3Grid **grids, int ngrids, int normalization_type)
{
  // Check arguments
  if (ngrids == 0) return 1;
  if (normalization_type == 0) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Normalize each grid 
  if ((normalization_type % 2) == 1) {
    for (int i = 0; i < ngrids; i++) {
      grids[i]->Normalize();
    }
  }

  // Normalize all grids together
  if (normalization_type == 2) {
    // Compute sum of all grids
    RNScalar sum = 0;
    for (int i = 0; i < ngrids; i++) {
      sum += grids[i]->L2Norm();
    }

    // Rescale so that they cumulatively sum to one
    if (sum > 0) {
      for (int i = 0; i < ngrids; i++) {
        grids[i]->Divide(sum);
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Normalized grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Normalization type = %d\n", normalization_type);
    printf("  L2Norms: ");  
    for (int i = 0; i < ngrids; i++) printf("%g ", grids[i]->L2Norm()); 
    printf("\n");
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
AlignGrids(R3Grid **grids1, R3Grid **grids2, int ngrids, R4Matrix& rotation_matrix, RNScalar& max_correlation)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Check number of grids
  if (ngrids == 0) return 1;

  // Get grid resolution
  int resolution = grids1[0]->XResolution();

  // Allocate spherical grids
  R3SphericalGrid sgrid1(resolution, resolution, resolution/2);
  R3SphericalGrid sgrid2(resolution, resolution, resolution/2);

  // Allocate spherical harmonicss
  R3SphericalHarmonics harmonics1(resolution/2, resolution/2);
  R3SphericalHarmonics harmonics2(resolution/2, resolution/2);

  // Allocate WignerD structure
  R3WignerD wignerD(resolution/2);

  // Fill grids, sgrids, and harmonics for each element (C, N, O, P, S) and add them to WignerD structure
  for (int i = 0; i < ngrids; i++) {
    // Clear spherical grids
    sgrid1.Clear();
    sgrid2.Clear();

    // Clear spherical harmonics
    harmonics1.Clear();
    harmonics2.Clear();

    // Set spherical grid transformations
    sgrid1.SetWorldToGridTransformation(grids1[i]->WorldBox());
    sgrid2.SetWorldToGridTransformation(grids2[i]->WorldBox());

    // Compute spherical grids
    sgrid1.Add(*(grids1[i]));
    sgrid2.Add(*(grids2[i]));

    // Compute spherical harmonics
    sgrid1.ForwardSphericalHarmonicTransform(harmonics1);
    sgrid2.ForwardSphericalHarmonicTransform(harmonics2);

    // Add spherical harmonics to WignerD structure
    for (int j = 0; j < resolution/2; j++) {
      wignerD.AddSphericalHarmonics(harmonics1.Shell(j), harmonics2.Shell(j));
    }
  }

  // Create an euler grid
  double *wigner_grid = new double [resolution * resolution * resolution]; 
  assert(wigner_grid); 

  // Compute the inverse WignerD transform
  wignerD.InverseWignerDTransform(wigner_grid);

  // Search for the rotation with maximum correlation
  int max_index = 0;
  max_correlation = -FLT_MAX;
  for (int i = 0; i < resolution * resolution * resolution; i++) {
    if (wigner_grid[i] > max_correlation) {
      max_correlation = wigner_grid[i];
      max_index = i;
    }
  }

  // Write the euler grid
  if (wigner_grid_name) {
    FILE *fp = fopen(wigner_grid_name, "wb");
    if (!fp) {
      fprintf(stderr, "Unable to open %s\n", wigner_grid_name);
    }
    else {
      // Write a .grd file
      fwrite(&resolution, sizeof(int), 1, fp);
      fwrite(&resolution, sizeof(int), 1, fp);
      fwrite(&resolution, sizeof(int), 1, fp);
      fwrite(&(R4identity_matrix[0][0]), sizeof(RNScalar), 16, fp);
      fwrite(wigner_grid, sizeof(double), resolution * resolution * resolution, fp);
      fclose(fp);
    }
  }

  // Delete euler grid
  delete [] wigner_grid;

  // Compute rotation matrix
  R3EulerRotationMatrix(max_index, resolution, rotation_matrix);

  // Compute the error
  RNScalar squared_error = 0;
  for (int g = 0; g < ngrids; g++) {
    for (int k = 0; k < sgrid1.RadiusResolution(); k++) {
      for (int j = 0; j < sgrid1.ThetaResolution(); j++) {
        for (int i = 0; i < sgrid1.PhiResolution(); i++) {
          R3Point position1 = sgrid1.WorldPosition(i, j, k);
          R3Point position2 = rotation_matrix * position1;
          RNScalar value1 = grids1[g]->WorldValue(position1);
          RNScalar value2 = grids2[g]->WorldValue(position2);
          RNScalar delta = value1 - value2;
          squared_error += delta * delta;
        }
      }
    }
  }
  RNScalar error = sqrt(squared_error);

  // Print statistics
  if (print_verbose) {
    printf("Aligned grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Grids = %d\n", ngrids);
    printf("  Resolution = %d\n", resolution);
    printf("  Maximum Correlation = %g\n", max_correlation);
    printf("  Error = %g\n", error);
    const R4Matrix& m = rotation_matrix;
    printf("  Matrix[0][0-3] = %g %g %g %g\n", m[0][0], m[0][1], m[0][2], m[0][3]);
    printf("  Matrix[1][0-3] = %g %g %g %g\n", m[1][0], m[1][1], m[1][2], m[1][3]);
    printf("  Matrix[2][0-3] = %g %g %g %g\n", m[2][0], m[2][1], m[2][2], m[2][3]);
    printf("  Matrix[3][0-3] = %g %g %g %g\n", m[3][0], m[3][1], m[3][2], m[3][3]);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: grdalign input1 input2 [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-debug")) print_debug = 1; 
      else if (!strcmp(*argv, "-normalize")) { argc--; argv++; normalization_type = atoi(*argv); }
      else if (!strcmp(*argv, "-wigner_grid")) { argc--; argv++; wigner_grid_name = *argv; }
      else { fprintf(stderr, "Invalid program argument: %s\n", *argv); exit(1); }
    }
    else {
      if (ngrids < max_grids) { 
        grid_names[ngrids++] = *argv; 
      }
      else {
        fprintf(stderr, "You specified too many grids (max is %d)\n", max_grids);
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check input 
  if (ngrids < 2) {
    fprintf(stderr, "You did not specify at least two grid files.\n");
    return 0;
  }

  // Check input 
  if ((ngrids % 2) == 1) {
    fprintf(stderr, "You did not specify an even number of grid files.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read grids
  R3Grid **grids = ReadGrids(grid_names, ngrids);
  if (!grids) exit(-1);

  // Normalize grids
  if (!NormalizeGrids(grids, ngrids/2, normalization_type)) exit(-1);
  if (!NormalizeGrids(&(grids[ngrids/2]), ngrids/2, normalization_type)) exit(-1);

  // Align grids
  RNScalar max_correlation = 0;
  R4Matrix rotation_matrix = R4identity_matrix;
  int status = AlignGrids(grids, &(grids[ngrids/2]), ngrids/2, rotation_matrix, max_correlation);
  if (!status) exit(-1);

  // Return success
  return 0;
}


