// Program to compare histograms



// Include files

#include <RNBasics/RNBasics.h>



// Program arguments

static const int max_histograms = 4096;
static char *referencelistname  = NULL;
static char *querylistname  = NULL;
static char *matrixname = NULL;
static RNBoolean dot_product = FALSE;
static int print_debug = 0;
static int print_verbose = 0;



int
ReadHistograms(const char *listname, int max_hist, float *histograms[max_histograms], int& nhistograms, int& nbins)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file with list of histograms
  FILE *list_file = fopen(listname, "r");
  if (!list_file) {
    fprintf(stderr, "Unable to open list %s\n", listname);
    return 0;
  }

  // Read list of histograms
  nhistograms = 0;
  char signame[128];
  while (fscanf(list_file, "%s", signame) == 1) {
    // Open histogram file
    FILE *histogram_file = fopen(signame, "rb");
    if (!histogram_file) {
      fprintf(stderr, "Unable to open file %s\n", signame);
      return 0;
    }
    
    // Read number of bins
    int tmp;
    if (fread(&tmp, sizeof(int), 1, histogram_file) != 1) {
      fprintf(stderr, "Unable to read number of bins in histogram file %s\n", signame);
      return 0;
    }

    // Check number of bins
    if (nhistograms == 0) nbins = tmp;
    else if (tmp != nbins) {
      fprintf(stderr, "Mismatching number of bins in histogram file %s (%d versus %d)\n", signame, tmp, nbins);
      return 0;
    }

    // Check number of histograms
    if (nhistograms >= max_hist) { 
      fprintf(stderr, "Too many histograms (%d), need to increase size of array (%d).\n", nhistograms, max_hist);
      return 0;
    }

    // Allocate bins
    histograms[nhistograms] = new float [ nbins ];
    assert(histograms[nhistograms]);

    // Read bins
    if (fread(histograms[nhistograms], sizeof(float), nbins, histogram_file) != (unsigned int) nbins) {
      fprintf(stderr, "Unable to read %d bins in histogram file %s\n", nbins, signame);
      return 0;
    }

    // Close file
    fclose(histogram_file);
    
    // Increment number of histograms
    nhistograms++;
  }
  
  // Close file
  fclose(list_file);
  
  // Print statistics
  if (print_verbose) {
    printf("Read histograms ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Histograms = %d\n", nhistograms);
    fflush(stdout);
  }

  // Return number of histograms
  return nhistograms;
}



float *
ComputeMatrix(float *reference_histograms[max_histograms], int reference_nhistograms, 
              float *query_histograms[max_histograms], int query_nhistograms, int nbins)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate matrix
  float *matrix = new float [reference_nhistograms * query_nhistograms];
  assert(matrix);

  // Compute distance matrix (L2 distance)
  float maxdistance = 0.0;
  for (int i = 0; i < reference_nhistograms; i++) {
    for (int j = 0; j < query_nhistograms; j++) {
      float distance = 0;
      if (dot_product) {
        float sum = 0.0;
        for (int k = 0; k < nbins; k++) {
          sum += reference_histograms[i][k] * query_histograms[j][k];
        }
        distance = 1 - sum;
      }
      else {
        float sum = 0.0;
        for (int k = 0; k < nbins; k++) {
          float delta = reference_histograms[i][k] - query_histograms[j][k];
          sum += delta * delta;
        }
        distance = sqrt(sum);
      }
      matrix[j*reference_nhistograms + i] = distance;
      if (distance > maxdistance) maxdistance = distance;

      if (print_debug) {
        float dot = 0; 
        float query_l2norm = 0; 
        float ref_l2norm = 0; 
        for (int k = 0; k < nbins; k++) {
          dot += reference_histograms[i][k] * query_histograms[j][k];
          ref_l2norm += reference_histograms[i][k] * reference_histograms[i][k]; 
          query_l2norm += reference_histograms[j][k] * reference_histograms[j][k]; 
        }
        fprintf(stderr, "%d %d : %g %g : %g %g\n", i, j, sqrt(ref_l2norm), sqrt(query_l2norm), dot, distance);
      }
    }
  }

  // Normalize distance matrix (divide by maxdist)
  // for (int i = 0; i < reference_nhistograms * query_nhistograms; i++) 
  //   matrix[i] /= maxdistance;

  // Print statistics
  if (print_verbose) {
    printf("Computed matrix ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Entries = %d\n", reference_nhistograms * query_nhistograms);
    fflush(stdout);
  }

  // Return success
  return matrix;
}



int
WriteMatrix(float *matrix, int reference_nhistograms, int query_nhistograms, const char *matrixname)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open matrix file 
  FILE *fp = fopen(matrixname, "wb");
  if (!fp) {
    fprintf(stderr, "Unable to open matrix %s\n", matrixname);
    return 0;
  }

  // Write matrix
  int nvalues = reference_nhistograms * query_nhistograms;
  if (fwrite(matrix, sizeof(float), nvalues, fp) != (unsigned int) nvalues) {
    fprintf(stderr, "Unable to write %d values to matrix %s\n", nvalues, matrixname);
    return 0;
  }

  // Close matrix file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote matrix ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", reference_nhistograms * query_nhistograms * sizeof(float));
    fflush(stdout);
  }

  // Return success
  return 1;
}



int
ParseArgs (int argc, char **argv)
{
  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { print_verbose = 1; }
      else if (!strcmp(*argv, "-debug")) { print_debug = 1; }
      else if (!strcmp(*argv, "-dot")) { dot_product = TRUE; }
      else if (!strcmp(*argv, "-querylist")) { ++argv; --argc; querylistname = *argv; }
      else { fprintf(stderr, "Invalid program argument: %s", *argv); return 0; }
      argv++; argc--;
    }
    else {
      if (!referencelistname) referencelistname = *argv;
      else if (!matrixname) matrixname = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); return 0; }
      argv++; argc--;
    }
  }

  // Check filename arguments
  if (!referencelistname || !matrixname) {
    // Print usage statement
    fprintf(stderr, 
      "Usage: list2matrix <options> referencelist matrix\n\n"
      "Options:\n"
      " -querylist <listfile>\n"
      " -v       # print verbose\n"      
      "\n");
    return 0;
  }

  // Return success
  return 1;
}



int
main(int argc, char **argv)
{
  // Parse args
  if(!ParseArgs(argc, argv)) exit(-1);

  // Read reference histograms
  float *reference_histograms[max_histograms] = { NULL };
  int reference_nhistograms = 0;
  int reference_nbins = 0;
  int status1 = ReadHistograms(referencelistname, max_histograms, reference_histograms, reference_nhistograms, reference_nbins);
  if (!status1) exit(-1);

  // Read query histograms
  float **query_histograms = reference_histograms;
  int query_nhistograms = reference_nhistograms;
  int query_nbins = reference_nbins;
  if (querylistname) {
    query_histograms = new float * [ max_histograms ];
    int status = ReadHistograms(querylistname, max_histograms, query_histograms, query_nhistograms, query_nbins);
    if (!status) exit(-1);
  }

  // Check if the number of bins in the two sets of histograms are the same
  if (reference_nbins != query_nbins) {
    fprintf(stderr, "Number of bins in histograms does not match\n");
    exit(-1);
  }

  // Compute the distance matrix
  float *matrix = ComputeMatrix(reference_histograms, reference_nhistograms, query_histograms, query_nhistograms, reference_nbins);
  if (!matrix) exit(-1);

  // Write the distance matrix
  int status2 = WriteMatrix(matrix, reference_nhistograms, query_nhistograms, matrixname);
  if (!status2) exit(-1);

  // Return success
  return 0;
}
















