// Program to compare histograms



// Include files

#include <RNBasics/RNBasics.h>



// Program arguments

static const int max_histograms = 4096;
static char *referencelistname  = NULL;
static char *querylistname  = NULL;
static char *matrixname = NULL;
static RNBoolean dot_product = FALSE;
static int print_debug = 0;
static int print_verbose = 0;



int
ReadHistogramNames(const char *listname, int max_hist, char *histogram_names[max_histograms])
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file with list of histograms
  FILE *list_file = fopen(listname, "r");
  if (!list_file) {
    fprintf(stderr, "Unable to open list %s\n", listname);
    return 0;
  }

  // Read list of histograms
  int nhistograms = 0;
  char signame[128];
  while (fscanf(list_file, "%s", signame) == 1) {
    histogram_names[nhistograms++] = strdup(signame);
  }
  
  // Close file
  fclose(list_file);
  
  // Print statistics
  if (print_verbose) {
    printf("Read histogram names ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Histograms = %d\n", nhistograms);
    fflush(stdout);
  }

  // Return number of histograms
  return nhistograms;
}



int
ReadHistogram(const char *histogram_name, float *&histogram, int& nbins)
{
  // Open histogram file
  FILE *histogram_file = fopen(histogram_name, "rb");
  if (!histogram_file) {
    fprintf(stderr, "Unable to open file %s\n", histogram_name);
    return 0;
  }
  
  // Read number of bins
  if (fread(&nbins, sizeof(int), 1, histogram_file) != 1) {
    fprintf(stderr, "Unable to read number of bins in histogram file %s\n", histogram_name);
    return 0;
  }

  // Check number of bins
  if (nbins <= 0) {
    fprintf(stderr, "Invalid number of bins (%d) in %s\n", nbins, histogram_name);
    return 0;
  }

  // Allocate bins
  if (!histogram) histogram = new float [ nbins ];
  assert(histogram);

  // Read bins
  if (fread(histogram, sizeof(float), nbins, histogram_file) != (unsigned int) nbins) {
    fprintf(stderr, "Unable to read %d bins in histogram file %s\n", nbins, histogram_name);
    return 0;
  }

  // Close file
  fclose(histogram_file);

  // Return number of bins
  return nbins;
}
    


float *
ComputeMatrix(char *reference_histogram_names[max_histograms], int reference_nhistograms, 
              char *query_histogram_names[max_histograms], int query_nhistograms)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate matrix
  float *matrix = new float [reference_nhistograms * query_nhistograms];
  assert(matrix);

  // Buffers for histograms
  static float *histogram1 = NULL;
  static float *histogram2 = NULL;

  // Compute distance matrix
  float maxdistance = 0.0;
  for (int i = 0; i < reference_nhistograms; i++) {
    // Read histogram1
    int nbins1 = 0;
    if (!ReadHistogram(reference_histogram_names[i], histogram1, nbins1)) {
      fprintf(stderr, "Unable to read histogram: %s\n", reference_histogram_names[i]);
      return NULL;
    }

    // Compare to all other histograms
    for (int j = 0; j < query_nhistograms; j++) {
      // Read histogram2
      int nbins2 = 0;
      if (!ReadHistogram(query_histogram_names[j], histogram2, nbins2)) {
        fprintf(stderr, "Unable to read histogram: %s\n", reference_histogram_names[j]);
        return NULL;
      }

      // Check number of bins
      if (nbins1 != nbins2) {
        fprintf(stderr, "Mismatching number of bins (%d) in %s\n", nbins2, query_histogram_names[j]);
        return NULL;
      }

      // Compute distance
      float distance = 0;
      if (dot_product) {
        float sum = 0.0;
        for (int k = 0; k < nbins1; k++) {
          sum += histogram1[k] * histogram2[k];
        }
        distance = 1 - sum;
      }
      else {
        float sum = 0.0;
        for (int k = 0; k < nbins1; k++) {
          float delta = histogram1[k] - histogram2[k];
          sum += delta * delta;
        }
        distance = sqrt(sum);
      }
      matrix[j*reference_nhistograms + i] = distance;
      if (distance > maxdistance) maxdistance = distance;
    }
  }

  // Delete histograms
  delete [] histogram1;    
  delete [] histogram2;    
  histogram1 = NULL;
  histogram2 = NULL;

  // Print statistics
  if (print_verbose) {
    printf("Computed matrix ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Entries = %d\n", reference_nhistograms * query_nhistograms);
    fflush(stdout);
  }

  // Return success
  return matrix;
}



int
WriteMatrix(float *matrix, int reference_nhistograms, int query_nhistograms, const char *matrixname)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open matrix file 
  FILE *fp = fopen(matrixname, "wb");
  if (!fp) {
    fprintf(stderr, "Unable to open matrix %s\n", matrixname);
    return 0;
  }

  // Write matrix
  int nvalues = reference_nhistograms * query_nhistograms;
  if (fwrite(matrix, sizeof(float), nvalues, fp) != (unsigned int) nvalues) {
    fprintf(stderr, "Unable to write %d values to matrix %s\n", nvalues, matrixname);
    return 0;
  }

  // Close matrix file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote matrix ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", reference_nhistograms * query_nhistograms * sizeof(float));
    fflush(stdout);
  }

  // Return success
  return 1;
}



int
ParseArgs (int argc, char **argv)
{
  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { print_verbose = 1; }
      else if (!strcmp(*argv, "-debug")) { print_debug = 1; }
      else if (!strcmp(*argv, "-dot")) { dot_product = TRUE; }
      else if (!strcmp(*argv, "-querylist")) { ++argv; --argc; querylistname = *argv; }
      else { fprintf(stderr, "Invalid program argument: %s", *argv); return 0; }
      argv++; argc--;
    }
    else {
      if (!referencelistname) referencelistname = *argv;
      else if (!matrixname) matrixname = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); return 0; }
      argv++; argc--;
    }
  }

  // Check filename arguments
  if (!referencelistname || !matrixname) {
    // Print usage statement
    fprintf(stderr, 
      "Usage: list2matrix <options> referencelist matrix\n\n"
      "Options:\n"
      " -querylist <listfile>\n"
      " -v       # print verbose\n"      
      "\n");
    return 0;
  }

  // Return success
  return 1;
}



int
main(int argc, char **argv)
{
  // Parse args
  if(!ParseArgs(argc, argv)) exit(-1);

  // Read reference histograms
  char *reference_histogram_names[max_histograms] = { NULL };
  int reference_nhistograms = ReadHistogramNames(referencelistname, max_histograms, reference_histogram_names);
  if (!reference_nhistograms) exit(-1);

  // Read query histograms
  char **query_histogram_names = reference_histogram_names;
  int query_nhistograms = reference_nhistograms;
  if (querylistname) {
    query_histogram_names = new char * [ max_histograms ];
    query_nhistograms = ReadHistogramNames(querylistname, max_histograms, query_histogram_names);
    if (!query_nhistograms) exit(-1);
  }

  // Compute the distance matrix
  float *matrix = ComputeMatrix(reference_histogram_names, reference_nhistograms, query_histogram_names, query_nhistograms);
  if (!matrix) exit(-1);

  // Write the distance matrix
  int status2 = WriteMatrix(matrix, reference_nhistograms, query_nhistograms, matrixname);
  if (!status2) exit(-1);

  // Return success
  return 0;
}
















