// Source file for the PCA analysis program



// Include files 

#include "RNBasics/RNBasics.h"
#include "SVD/SVDFit.h"



// Program arguments

static const int max_histograms = 4096;
static char *input_filename = NULL;
static char *output_filename = NULL;
static char *eigenvector_filename = NULL;
static int input_nbins = 0;
static int output_nbins = 0;
static RNBoolean input_binary = 0;
static RNBoolean output_binary = 0;
static RNBoolean normalization = 0;
static RNBoolean print_verbose = 0;



int
GetHistogramNBins(const char *histogram_name, RNBoolean binary)
{
  int nbins = 0;

  // Check if binary
  if (binary) {
    // Open histogram file
    FILE *fp = fopen(histogram_name, "rb");
    if (!fp) {
      fprintf(stderr, "Unable to open file %s\n", histogram_name);
      return 0;
    }
  
    // Read number of bins
    if (fread(&nbins, sizeof(int), 1, fp) != 1) {
      fprintf(stderr, "Unable to read histogram file %s\n", histogram_name);
      return 0;
    }

    // Close file
    fclose(fp);
  }
  else {
    // Open histogram file
    FILE *fp = fopen(histogram_name, "r");
    if (!fp) {
      fprintf(stderr, "Unable to open file %s\n", histogram_name);
      return 0;
    }
  
    // Read number of bins
    if (fscanf(fp, "%d", &nbins) != 1) {
      fprintf(stderr, "Unable to read histogram file %s\n", histogram_name);
      return 0;
    }

    // Close file
    fclose(fp);
  }

  // Return number of bins
  return nbins;
}



int
ReadHistogramList(const char *listname, int max_histograms, char **histogram_names, int& nhistograms)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file with list of histograms
  FILE *list_file = fopen(listname, "r");
  if (!list_file) {
    fprintf(stderr, "Unable to open list %s\n", listname);
    return 0;
  }

  // Read list of histograms
  char histogram_name[128];
  while (fscanf(list_file, "%s", histogram_name) == 1) {
    histogram_names[nhistograms++] = strdup(histogram_name);
    if (nhistograms >= max_histograms) {
      fprintf(stderr, "Too many histograms in %s (max is %d)\n", listname, max_histograms);
      return 0;
    }
  }
  
  // Close file
  fclose(list_file);

  // Print statistics
  if (print_verbose) {
    printf("Read %s ...\n", listname);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Histograms = %d\n", nhistograms);
    fflush(stdout);
  }

  // Return success
  return 1;
}



int
ReadHistogram(char *histogram_name, float *bins, int nbins, RNBoolean binary)
{
  // Initialize variable telling how many bins are in histogram
  int actual_nbins = 0;

  // Check if binary or ascii
  if (binary) {
    // Open histogram file
    FILE *fp = fopen(histogram_name, "rb");
    if (!fp) {
      fprintf(stderr, "Unable to open file %s\n", histogram_name);
      return 0;
    }
  
    // Read number of bins
    if (fread(&actual_nbins, sizeof(int), 1, fp) != 1) {
      fprintf(stderr, "Unable to read number of bins in histogram file %s\n", histogram_name);
      return 0;
    }

    // Determine number of bins to read
    if (nbins < actual_nbins) actual_nbins = nbins;

    // Read bins
    if (fread(bins, sizeof(float), actual_nbins, fp) != (unsigned int) nbins) {
      fprintf(stderr, "Unable to read %d bins in histogram file %s\n", nbins, histogram_name);
      return 0;
    }
  }
  else {
    // Open histogram file
    FILE *fp = fopen(histogram_name, "r");
    if (!fp) {
      fprintf(stderr, "Unable to open file %s\n", histogram_name);
      return 0;
    }
  
    // Read number of bins
    if (fscanf(fp, "%d", &actual_nbins) != 1) {
      fprintf(stderr, "Unable to read number of bins in histogram file %s\n", histogram_name);
      return 0;
    }

    // Determine number of bins to read
    if (nbins < actual_nbins) actual_nbins = nbins;

    // Read bins
    for (int i = 0; i < actual_nbins; i++) {
      if (fscanf(fp, "%f", &bins[i]) != 1) {
        fprintf(stderr, "Unable to read bin in histogram file %s\n", histogram_name);
        return 0;
      }
    }
  }

  // Zero remaining bins
  for (int i = actual_nbins; i < nbins; i++) {
    bins[i] = 0;
  }
  
  // Return success
  return 1;
}



int
WriteHistogram(char *filename, float *bins, int nbins, RNBoolean binary)
{
  // Check if binary or ascii
  if (binary) {
    // Open file
    FILE *fp = fopen(filename, "wb");
    if (!fp) {
      fprintf(stderr, "Unable to open histogram file %s\n", filename);
      return 0;
    }
 
    // Write number of bins
    if (fwrite(&nbins, sizeof(int), 1, fp) != 1) {
      fprintf(stderr, "Unable to write number of bins to histogram file %s\n", filename);
      return 0;
    }

    // Write bins
    if (fwrite(bins, sizeof(float), nbins, fp) != (unsigned long) nbins) {
      fprintf(stderr, "Unable to write bins to histogram file %s\n", filename);
      return 0;
    }

    // Close file
    fclose(fp);
  }
  else {
    // Open file
    FILE *fp = fopen(filename, "w");
    if (!fp) {
      fprintf(stderr, "Unable to open histogram file %s\n", filename);
      return 0;
    }
 
    // Write number of bins
    fprintf(fp, "%d ", nbins);

    // Write bins
    for (int j = 0; j < nbins; j++) 
      fprintf(fp, "%f ", bins[j]);
    fprintf(fp, "\n");

    // Close file
    fclose(fp);
  }

  // Return success
  return 1;
}



int
ReadHistograms(char **histogram_names, int nhistograms, int nbins, GXMatrixMNd& data, RNBoolean binary)
{
  // Check number of histograms
  if (nhistograms == 0) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Initialize matrix
  data.SetDim(nhistograms, nbins, 0);
  
  // Allocate temporary buffer for bins
  float *bins = new float [ nbins ];
  assert(bins);

  // Read all histograms
  for (int i = 0; i < nhistograms; i++) {
    char *histogram_name = histogram_names[i];

    // Read histogram
    if (!ReadHistogram(histogram_name, bins, nbins, binary)) return 0;

    // Fill matrix with histogram data
    for (int j = 0; j < nbins; j++) 
      data(i,j) = bins[j];
  }

  // Free temporary buffer for bins
  delete [] bins;
  
  // Print statistics
  if (print_verbose) {
    printf("Read histograms ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Histograms = %d\n", nhistograms);
    printf("  # Bins = %d\n", nbins);
    fflush(stdout);
  }

  // Return success
  return 1;
}



int
WriteHistograms(char **histogram_filenames, int nhistograms, int nbins, GXMatrixMNd& data, RNBoolean binary)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate temporary buffer for bins
  float *bins = new float [ nbins ];
  assert(bins);

  // Write histograms to files 
  for (int i = 0; i < nhistograms; i++) {
    char *histogram_filename = histogram_filenames[i];

    // Fill array of bins
    for (int j = 0; j < nbins; j++) {
      bins[j] = data(i, j);
    }

    // Write histogram
    if (!WriteHistogram(histogram_filename, bins, nbins, binary)) return 0;
  }

  // Delete temporary array of bins
  delete [] bins;

  // Print statistics
  if (print_verbose) {
    printf("Wrote histograms ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Histograms = %d\n", nhistograms);
    printf("  # Bins = %d\n", nbins);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
Covariance(GXMatrixMNd& data, GXMatrixMNd& covariance)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get convenient variables
  int m = (int) data.Rows();
  int n = (int) data.Cols();

  // Compute center of mass
  GXMatrixMNd mean(1, n, 0);
  for (int i = 0; i < n; i++) {
    for (int k = 0; k < m; k++) {
      mean(0,i) += data(k,i);
    }
    mean(0,i) /= m;
  }

  // Compute covariance matrix
  covariance.SetDim(n, n, 0);
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < n; j++) {
      double sum = 0;
      for (int k = 0; k < m; k++) {
        double di = data(k,i) - mean(0,i);
        double dj = data(k,j) - mean(0,j);
        sum += di * dj; 
      }
      covariance(i,j) = sum / (m-1);
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Computed covariance matrix ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
EigenAnalysis(GXMatrixMNd& matrix, GXMatrixMNd& eigenvalues, GXMatrixMNd& eigenvectors)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get convenient variables
  int m = (int) matrix.Rows();
  int n = (int) matrix.Cols();

  // Singular value decomposition
  GXMatrixMNd U(m, n, 0);
  eigenvalues.SetDim(n, 1, 0);
  eigenvectors.SetDim(n, n, 0);
  num_svd<double> (matrix.Data(), m, n, U.Data(), eigenvalues.Data(), eigenvectors.Data());

  // Normalize eigenvectors (columns)
  if (normalization == 1) {
    for (int j = 0; j < n; j++) {
      // Compute squared length
      double squared_length = 0;
      for (int i = 0; i < n; i++) {
        double v = eigenvectors(i, j);
        squared_length += v * v;
      }

      // Divide through by length
      if (squared_length > 0) {
        double length = sqrt(squared_length);
        for (int i = 0; i < n; i++) 
          eigenvectors(i, j) = eigenvectors(i, j) / length;
      }
      else {
        fprintf(stderr, "eigenvector has null length\n");
        return 0;
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Computed eigenvalues and eigenvectors ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  1st Eigenvalue = %g\n", eigenvalues(0,0));
    printf("  2nd Eigenvalue = %g\n", eigenvalues(1,0));
    printf("  3rd Eigenvalue = %g\n", eigenvalues(2,0));
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ProjectHistograms(GXMatrixMNd& input_data, GXMatrixMNd& eigenvalues, GXMatrixMNd& eigenvectors, GXMatrixMNd& projected_data)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get convenient variables
  int m = (int) input_data.Rows();
  int n = (int) input_data.Cols();

  // Initalize projected values
  projected_data.SetDim(m, n, 0);

  // Project input files onto principal vectors
  for (int k = 0; k < m; k++) {
    for (int j = 0; j < n; j++) {
      double dot = 0;
      for (int i = 0; i < n; i++) 
        dot += input_data(k, i) * eigenvectors(i, j);
      projected_data(k, j) = dot;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Projected histograms ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



int
WriteEigenVectors(char *filename, GXMatrixMNd& eigenvalues, GXMatrixMNd& eigenvectors, int nbins)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(filename, "w");
  if (!fp) {
    fprintf(stderr, "Unable to open eigenvector file %s\n", filename);
    return 0;
  }

  // Print eigenvalues and eigenvectors
  for (int j = 0; j < nbins; j++) {
    fprintf(fp, "%12f ( ", eigenvalues(j, 0));
    for (int i = 0; i < (int) eigenvectors.Rows(); i++) fprintf(fp, "%12f ", eigenvectors(i, j));
    fprintf(fp, ")\n");
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote eigenvectors ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: hst2pca hstlist pcalist [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = TRUE; 
      else if (!strcmp(*argv, "-input_binary")) { input_binary = TRUE; }
      else if (!strcmp(*argv, "-output_binary")) { output_binary = TRUE; }
      else if (!strcmp(*argv, "-input_nbins")) { argc--; argv++; input_nbins = atoi(*argv); }
      else if (!strcmp(*argv, "-output_nbins")) { argc--; argv++; output_nbins = atoi(*argv); }
      else if (!strcmp(*argv, "-normalization")) { argc--; argv++; normalization = atoi(*argv); }
      else if (!strcmp(*argv, "-eigenvectors")) { argc--; argv++; eigenvector_filename = *argv; }
      else { fprintf(stderr, "Invalid program argument: %s", *argv); return 0; }
    }
    else {
      if (!input_filename) input_filename = *argv;
      else if (!output_filename) output_filename = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); return 0; }
    }
    argv++; argc--;
  }

  // Check input list filename
  if (!input_filename) {
    fprintf(stderr, "You did not specify a list of input histogram files.\n");
    return 0;
  }

  // Check output list filename
  if (!output_filename) {
    fprintf(stderr, "You did not specify a list of output histogram files.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read list of input histograms
  int input_nhistograms = 0;
  const int max_histograms = 4096;
  static char *input_histogram_names[max_histograms] = { NULL };
  int status = ReadHistogramList(input_filename, max_histograms, input_histogram_names, input_nhistograms);
  if (!status) exit(-1);

  // Read list of output histograms
  int output_nhistograms = 0;
  static char *output_histogram_names[max_histograms] = { NULL };
  status = ReadHistogramList(output_filename, max_histograms, output_histogram_names, output_nhistograms);
  if (!status) exit(-1);

  // Check lists
  if (input_nhistograms == 0) {
    fprintf(stderr, "Empty list of input histograms: %s \n", input_filename);
    exit(-1);
  }
  if (input_nhistograms != output_nhistograms) {
    fprintf(stderr, "Mismatching numbers of file names in input and output lists: %d %d \n", input_nhistograms, output_nhistograms);
    exit(-1);
  }

  // Determine number of input bins
  if (input_nbins <= 0) {
    input_nbins = GetHistogramNBins(input_histogram_names[0], input_binary);
    if (input_nbins == 0) exit(-1);
  }

  // Determine number of output bins
  if ((output_nbins <= 0) || (output_nbins > input_nbins)) {
    output_nbins = input_nbins;
  }

  // Read input histograms
  int nbins = 0;
  GXMatrixMNd input_data;
  status = ReadHistograms(input_histogram_names, input_nhistograms, input_nbins, input_data, input_binary);
  if (!status) exit(-1);

  // Compute covariance matrix
  GXMatrixMNd covariance;
  status = Covariance(input_data, covariance);
  if (!status) exit(-1);

  // Compute eigenvalues and eigenvectors
  GXMatrixMNd eigenvalues, eigenvectors;
  status = EigenAnalysis(covariance, eigenvalues, eigenvectors);
  if (!status) exit(-1);

  // Project input data onto new basis vectors
  GXMatrixMNd output_data;
  status = ProjectHistograms(input_data, eigenvalues, eigenvectors, output_data);
  if (!status) exit(-1);

  // Write projected histograms
  status = WriteHistograms(output_histogram_names, output_nhistograms, output_nbins, output_data, output_binary);
  if (!status) exit(-1);

  // Write eigenvalues and eigenvectors
  if (eigenvector_filename) {
    status = WriteEigenVectors(eigenvector_filename, eigenvalues, eigenvectors, output_nbins);
    if (!status) exit(-1);
  }

  // Return success
  return 0;
}


