// Source file for pocketfinder [An05]



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *pdb_name = NULL;
static char *grid_name = NULL;
static int grid_resolution[3] = { 0, 0, 0 };
static int max_grid_resolution = 256;
static RNLength grid_spacing = 0;
static R3Point *world_center = NULL;
static RNLength world_radius = 0;
static int print_verbose = 0;



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static R3Grid *
CreateGrid(PDBFile *file)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get array of atoms
  if (file->NModels() == 0) return NULL;
  PDBModel *model = file->Model(0);

  // Compute world box
  R3Box world_box;
  if (world_center || (world_radius != 0)) {
    // Compute world centroid
    R3Point world_centroid;
    if (world_center) world_centroid = *world_center;
    else world_centroid = PDBCentroid(model->atoms);

    // Compute world radius
    if (world_radius == 0) {
       world_radius = 3.0 * PDBAverageDistance(model->atoms, world_centroid);
    }

    // Compute world box
    R3Point world_corner1(world_centroid - R3ones_vector * world_radius);
    R3Point world_corner2(world_centroid + R3ones_vector * world_radius);
    world_box = R3Box(world_corner1, world_corner2);
  }
  else {
    // Compute world box
    world_box = model->BBox();
  }

  // Compute grid resolution
  if (grid_spacing == 0) grid_spacing = 0.5;
  int xres = grid_resolution[0];
  if (xres == 0) {
    xres = (int) (world_box.XLength() / grid_spacing + 1);
    if (xres > max_grid_resolution) xres = max_grid_resolution;
  }
  int yres = grid_resolution[1];
  if (yres == 0) {
    yres = (int) (world_box.YLength() / grid_spacing + 1);
    if (yres > max_grid_resolution) yres = max_grid_resolution;
  }
  int zres = grid_resolution[2];
  if (zres == 0) {
    zres = (int) (world_box.ZLength() / grid_spacing + 1);
    if (zres > max_grid_resolution) zres = max_grid_resolution;
  }

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set transform 
  grid->SetWorldToGridTransformation(world_box);

  // Rasterize all atoms into interior grid
  R3Grid interior_grid(xres, yres, zres);
  interior_grid.SetWorldToGridTransformation(world_box);
  for (int a = 0; a < model->NAtoms(); a++) {
    // Get atom
    PDBAtom *atom = model->Atom(a);
    if (atom->IsHetAtom()) continue;
    interior_grid.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1);
  }

  // Scan in three axial dimensions
  for (RNDimension dim = RN_X; dim <= RN_Z; dim++) {
    int index[3];
    RNDimension dim1 = (dim+1)%3;
    RNDimension dim2 = (dim+2)%3;
    RNBoolean *found1_buffer = new RNBoolean [grid->Resolution(dim)];
    for (index[dim2] = 0; index[dim2] < grid->Resolution(dim2); index[dim2]++) {
      for (index[dim1] = 0; index[dim1] < grid->Resolution(dim1); index[dim1]++) {
        RNBoolean found1 = FALSE;
        for (index[dim] = 0; index[dim] < grid->Resolution(dim); index[dim]++) {
          if (RNIsPositive(interior_grid.GridValue(index[0], index[1], index[2]))) found1 = TRUE;
          else found1_buffer[index[dim]] = found1;
        }
        RNBoolean found2 = FALSE;
        for (index[dim] = grid->Resolution(dim)-1; index[dim] >= 0; index[dim]--) {
          if (RNIsPositive(interior_grid.GridValue(index[0], index[1], index[2]))) found2 = TRUE;
          else if (found2 && found1_buffer[index[dim]]) grid->AddGridValue(index[0], index[1], index[2], 1);
        }
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



static int 
WriteGrid(R3Grid *grid, const char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write grid
  int status = grid->WriteFile(grid_name);

  // Print statistics
  if (print_verbose) {
    printf("Wrote grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", status * sizeof(RNScalar));
    fflush(stdout);
  }

  // Return status
  return status;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: pdb2grd pdbfile gridfile [-resolution x y z] [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-radius")) { 
        argc--; argv++; world_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-center")) { 
        world_center = new R3Point();
        argc--; argv++; (*world_center)[0] = atof(*argv); 
        argc--; argv++; (*world_center)[1] = atof(*argv); 
        argc--; argv++; (*world_center)[2] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; grid_resolution[0] = atoi(*argv); 
        argc--; argv++; grid_resolution[1] = atoi(*argv); 
        argc--; argv++; grid_resolution[2] = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-max_resolution")) { 
        argc--; argv++; max_grid_resolution = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-spacing")) { 
        argc--; argv++; grid_spacing = atof(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else if (!grid_name) grid_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check grid resolution
  if ((grid_resolution[0] < 0) || (grid_resolution[1] < 0) || (grid_resolution[2] < 0)) {
    fprintf(stderr, "Invalid grid resolution: %d %d %d\n", grid_resolution[0], grid_resolution[1], grid_resolution[2]);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read PDB file
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Create grid 
  R3Grid *grid = CreateGrid(file);
  if (!grid) exit(-1);

  // Write grid
  int status = WriteGrid(grid, grid_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}
