// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"
#include "KDTree.h"
#include "map.h"



// Program variables

static char *pdb_name = NULL;
static char *map_name = NULL;
static char *grid_name = NULL;
static char *asa_name = NULL;
static char *ligand_name = NULL;
static int max_resolution = 256;
static RNScalar grid_spacing = 0.5; // angstroms
static RNScalar world_sigma = 0; // angstroms
static RNScalar world_radius = PDB_UNKNOWN;
static R3Point world_centroid(PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN);
static RNScalar nearby_radius = 3.5;
static RNBoolean normalization_type = 0; 
static RNBoolean mask_to_ligand = FALSE;
static RNBoolean mask_to_target_volume = FALSE;
static RNBoolean target_volume_from_ligand = FALSE;
static RNBoolean target_volume_from_ligand_elements = FALSE;
static RNBoolean largest_connected_component_only = FALSE;
static RNScalar target_volume_default = 100;
static int print_verbose = 0;
static int print_debug = 0;
static int print_plots = 0;



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static int
ReadASAFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read ASA file
  int natoms = file->ReadASAFile(filename);
  if (natoms == 0) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read accessible surface area file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Surface Atoms = %d\n", natoms);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static PDBResidue *
FindLigand(PDBFile *file, const char *ligand_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Find ligand
  PDBResidue *ligand = file->FindResidue(ligand_name);
  if (!ligand) {
    fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, pdb_name); 
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Found ligand ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", ligand->NAtoms());
    fflush(stdout);
  }

  // Return ligand
  return ligand;
}



static Map *
ReadMap(const char *map_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate map
  Map *map = new Map();
  if (!map) {
    fprintf(stderr, "Unable to allocate map.\n");
    return NULL;
  }

  // Read map
  int status = map->ReadFile(map_name);
  if (!status) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read map ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Frame Type = %d\n", map->FrameType());
    printf("  Storage Type = %d\n", map->StorageType());
    printf("  # Frames = %d\n", map->NFrames());
    printf("  # Elements = %d\n", map->NElements());
    printf("  World Radius = %g\n", map->WorldRadius());
    printf("  World Spacing = %g\n", map->WorldSpacing());
    printf("  Resolution = %d\n", map->Resolution());
    fflush(stdout);
  }

  // Return map
  return map;
}



static int
SplatAtoms(PDBFile *file, Map *map, R3Grid **grids, R3Grid *weight, const R3Box& world_box)
{
  // Start statistics
  int atom1_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);
  if (model->NAtoms() == 0) return 0;

  // Create KD tree of all protein atoms
  PDBAtom *tmp = model->Atom(0);
  int position_offset = (unsigned char *) &(tmp->position) - (unsigned char *) tmp;
  KDTree<PDBAtom *> *kdtree = new KDTree<PDBAtom *>(model->atoms, position_offset);
  if (!kdtree) { fprintf(stderr, "Unable to create KD tree for %s\n", file->Name()); return 0; }

  // Consider every atom as potential origin of coordinate system
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom1 = model->Atom(i);
    if (atom1->IsHetAtom()) continue;
    if (!atom1->AminoAcid()) continue;
    if ((asa_name) && (atom1->accessible_surface_area == 0)) continue;
    if (R3Distance(atom1->Position(), world_box) > map->WorldRadius()) continue;
    int atom1_index = map->AtomType(atom1);
    assert(atom1_index < map->NAtomTypes());
    if (atom1_index < 0) continue;
    atom1_count++;

    // Find all nearby atoms 
    const R3Point& p1 = atom1->Position();
    RNArray<PDBAtom *> nearby_atoms;
    kdtree->FindAll(atom1, RN_EPSILON, nearby_radius, nearby_atoms);
    if (nearby_atoms.IsEmpty()) continue;

    // Compute first vector of atom coordinate system 
    R3Vector vector1 = R3zero_vector;
    for (int a = 0; a < nearby_atoms.NEntries(); a++) {
      PDBAtom *atom2 = nearby_atoms[a];
      if (atom2->IsHetAtom()) continue;
      if (!atom2->AminoAcid()) continue;
      R3Point p2 = atom2->Position();
      R3Vector v = p2 - p1;
      RNLength d = v.Length();
      if (RNIsZero(d)) continue;
      vector1 += v / d;
    }

    // Compute second vector of atom coordinate system 
    if (vector1.IsZero()) continue;
    R3Vector vector2 = R3zero_vector;
    R3Plane plane1(atom1->Position(), vector1);
    for (int a = 0; a < nearby_atoms.NEntries(); a++) {
      PDBAtom *atom2 = nearby_atoms[a];
      if (atom2->IsHetAtom()) continue;
      if (!atom2->AminoAcid()) continue;
      R3Point p2 = atom2->Position();
      p2.Project(plane1);
      R3Vector v = p2 - p1;
      RNLength d = v.Length();
      if (RNIsZero(d)) continue;
      vector2 += v / d;
    }

    // Construct coordinate system 
    if (vector2.IsZero()) continue;
    R3Triad triad(vector1, vector2);
    R3CoordSystem cs(p1, triad);

    // Construct transformation from frame to world
    R3Affine frame_to_world(cs.Matrix());

    // Splat map into grids
    for (int m = 0; m < map->NElements(); m++) {
      map->SplatIntoGrid(atom1_index, m, grids[m], frame_to_world);
    }

    // Splat map into frame overlap grid
    if (weight) {
      R3Point world_position = cs.Origin();
      world_position.Transform(frame_to_world);
      weight->RasterizeWorldSphere(world_position, map->WorldRadius(), 1);
    }
  }

  // Return number of splats
  return atom1_count;
}



static int
SplatTriples(PDBFile *file, Map *map, R3Grid **grids, R3Grid *weight, const R3Box& world_box)
{
  // Start statistics
  int atom1_count = 0;
  int atom2_count = 0;
  int atom3_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);
  if (model->NAtoms() == 0) return 0;

  // Get the number of possible atom types
  int natomtypes = map->NAtomTypes();

  // Create KD tree of all protein atoms
  PDBAtom *tmp = model->Atom(0);
  int position_offset = (unsigned char *) &(tmp->position) - (unsigned char *) tmp;
  KDTree<PDBAtom *> *kdtree = new KDTree<PDBAtom *>(model->atoms, position_offset);
  if (!kdtree) { fprintf(stderr, "Unable to create KD tree for %s\n", file->Name()); return 0; }

  // Consider every atom as potential origin of coordinate system
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom1 = model->Atom(i);
    if (atom1->IsHetAtom()) continue;
    if (!atom1->AminoAcid()) continue;
    if ((asa_name) && (atom1->accessible_surface_area == 0)) continue;
    if (R3Distance(atom1->Position(), world_box) > map->WorldRadius()) continue;
    int atom1_index = map->AtomType(atom1);
    assert(atom1_index < natomtypes);
    if (atom1_index < 0) continue;
    atom1_count++;

    // Find all other atoms nearby atom1 
    RNArray<PDBAtom *> nearby_atoms;
    kdtree->FindAll(atom1, RN_EPSILON, 2 * nearby_radius, nearby_atoms);

    // Consider all nearby protein atoms
    for (int j = 0; j < nearby_atoms.NEntries(); j++) {
      PDBAtom *atom2 = nearby_atoms[j];
      if (atom2 == atom1) continue;
      if (atom2->IsHetAtom()) continue;
      if (!atom2->AminoAcid()) continue;
      if ((asa_name) && (atom2->accessible_surface_area == 0)) continue;
      if (R3Distance(atom2->Position(), atom1->Position()) > nearby_radius) continue;
      int atom2_index = map->AtomType(atom2);
      assert(atom2_index < natomtypes);
      if (atom2_index < 0) continue;
      atom2_count++;

      // Consider all pairs of nearby protein atoms
      for (int k = j+1; k < nearby_atoms.NEntries(); k++) {
        PDBAtom *atom3 = nearby_atoms[k];
        if (atom3 == atom1) continue;
        if (atom3 == atom2) continue;
        if (atom3->IsHetAtom()) continue;
        if (!atom3->AminoAcid()) continue;
        if ((asa_name) && (atom3->accessible_surface_area == 0)) continue;
        if (R3Distance(atom3->Position(), atom2->Position()) > nearby_radius) continue;
        int atom3_index = map->AtomType(atom3);
        assert(atom3_index < natomtypes);
        if (atom3_index < 0) continue;
        atom3_count++;

        // Construct world to frame transformation
        R3Point origin = atom1->Position();
        R3Vector posx_vector = origin - atom2->Position();
        RNLength posx_length = posx_vector.Length();
        if (RNIsZero(posx_length)) continue;
        else posx_vector /= posx_length;
        R3Vector posy_vector = atom3->Position() - atom2->Position();
        R3Vector posz_vector = posx_vector % posy_vector;
        RNLength posz_length = posz_vector.Length();
        if (RNIsZero(posz_length)) return 0;
        else posz_vector /= posz_length;
        posy_vector = posz_vector % posx_vector;
        RNLength posy_length = posy_vector.Length();
        if (RNIsZero(posy_length)) return 0;
        else posy_vector /= posy_length;
        R3Triad triad(posx_vector, posy_vector, posz_vector);
        R3CoordSystem cs(origin, triad);
        R3Affine frame_to_world(cs.Matrix());

        // Determine frame index 
        int frame_index = atom3_index;
        frame_index += atom2_index * natomtypes;
        frame_index += atom1_index * natomtypes * natomtypes;
        assert((frame_index >= 0) && (frame_index < map->NFrames()));

        // Splat map into grids
        for (int m = 0; m < map->NElements(); m++) {
          map->SplatIntoGrid(frame_index, m, grids[m], frame_to_world);
        }

        // Splat map into frame overlap grid
        if (weight) {
          R3Point world_position = cs.Origin();
          world_position.Transform(frame_to_world);
          weight->RasterizeWorldSphere(world_position, map->WorldRadius(), 1);
        }
      }
    }
  }

  // Return number of splats
  return atom3_count;
}



static int
SplatResidues(PDBFile *file, Map *map, R3Grid **grids, R3Grid *weight, const R3Box& world_box)
{
  // Start statistics
  int splat_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);
  if (model->NAtoms() == 0) return 0;

  // Consider each protein residue
  for (int i = 0; i < model->NResidues(); i++) {
    PDBResidue *residue1 = model->Residue(i);
    if (residue1->HasHetAtoms()) continue;

    // Determine amino acid
    PDBAminoAcid *aminoacid1 = residue1->AminoAcid();
    if (!aminoacid1) continue;
    int aminoacid_index = aminoacid1 - PDBaminoacids;
    assert((aminoacid_index >= 0) && (aminoacid_index < PDBnaminoacids));
    assert(aminoacid_index < map->NFrames());

    // Check whether the residue is on the surface
    if (asa_name) {
      RNBoolean on_surface = FALSE;
      for (int j = 0; j < residue1->NAtoms(); j++) 
        if (residue1->Atom(j)->accessible_surface_area > 0) { on_surface = TRUE; break; }
      if (!on_surface) continue;
    }

    // Construct residue coordinate system 
    R3CoordSystem cs;
    if (!residue1->ConstructCoordinateSystem(cs, map->FrameType())) {
      // fprintf(stderr, "Unable to construct coordinate system for residue %s %d\n", residue1->Name(), residue1->Sequence());
      continue;
    }

    // Check if splat can touch world_box
    if (R3Distance(cs.Origin(), world_box) > map->WorldRadius()) continue;

    // Construct transformation from world to frame
    R3Affine frame_to_world(cs.Matrix());

    // Splat map into grids
    for (int m = 0; m < map->NElements(); m++) {
      map->SplatIntoGrid(aminoacid_index, m, grids[m], frame_to_world);
    }

    // Splat map into frame overlap grid
    if (weight) {
      R3Point world_position = cs.Origin();
      world_position.Transform(frame_to_world);
      weight->RasterizeWorldSphere(world_position, map->WorldRadius(), 1);
    }

    // Increment counter
    splat_count++;
  }

  // Return number of splats
  return splat_count;
}



static R3Grid **
CreateGrids(PDBFile *file, PDBResidue *ligand, Map *map)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get PDB model from file
  if (file->NModels() < 1) return NULL;
  PDBModel *model = file->Model(0);
  if (model->NAtoms() == 0) return 0;

  // Construct world box for grids (must be square for shape matching)
  R3Point c = world_centroid;
  RNScalar r = world_radius;
  if (r == PDB_UNKNOWN) r = (ligand) ? 15 : 0.5 * model->BBox().LongestAxisLength();
  if (c[0] == PDB_UNKNOWN) c = (ligand) ? ligand->Centroid() : model->BBox().Centroid();
  R3Box world_box(c[0]-r, c[1]-r, c[2]-r, c[0]+r, c[1]+r, c[2]+r);

  // Determine grid resolution (must be power of 2 for shape matching)
  int res = 4;
  while (((world_box.XLength() / res) > grid_spacing) && (res < max_resolution))
    res *= 2;

  // Allocate array of grids
  R3Grid **grids = new R3Grid * [ map->NElements() ];
  if (!grids) {
    fprintf(stderr, "Unable to allocate array of grids\n");
    return NULL;
  }

  // Initialize grids
  for (int i = 0; i < map->NElements(); i++) {
    // Allocate grid
    grids[i] = new R3Grid(res, res, res);
    if (!grids[i]) {
      fprintf(stderr, "Unable to allocate grid\n");
      return NULL;
    }

    // Set transformation
    grids[i]->SetWorldToGridTransformation(world_box);
  }

  // Create grid to store frame overlaps
  R3Grid *weight = NULL;
  if (normalization_type & 0x4) {
    weight = new R3Grid(res, res, res);
    assert(weight);
    weight->SetWorldToGridTransformation(world_box);
  }

  // Splat maps into grids
  int splat_count = 0;
  if (map->FrameType() == ATOM_MAP_FRAME) splat_count = SplatAtoms(file, map, grids, weight, world_box);
  else if (map->FrameType() == TRIPLE_MAP_FRAME) splat_count = SplatTriples(file, map, grids, weight, world_box);
  else if (map->FrameType() == RESIDUE_MAP_FRAME) splat_count = SplatResidues(file, map, grids, weight, world_box);
  else RNAbort("Invalid frame type: %d", map->FrameType());
  if (!splat_count) return 0;

  // Normalize by frame overlaps
  if (normalization_type & 0x4) {
    assert(weight);
    weight->Blur(world_sigma * weight->WorldToGridScaleFactor());
    for (int i = 0; i < map->NElements(); i++) 
      grids[i]->Divide(*weight);
    delete weight;
  }

  // Print statistics
  if (print_verbose) {
    RNScalar L1 = 0;
    RNScalar L2_squared = 0;
    for (int i = 0; i < map->NElements(); i++) { 
      L1 += grids[i]->L1Norm(); 
      L2_squared += grids[i]->L2NormSquared(); 
    }
    printf("Created grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Grids = %d\n", map->NElements());
    printf("  Grid Resolution = %d %d %d\n", res, res, res);
    printf("  World Spacing = %g\n", grids[0]->GridToWorldScaleFactor());
    printf("  World Centroid = %g %g %g\n", c[0], c[1], c[2]);
    printf("  World Radius = %g\n", r);
    printf("  # Splats = %d\n", splat_count);
    printf("  L1Norm = %g\n", L1);
    printf("  L2Norm = %g\n", sqrt(L2_squared));
    fflush(stdout);
  }

  // Return grids
  return grids;
}



static int 
BlurGrids(R3Grid **grids, int ngrids, RNScalar world_sigma)
{
  // Check sigma
  if (world_sigma == 0) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Blur the grid
  for (int i = 0; i < ngrids; i++) {
    grids[i]->Blur(world_sigma * grids[i]->WorldToGridScaleFactor());
  }

  // Print statistics
  if (print_verbose) {
    RNScalar L1 = 0;
    RNScalar L2_squared = 0;
    for (int i = 0; i < ngrids; i++) { 
      L1 += grids[i]->L1Norm(); 
      L2_squared += grids[i]->L2NormSquared(); 
    }
    printf("Blurred grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  World Sigma = %g\n", world_sigma);
    printf("  L1Norm = %g\n", L1);
    printf("  L2Norm = %g\n", sqrt(L2_squared));
    fflush(stdout);
  }

  // Return success
  return 1;
}



static RNScalar 
GridVolume(R3Grid *grid, RNScalar threshold, RNBoolean largest_connected_component_only)
{
  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Count number of grid samples at or above threshold
  int grid_count = 0;
  if (largest_connected_component_only) {
    const int max_components = 1024;
    int component_sizes[max_components];
    int ncomponents = grid->ConnectedComponents(threshold, max_components, NULL, component_sizes);
    for (int i = 0; i < ncomponents; i++) {
      if (component_sizes[i] > grid_count) grid_count = component_sizes[i];
    }
  }
  else {
    for (int i = 0; i < grid->NEntries(); i++) {
      if (grid->GridValue(i) >= threshold) grid_count++;
    }
  }

  // Return volume of grid samples at or above threshold
  return grid_count * volume_scale_factor;
}



static RNScalar 
GridThreshold(R3Grid *grid, RNScalar target_volume, RNBoolean largest_connected_component_only)
{
  // Determine ranges of thresholds and volumes
  RNInterval threshold_range = grid->Range();
  if (threshold_range.Min() == 0) threshold_range.SetMin(1.0E-20);
  RNScalar min_volume = GridVolume(grid, threshold_range.Max(), largest_connected_component_only);
  RNScalar max_volume = GridVolume(grid, threshold_range.Min(), largest_connected_component_only);
  RNInterval volume_range(min_volume, max_volume); 

  // Check if target volume is within range
  if (target_volume >= volume_range.Max()) return threshold_range.Min();
  if (target_volume <=volume_range.Min()) return threshold_range.Max();

  // Binary search for threshold that provides target volume
  for (int i = 0; i < 1000; i++) {
    RNScalar t = (target_volume - volume_range.Min()) / (volume_range.Max() - volume_range.Min());
    RNScalar threshold = threshold_range.Max() - t * (threshold_range.Max() - threshold_range.Min());
    RNScalar volume = GridVolume(grid, threshold, largest_connected_component_only);
    if (RNIsGreater(volume, target_volume, 1)) {
      threshold_range.SetMin(threshold);
      volume_range.SetMax(volume);
    }
    else if (RNIsLess(volume, target_volume, 1)) {
      threshold_range.SetMax(threshold);
      volume_range.SetMin(volume);
    }
    else return threshold;
  }

  // Return middle of remaining range
  return threshold_range.Mid();
}



static void
MaskGridsToLigand(PDBFile *file, PDBResidue *ligand, R3Grid **grids, int ngrids)
{
  // Create mask
  R3Grid mask(grids[0]->XResolution(), grids[0]->YResolution(), grids[0]->ZResolution());
  mask.SetWorldToGridTransformation(grids[0]->WorldToGridTransformation());
  for (int i = 0; i < ligand->NAtoms(); i++) {
    PDBAtom *atom = ligand->Atom(i);
    mask.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1);
  }

  // Mask the grids
  for (int i = 0; i < ngrids; i++) {
    grids[i]->Mask(mask);
  }
}



static void
MaskGridsToCavities(PDBFile *file, R3Grid **grids, int ngrids, RNScalar max_offset_distance, RNScalar min_cavity_volume)
{
  // Get PDB model
  if (file->NModels() == 0) return;
  PDBModel *model = file->Model(0);

  // Create exterior mask
  RNScalar water_radius = 1.4;
  R3Grid mask(grids[0]->XResolution(), grids[0]->YResolution(), grids[0]->ZResolution());
  mask.SetWorldToGridTransformation(grids[0]->WorldToGridTransformation());
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (atom->IsHetAtom()) continue;
    mask.RasterizeWorldSphere(atom->Position(), atom->Radius() + water_radius, 1);
  }

  // Invert and dilate mask to get exterior
  mask.Threshold(0.5, 1, 0);
  mask.Dilate(water_radius * mask.WorldToGridScaleFactor());

  // Create offset mask
  R3Grid offset_mask(grids[0]->XResolution(), grids[0]->YResolution(), grids[0]->ZResolution());
  offset_mask.SetWorldToGridTransformation(grids[0]->WorldToGridTransformation());
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (atom->IsHetAtom()) continue;
    offset_mask.RasterizeWorldSphere(atom->Position(), atom->Radius() + max_offset_distance, 1);
  }

  // Combine the masks
  mask.Mask(offset_mask);

  // Keep only cavities large enough to hold a certain number of atoms
  RNScalar world_grid_cell_volume = mask.GridToWorldScaleFactor();
  world_grid_cell_volume = world_grid_cell_volume * world_grid_cell_volume * world_grid_cell_volume;
  int min_cavity_size = (int) (min_cavity_volume / world_grid_cell_volume);
  int max_cavities = mask.NEntries();
  int *cavity_size = new int [ max_cavities ];
  int *cavity_grid = new int [ mask.NEntries() ];
  int ncavities = mask.ConnectedComponents(0.5, max_cavities, NULL, cavity_size, cavity_grid);
  RNScalar *mask_valuesp = (RNScalar *) mask.GridValues();
  for (int i = 0; i < mask.NEntries(); i++) {
    int cavity = cavity_grid[i];
    assert(cavity < ncavities);
    if (cavity < 0) continue;
    if (cavity_size[cavity] < min_cavity_size) *mask_valuesp = 0;
    mask_valuesp++;
  }
  delete [] cavity_size;
  delete [] cavity_grid;

  // Mask the grids
  for (int i = 0; i < ngrids; i++) {
    grids[i]->Mask(mask);
  }
}



static int 
MaskGrids(PDBFile *file, PDBResidue *ligand, R3Grid **grids, int ngrids)
{
  // Check number of grids
  if (ngrids == 0) return 0;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Mask the grids
  if (mask_to_ligand && ligand) {
    // Mask the grids to cover only the ligand volume
    MaskGridsToLigand(file, ligand, grids, ngrids);
  }
  else {
    // Mask the grids to cover only cavities exterior to the protein
    RNScalar max_offset_distance = 6;
    RNScalar min_ligand_atoms = 4;
    RNScalar min_cavity_volume = min_ligand_atoms * 4 * RN_PI * 1.8 * 1.8 * 1.8 / 3;
    MaskGridsToCavities(file, grids, ngrids, max_offset_distance, min_cavity_volume);
  }

  // Mask the grid to adhere to a target volume
  if (mask_to_target_volume) {
    for (int i = 0; i < ngrids; i++) {
      // Get grid
      R3Grid *grid = grids[i];

      // Determine scale factor to take number of grid cells to a volume
      RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
      RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

      // Determine target volume
      RNScalar target_volume = target_volume_default;
      if (ligand && (target_volume_from_ligand || target_volume_from_ligand_elements)) {
        // Create grid with ligand atoms of element type rasterized
        R3Grid ligand_grid(grid->XResolution(), grid->YResolution(), grid->ZResolution());
        ligand_grid.SetWorldToGridTransformation(grid->WorldToGridTransformation());
        for (int j = 0; j < ligand->NAtoms(); j++) {
          PDBAtom *atom = ligand->Atom(j);
          PDBElement *element = atom->Element();
          if (!element) continue;
          int element_index = element - PDBelements;
          assert((element_index >= 0) && (element_index < PDBnelements));
          if ((element_index < 0) || (element_index >= ngrids)) element_index = 0;
          if ((target_volume_from_ligand_elements) && (i != element_index)) continue;
          ligand_grid.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
        }

        // Determine target volume from ligand
        target_volume = volume_scale_factor * ligand_grid.Cardinality();
      }

      // Determine threshold to achieve target volume
      RNScalar threshold = GridThreshold(grid, target_volume, largest_connected_component_only);

      // Mask the grid to threshold
      if (largest_connected_component_only) {
        // Allocate arrays to store connected component information
        int *component_sizes = new int [ grid->NEntries() ];
        int *component_ids = new int [ grid->NEntries() ];
        assert(component_sizes && component_ids);

        // Find connected components 
        int ncomponents = grid->ConnectedComponents(threshold, grid->NEntries(), NULL, component_sizes, component_ids);
        assert(ncomponents <= grid->NEntries());

        // Find maximum connected component
        int biggest_component = -1;
        int biggest_component_size = -1;
        for (int j = 0; j < ncomponents; j++) {
          if (component_sizes[j] > biggest_component_size) {
            biggest_component_size = component_sizes[j];
            biggest_component = j;
          }
        }

        // Assign '0' to values not in biggest component at threshold
        if (biggest_component >= 0) {
          int *component_idsp = component_ids;
          RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
          for (int j = 0; j < grid->NEntries(); j++) {
            if (*component_idsp != biggest_component) *grid_valuesp = 0;
            component_idsp++;
            grid_valuesp++;
          }
        }

        // Delete arrays to store connected component information
        delete [] component_sizes;
        delete [] component_ids;
      }
      else {
        // Threshold the grid
        grid->Threshold(threshold, 0, R3_GRID_KEEP_VALUE);
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    RNScalar L1 = 0;
    RNScalar L2_squared = 0;
    for (int i = 0; i < ngrids; i++) { 
      L1 += grids[i]->L1Norm(); 
      L2_squared += grids[i]->L2NormSquared(); 
    }
    printf("Masked grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Mask to ligand = %d\n", (mask_to_ligand && ligand) ? 1 : 0);
    printf("  L1Norm = %g\n", L1);
    printf("  L2Norm = %g\n", sqrt(L2_squared));
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
NormalizeGrids(R3Grid **grids, int ngrids, int normalization_type)
{
  // Check arguments
  if (ngrids == 0) return 1;
  if (normalization_type == 0) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Check normalization type
  if ((normalization_type & 0x3) == 1) {
    // Normalize each grid 
    for (int i = 0; i < ngrids; i++) {
      grids[i]->Normalize();
    }
  }
  else if ((normalization_type & 0x3) == 2) {
    // Normalize each grid cell
    for (int x = 0; x < grids[0]->XResolution(); x++) {
      for (int y = 0; y < grids[0]->YResolution(); y++) {
        for (int z = 0; z < grids[0]->ZResolution(); z++) {
          // Compute sum of all grid entries at position
          RNScalar sum = 0;
          for (int i = 0; i < ngrids; i++) {
            RNScalar value = grids[i]->GridValue(x, y, z);
            sum += value * value;
          }
  
          // Rescale so that they sum to one
          if (sum > 0) {
            RNScalar l2norm = sqrt(sum);
            for (int i = 0; i < ngrids; i++) {
               grids[i]->SetGridValue(x, y, z, grids[i]->GridValue(x, y, z) / l2norm);
            }
          }
        }
      }
    }
  }
  else if ((normalization_type & 0x3) == 3) {
    // Compute sum of all grids
    RNScalar sum = 0;
    for (int i = 0; i < ngrids; i++) {
      sum += grids[i]->L2Norm();
    }

    // Rescale so that they cumulatively sum to one
    if (sum > 0) {
      for (int i = 0; i < ngrids; i++) {
        grids[i]->Divide(sum);
      }
    }
  }
  else {
    RNAbort("Invalid normalization type");
  }

  // Print statistics
  if (print_verbose) {
    printf("Normalized grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Normalization type = %d\n", normalization_type);
    printf("  Norm by element ...\n");
    RNScalar L1_sum = 0;
    RNScalar L2_squared_sum = 0;
    for (int i = 0; i < ngrids; i++) { 
      RNScalar L1 = grids[i]->L1Norm(); L1_sum += L1;
      RNScalar L2_squared = grids[i]->L2NormSquared(); L2_squared_sum += L2_squared;
      printf("    Norms %s = %g %g\n", PDBelements[i].Name(), L1, sqrt(L2_squared));
    }
    printf("  L1Norm = %g\n", L1_sum);
    printf("  L2Norm = %g\n", sqrt(L2_squared_sum));
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
WriteGrids(R3Grid **grids, int ngrids, const char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write grids
  for (int i = 0; i < ngrids; i++) {
    char name[256];
    strcpy(name, grid_name);
    char *namep = strstr(name, ".grd");
    if (!namep) namep = &name[strlen(name)];
    if (ngrids > 1) sprintf(namep, "-%s.grd", PDBelements[i].Name());
    int status = grids[i]->WriteFile(name);
    if (!status) return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Wrote grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Grids = %d\n", ngrids);
    printf("  # Bytes = %d\n", ngrids * grids[0]->NEntries() * sizeof(RNScalar));
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
PrintPlot(R3Grid *grid, RNScalar *ligand_values, int ligand_nvalues, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = stdout;
  if (filename) {
    fp = fopen(filename, "w");
    if (!fp) {
      fprintf(stderr, "Unable to open site plot file: %s\n", filename);
      return 0;
    }
  }

  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Make sorted array of grid values
  int grid_nvalues = grid->NEntries();
  RNScalar *grid_values = new RNScalar [ grid_nvalues ];
  assert(grid_values);
  for (int i = 0; i < grid_nvalues; i++) grid_values[i] = grid->GridValue(i);
  qsort(grid_values, grid_nvalues, sizeof(RNScalar), RNCompareScalars);

  // Print plot
  fprintf(fp, "0 0\n");
  int grid_index = grid_nvalues-1;
  for (int ligand_index = ligand_nvalues-1; ligand_index >= 0; ligand_index--) {
    // Check grid index
    if (grid_index <= 0) break;

    // Check ligand value
    if (ligand_values[ligand_index] == 0) break;

    // Find next grid value less or equal to ligand value
    for (; grid_index >= 0; grid_index--) {
      if (RNIsLess(grid_values[grid_index], ligand_values[ligand_index], 1.0E-8)) break;
    }

    // Check if did not find any of grid 
    if (grid_index < 0) break;

    // Determine fraction of ligand values considered
    RNScalar ligand_fraction = (RNScalar) (ligand_nvalues - ligand_index) / (RNScalar) ligand_nvalues;

    // Determine number of grid cells with value greater of equal to ligand value
    int grid_count = grid_nvalues - grid_index;

    // Convert number of grid cells to a volume
    RNScalar grid_volume = grid_count * volume_scale_factor;

    // Print plot numbers
    fprintf(fp, "%g %g\n", ligand_fraction, grid_volume);
  }

  // Close file
  if (filename) fclose(fp);

  // Deallocate array of grid values
  delete [] grid_values;

  // Print statistics
  if (print_verbose) {
    printf("Wrote plot ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", ligand_nvalues);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static void
AnalyzeLigand(PDBResidue *ligand, R3Grid **grids, int ngrids)
{
  // Check ligand
  if (!ligand) return;
  if (ngrids == 0) return;

  // Check grids
  if (ngrids == 0) return;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Print leader message
  if (print_verbose) {
    printf("Analyzing grids/ligand ...\n");
    fflush(stdout);
  }

  // Compute count and sum for each grid
  int total_nsamples = 0;
  RNScalar total_sum = 0;
  int *per_grid_nsamples = new int [ ngrids ];
  RNScalar *per_grid_sum = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) per_grid_nsamples[g] = 0;
  for (int g = 0; g < ngrids; g++) per_grid_sum[g] = 0;
  for (int g = 0; g < ngrids; g++) {
    for (int i = 0; i < grids[g]->XResolution(); i++) {
      for (int j = 0; j < grids[g]->YResolution(); j++) {
        for (int k = 0; k < grids[g]->ZResolution(); k++) {
          RNScalar value = grids[g]->GridValue(i, j, k);
          if (value == 0) continue;
          total_sum += value;
          total_nsamples++;
          per_grid_sum[g] += value;
          per_grid_nsamples[g]++;
        }
      }
    }
  }

  // Check number of samples
  if (total_nsamples == 0) return;

  // Compute avg for each grid
  RNScalar total_avg = total_sum / total_nsamples;
  RNScalar *per_grid_avg = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) {
    if (per_grid_nsamples[g] == 0) per_grid_avg[g] = 0;
    else per_grid_avg[g] = per_grid_sum[g] / per_grid_nsamples[g];
  }

  // Compute sum of squared errors ofor each grid
  RNScalar total_sse = 0;
  RNScalar *per_grid_sse = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) per_grid_sse[g] = 0;
  for (int g = 0; g < ngrids; g++) {
    for (int i = 0; i < grids[g]->XResolution(); i++) {
      for (int j = 0; j < grids[g]->YResolution(); j++) {
        for (int k = 0; k < grids[g]->ZResolution(); k++) {
          RNScalar value = grids[g]->GridValue(i, j, k);
          RNScalar total_delta = value - total_avg;
          total_sse += total_delta * total_delta;
          RNScalar per_grid_delta = value - per_grid_avg[g];
          per_grid_sse[g] += per_grid_delta * per_grid_delta;
        }
      }
    }
  }

  // Compute standard deviation for each grid
  RNScalar total_stddev = sqrt(total_sse / total_nsamples);
  RNScalar *per_grid_stddev = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) {
    if (per_grid_nsamples[g] == 0) per_grid_stddev[g] = 0;
    else per_grid_stddev[g] = sqrt(per_grid_sse[g] / per_grid_nsamples[g]);
  }

  // Print grid scores
  if (print_verbose) {
    printf("  Grid Statistics ...\n");
    printf("    Grid Statistics %s = %12g ( %12g )  %12g %12g\n", "All",  
      total_sum, 100.0, total_avg, total_stddev);
    for (int g = 0; g < ngrids; g++) {
      printf("    Grid Statistics %s = %12g ( %12g )  %12g %12g\n", PDBelements[g].Name(), 
        per_grid_sum[g], 100 * per_grid_sum[g] / total_sum, per_grid_avg[g], per_grid_stddev[g]);
    }
    fflush(stdout);
  }

  // Compute grid scores
  int total_element_natoms = 0;
  RNScalar total_element_score = 0;
  int *per_grid_element_natoms = new int [ ngrids ];
  RNScalar *per_grid_element_score = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) per_grid_element_natoms[g] = 0; 
  for (int g = 0; g < ngrids; g++) per_grid_element_score[g] = 0; 
  for (int i = 0; i < ligand->NAtoms(); i++) {
    PDBAtom *atom = ligand->Atom(i);

    // Get element
    PDBElement *element = atom->Element();
    if (!element) continue;
    int element_index = element - PDBelements;
    assert((element_index >= 0) && (element_index < PDBnelements));
    if ((element_index < 0) || (element_index >= ngrids)) element_index = 0;

    // Log score for element at atom position
    if (per_grid_stddev[element_index] > 0) {
      const R3Point& position = atom->Position();
      RNScalar value = grids[element_index]->WorldValue(position);
      RNScalar score = (value - per_grid_avg[element_index]) / per_grid_stddev[element_index];
      total_element_score += score;
      total_element_natoms++;
      per_grid_element_score[element_index] += score;
      per_grid_element_natoms[element_index]++;
    }
  }

  // Divide scores by number of atoms to get average score per atom
  if (total_element_natoms > 0) total_element_score /= total_element_natoms;
  for (int g = 0; g < ngrids; g++) {
    if (per_grid_element_natoms[g] > 0) per_grid_element_score[g] /= per_grid_element_natoms[g];
  }

  // Print statistics
  if (print_verbose) {
    printf("  Element Scores ...\n");
    printf("    Element Score %s = %12g\n", "All",  total_element_score);
    for (int g = 0; g < ngrids; g++) 
      printf("    Element Score %s = %12g\n", PDBelements[g].Name(), per_grid_element_score[g]);
    fflush(stdout);
  }

  // Compute per cell statistics
  int total_cell_natoms = 0;
  RNScalar total_cell_score = 0;
  int *per_grid_cell_natoms = new int [ ngrids ];
  RNScalar *per_grid_cell_score = new RNScalar [ ngrids ];
  for (int g = 0; g < ngrids; g++) per_grid_cell_natoms[g] = 0; 
  for (int g = 0; g < ngrids; g++) per_grid_cell_score[g] = 0; 
  for (int i = 0; i < ligand->NAtoms(); i++) {
    PDBAtom *atom = ligand->Atom(i);

    // Get element
    PDBElement *element = atom->Element();
    if (!element) continue;
    int element_index = element - PDBelements;
    assert((element_index >= 0) && (element_index < PDBnelements));
    if ((element_index < 0) || (element_index >= ngrids)) element_index = 0;

    // Compute statistics for cells at grid position
    RNScalar cell_sum = 0;
    for (int g = 0; g < ngrids; g++) cell_sum += grids[g]->WorldValue(atom->Position());
    RNScalar cell_avg = cell_sum / ngrids;
    RNScalar cell_sse = 0;
    for (int g = 0; g < ngrids; g++) { RNScalar delta = grids[g]->WorldValue(atom->Position()) - cell_avg; cell_sse += delta * delta; }
    RNScalar cell_stddev = sqrt(cell_sse / ngrids);
    if (cell_stddev > 0) {
      RNScalar score = (grids[element_index]->WorldValue(atom->Position()) - cell_avg) / cell_stddev;
      total_cell_score += score;
      total_cell_natoms++;
      per_grid_cell_score[element_index] += score;
      per_grid_cell_natoms[element_index]++;
    }
  }

  // Divide cell score by number of atoms to get average score per atom
  if (total_cell_natoms > 0) total_cell_score /= total_cell_natoms;
  for (int g = 0; g < ngrids; g++) {
    if (per_grid_cell_natoms[g] > 0) per_grid_cell_score[g] /= per_grid_cell_natoms[g];
  }

  // Print statistics
  if (print_verbose) {
    printf("  Cell Scores ...\n");
    printf("    Cell Score %s = %12g\n", "All",  total_cell_score);
    for (int g = 0; g < ngrids; g++) 
      printf("    Cell Score %s = %12g\n", PDBelements[g].Name(), per_grid_cell_score[g]);
    fflush(stdout);
  }

  // Print plot
  if (print_plots) {
    // Create plot for each grid
    int plot_count = 0;
    for (int g = 0; g < ngrids; g++) {
      // Determine scale factor to take number of grid cells to a volume
      RNScalar length_scale_factor = grids[g]->GridToWorldScaleFactor();
      RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

      // Create grid with ligand rasterized
      R3Grid ligand_grid(grids[g]->XResolution(), grids[g]->YResolution(), grids[g]->ZResolution());
      ligand_grid.SetWorldToGridTransformation(grids[g]->WorldToGridTransformation());
      for (int i = 0; i < ligand->NAtoms(); i++) {
        PDBAtom *atom = ligand->Atom(i);
        PDBElement *element = atom->Element();
        if (!element) continue;
        int element_index = element - PDBelements;
        assert((element_index >= 0) && (element_index < PDBnelements));
        if ((element_index < 0) || (element_index >= ngrids)) element_index = 0;
        if (element_index != g) continue;
        ligand_grid.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
      }

      // Allocate array of values found at positions of ligand atoms
      int ligand_nvalues = ligand_grid.Cardinality();
      if (ligand_nvalues == 0) continue;
      RNScalar *ligand_values = new RNScalar [ ligand_nvalues ];
      assert(ligand_values);

      // Fill array of ligand values 
      int count = 0;
      for (int j = 0; j < grids[g]->NEntries(); j++) {
        if (ligand_grid.GridValue(j) == 0) continue;
        assert(count < ligand_nvalues);
        ligand_values[count++] = grids[g]->GridValue(j);
      }

      // Sort the values
      qsort(ligand_values, ligand_nvalues, sizeof(RNScalar), RNCompareScalars);

      // Print plot
      char plot_name[128];
      sprintf(plot_name, "%s/%s.plt", grid_name, PDBelements[g].Name());
      int status = PrintPlot(grids[g], ligand_values, ligand_nvalues, plot_name);
      assert(status);

      // Print ligand plot
      char ligand_plot_name[128];
      sprintf(ligand_plot_name, "%s/%s_L.plt", grid_name, PDBelements[g].Name());
      FILE *fp = fopen(ligand_plot_name, "w");
      assert(fp);
      fprintf(fp, "0 %g\n", ligand_nvalues * volume_scale_factor);
      fprintf(fp, "1 %g\n", ligand_nvalues * volume_scale_factor);
      fclose(fp);

      // Delete ligand values
      delete [] ligand_values;

      // Count plots
      plot_count++;
    }

    // Print statistics
    if (print_verbose) {
      printf("  # Plots = %d\n", plot_count);
      fflush(stdout);
    }
  }
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: map2grd pdbfile mapdir griddir [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-asa")) { 
        argc--; argv++; asa_name = *argv; 
      }
      else if (!strcmp(*argv, "-ligand")) { 
        argc--; argv++; ligand_name = *argv; 
      }
      else if (!strcmp(*argv, "-spacing")) { 
        argc--; argv++; grid_spacing = atof(*argv); 
      }
      else if (!strcmp(*argv, "-nearby")) { 
        argc--; argv++; nearby_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-radius")) { 
        argc--; argv++; world_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-center")) { 
        argc--; argv++; world_centroid[0] = atof(*argv); 
        argc--; argv++; world_centroid[1] = atof(*argv); 
        argc--; argv++; world_centroid[2] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; max_resolution = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-sigma")) { 
        argc--; argv++; world_sigma = atof(*argv); 
      }
      else if (!strcmp(*argv, "-normalization")) { 
        argc--; argv++; normalization_type = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-mask_to_ligand")) { 
        mask_to_ligand = TRUE; 
      }
      else if (!strcmp(*argv, "-mask_to_target_volume")) { 
        mask_to_target_volume = TRUE; 
      }
      else if (!strcmp(*argv, "-target_volume")) { 
        argc--; argv++; target_volume_default = atof(*argv); 
      }
      else if (!strcmp(*argv, "-target_volume_from_ligand")) { 
        target_volume_from_ligand = TRUE; 
      }
      else if (!strcmp(*argv, "-target_volume_from_ligand_elements")) { 
        target_volume_from_ligand_elements = TRUE; 
      }
      else if (!strcmp(*argv, "-largest_connected_component_only")) { 
        largest_connected_component_only = TRUE; 
      }
      else if (!strcmp(*argv, "-debug")) { 
        print_debug = 1; 
      }
      else if (!strcmp(*argv, "-plot")) { 
        print_plots = 1; 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else if (!map_name) map_name = *argv;
      else if (!grid_name) grid_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check map name
  if (!map_name) {
    fprintf(stderr, "You did not specify a grid base filename.\n");
    return 0;
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check world radius
  if (RNIsZero(world_radius)) {
    fprintf(stderr, "Invalid world radius: %g\n", world_radius);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read pdb
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Read accessible surface area file
  if (asa_name) {
    int status = ReadASAFile(file, asa_name);
    if (!status) exit(-1);
  }

  // Find ligand
  PDBResidue *ligand = NULL;
  if (ligand_name) {
    ligand = FindLigand(file, ligand_name);
    if (!ligand) exit(-1);
  }

  // Read map 
  Map *map = ReadMap(map_name);
  if (!map) exit(-1);

  // Create grids
  int ngrids = map->NElements();
  R3Grid **grids = CreateGrids(file, ligand, map);
  if (!grids) exit(-1);

  // Blur grids
  int status3 = BlurGrids(grids, ngrids, world_sigma);
  if (!status3) exit(-1);

  // Mask grids
  int status4 = MaskGrids(file, ligand, grids, ngrids);
  if (!status4) exit(-1);

  // Normalize grids
  int status5 = NormalizeGrids(grids, ngrids, normalization_type);
  if (!status5) exit(-1);

  // Analyze ligand
  if (ligand) AnalyzeLigand(ligand, grids, ngrids);

  // Write grids
  int status6= WriteGrids(grids, ngrids, grid_name);
  if (!status6) exit(-1);

  // Return success
  return 0;
}
