// Source file for the map class



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"
#include "map.h"



// Point map storage type

struct MapPoint {
  MapPoint(const R3Point& position, RNScalar value) 
    : position(position), value(value) {};
  R3Point position;
  RNScalar value;
};



Map:: 
Map(void)
  : frame_type(0),
    atom_type(0),
    storage_type(0),
    nframes(0),
    nelements(0),
    world_radius(0),
    distributions(NULL),
    res(0)
{
}



Map:: 
Map(int frame_type, int atom_type, int storage_type, int nelements, RNScalar world_radius, int res)
  : frame_type(frame_type),
    atom_type(atom_type),
    storage_type(storage_type),
    nframes(0),
    nelements(nelements),
    world_radius(world_radius),
    res(res)
{
  // Determine number of frames
  if (frame_type == RESIDUE_MAP_FRAME) nframes = PDBnaminoacids;
  else if (frame_type == ATOM_MAP_FRAME) nframes = NAtomTypes();
  else if (frame_type == TRIPLE_MAP_FRAME) nframes = NAtomTypes() * NAtomTypes() * NAtomTypes();
  else if (frame_type == SITE_MAP_FRAME) nframes = 1;
  else RNAbort("Invalid map frame type: %d", frame_type); 

  // Allocate grid pointers
  assert(nframes > 0);
  assert(nelements > 0);
  distributions = new void **[ nframes ];
  assert(distributions);
  for (int i = 0; i < nframes; i++) {
    distributions[i] = new void * [ nelements ];
    assert(distributions[i]);
    for (int j = 0; j < nelements; j++) {
      distributions[i][j] = NULL;
    }
  }
}



Map:: 
~Map(void)
{
  // Delete distributions
  for (int i = 0; i < nframes; i++) {
    for (int j = 0; j < nelements; j++) {
      if (distributions[i][j]) {
        if (storage_type == POINT_MAP_STORAGE) {
          RNArray<MapPoint *> *points = (RNArray<MapPoint *> *) distributions[i][j];
          for (int k = 0; k < points->NEntries(); k++) delete points->Kth(k);
          delete points;
        }
        else if (storage_type == GRID_MAP_STORAGE) {
          R3Grid *grid = (R3Grid *) distributions[i][j];
          delete grid;
        }
        else if (storage_type == DISTANCE_MAP_STORAGE) {
          RNScalar *histogram = (RNScalar *) distributions[i][j];
          delete [] histogram;
        }
        else {
          RNAbort("Unknown map storage type");
        }
      }
    }
    delete [] distributions[i];
  }
  delete [] distributions;
}



void Map::
AddPoint(int frame_index, int element_index, const R3Point& position, RNScalar value)
{
  // Check storage type
  if (storage_type == POINT_MAP_STORAGE) {
    // Insert into array of points
    RNArray<MapPoint *> *points = (RNArray<MapPoint *> *) distributions[frame_index][element_index]; 
    if (!points) {
      points = new RNArray<MapPoint *>();
      distributions[frame_index][element_index] = points;
    }
    points->Insert(new MapPoint(position, value));
  }
  else if (storage_type == GRID_MAP_STORAGE) {
    // Rasterize into grid
    R3Grid *grid = (R3Grid *) distributions[frame_index][element_index];
    if (!grid) {
      grid = new R3Grid(res, res, res);
      R3Point world_corner(world_radius, world_radius, world_radius);
      R3Box world_box(-world_corner, world_corner);
      grid->SetWorldToGridTransformation(world_box);
      distributions[frame_index][element_index] = grid;
    }
    grid->RasterizeWorldPoint(position, value);
  }
  else if (storage_type == DISTANCE_MAP_STORAGE) {
    // Insert into distance histogram
    RNScalar *histogram = (RNScalar *) distributions[frame_index][element_index];
    if (!histogram) {
      histogram = new RNScalar [ res ];
      for (int i = 0; i < res; i++) histogram[i] = 0; 
      distributions[frame_index][element_index] = histogram;
    }
    RNLength distance = sqrt(position[0]*position[0] + position[1]*position[1] + position[2]*position[2]);
    int bin1 = (int) (res * distance / world_radius);
    int bin2 = bin1 + 1;
    RNLength delta = distance - bin1 * world_radius / res;
    assert((bin1 >= 0) && (bin2 >= 0));
    assert((delta >= 0) && (delta <= 1));
    if (bin1 < res) histogram[bin1] += value * (1 - delta);
    if (bin2 < res) histogram[bin2] += value * delta;
  }
  else {
    RNAbort("Unknown map storage type");
  }
}



void Map::
SplatIntoGrid(int frame_index, int element_index, R3Grid *grid, const R3Affine& frame_to_world)
{
  // Check distributions
  if (!distributions[frame_index][element_index]) return;

  // Splat distributions depending on storage type
  if (storage_type == POINT_MAP_STORAGE) {
    RNArray<MapPoint *> *points = (RNArray<MapPoint *> *) distributions[frame_index][element_index];
    for (int k = 0; k < points->NEntries(); k++) {
      MapPoint *point = points->Kth(k);
      R3Point world_position = point->position;
      world_position.Transform(frame_to_world);
      grid->RasterizeWorldPoint(world_position, point->value);
    }
  }
  else if (storage_type == GRID_MAP_STORAGE) {
    R3Grid *map_grid = (R3Grid *) distributions[frame_index][element_index];
    assert(map_grid->XResolution() == res);
    assert(map_grid->YResolution() == res);
    assert(map_grid->ZResolution() == res);
    for (int iz = 0; iz < res; iz++) {
      for (int iy = 0; iy <  res; iy++) {
        for (int ix = 0; ix <  res; ix++) {
          RNScalar value = map_grid->GridValue(ix, iy, iz);
          if (value > 0) {
            R3Point world_position = map_grid->WorldPosition(ix, iy, iz);
            world_position.Transform(frame_to_world);
            grid->RasterizeWorldPoint(world_position, value);
          }
        }
      }
    }
  }
  else if (storage_type == DISTANCE_MAP_STORAGE) {
    RNScalar *histogram = (RNScalar *) distributions[frame_index][element_index];
    int res_squared = res * res;
    RNScalar scale = world_radius / res;
    for (int iz = -res+1; iz < res; iz++) {
      int iz_squared = iz * iz;
      for (int iy = -res+1; iy < res; iy++) {
        int iy_squared = iy * iy;
        for (int ix = -res+1; ix < res; ix++) {
          int ix_squared = ix * ix;
          int distance_squared = ix_squared + iy_squared + iz_squared;
          if (distance_squared < res_squared) {
            int bin = (int) sqrt((RNScalar) distance_squared);
            if (histogram[bin] > 0) {
              RNScalar value = histogram[bin] / distance_squared;
              R3Point world_position = R3Point(ix * scale + 0.5, iy * scale + 0.5, iz * scale + 0.5);
              world_position.Transform(frame_to_world);
              grid->RasterizeWorldPoint(world_position, value);
            }
          }
        }
      }
    }
  }
  else {
    RNAbort("Unknown map storage type");
  }
}



void Map::
Blend(void)
{
  // Determine sigma (try to keep L1 norm constant)
  RNScalar world_sigma = world_radius / 3;

  // Check storage type
  if (storage_type == POINT_MAP_STORAGE) {
    // Compute Gaussian variables
    RNScalar a = sqrt(2.0 * RN_PI) * world_sigma;
    RNScalar fac = 1.0 / (a * a * a);
    RNScalar denom = 2.0 * world_sigma * world_sigma;

    // Blend all distributions
    for (int i = 0; i < nframes; i++) {
      for (int j = 0; j < nelements; j++) {
        RNArray<MapPoint *> *points = (RNArray<MapPoint *> *) distributions[i][j];
        if (points) {
          for (int k = 0; k < points->NEntries(); k++) {
            MapPoint *point = points->Kth(k);
            RNScalar d = R3Distance(point->position, R3zero_point);
            RNScalar factor = fac * exp(-d * d / denom);
            point->value *= factor;
          }
        }
      }
    }
  }
  else if (storage_type == GRID_MAP_STORAGE) {
    // Construct grid with Gaussian extending from center
    R3Grid gauss_grid(res, res, res);
    RNScalar center = 0.5 * (res-1);
    RNScalar grid_sigma = world_sigma * (res-1) / (2.0 * world_radius);
    RNScalar a = sqrt(2.0 * RN_PI) * grid_sigma;
    RNScalar fac = 1.0 / (a * a * a);
    RNScalar denom = 2.0 * grid_sigma * grid_sigma;
    for (int i = 0; i < res; i++) {
      RNScalar dx = i - center;
      for (int j = 0; j < res; j++) {
        RNScalar dy = j - center;
        for (int k = 0; k < res; k++) {
          RNScalar dz = k - center;
          RNScalar d = sqrt(dx*dx + dy*dy + dz*dz);
          gauss_grid.SetGridValue(i, j, k, fac * exp(-d * d / denom));
        }
      }
    }

    // Blend all distributions
    for (int i = 0; i < nframes; i++) {
      for (int j = 0; j < nelements; j++) {
        R3Grid *grid = (R3Grid *) distributions[i][j];
        if (grid) grid->Multiply(gauss_grid);
      }
    }
  }
  else if (storage_type == DISTANCE_MAP_STORAGE) {
    // Compute Gaussian variables
    RNScalar a = sqrt(2.0 * RN_PI) * world_sigma;
    RNScalar fac = 1.0 / (a * a * a);
    RNScalar denom = 2.0 * world_sigma * world_sigma;

    // Blend all distributions
    for (int i = 0; i < nframes; i++) {
      for (int j = 0; j < nelements; j++) {
        RNScalar *histogram = (RNScalar *) distributions[i][j];
        if (histogram) {
          for (int k = 0; k < res; k++) {
            RNScalar d = world_radius * (RNScalar) k / (RNScalar) (res-1);
            RNScalar factor = fac * exp(-d * d / denom);
            histogram[k] *= factor;
          }
        }
      }
    }
  }
  else {
    RNAbort("Unknown map storage type");
  }
}



void Map::
Blur(RNLength world_sigma)
{
  // Check storage type
  if (storage_type == POINT_MAP_STORAGE) {
    RNAbort("Blur not implemented for point map storage type\n");
  }
  else if (storage_type == GRID_MAP_STORAGE) {
    // Blur all distributions
    for (int i = 0; i < nframes; i++) {
      for (int j = 0; j < nelements; j++) {
        R3Grid *grid = (R3Grid *) distributions[i][j];
        if (grid) grid->Blur(world_sigma * grid->WorldToGridScaleFactor());
      }
    }
  }
  else if (storage_type == DISTANCE_MAP_STORAGE) {
    RNAbort("Blur not implemented for distance map storage type\n");
  }
  else {
    RNAbort("Unknown map storage type");
  }
}



void Map::
Normalize(void)
{
  // Divide each distribution by L1Norm of whole map

  // Compute sum 
  RNScalar sum = 0;
  for (int i = 0; i < nframes; i++) {
    for (int j = 0; j < nelements; j++) {
      sum += L1Norm(i, j);
    }
  }

  // Divide by sum
  for (int i = 0; i < nframes; i++) {
    for (int j = 0; j < nelements; j++) {
      if (RNIsZero(sum)) continue;
      Scale(i, j, 1 / sum);
    }
  }
}



int Map::
WriteFile(const char *filename)
{
  // Open map file
  FILE *fp = fopen(filename, "wb");
  if (!fp) {
    fprintf(stderr, "Unable to open map file: %s\n", filename);
    return 0;
  }

  // Initialize error status
  int error = 0;

  // Write map header info
  if (fwrite(&frame_type, sizeof(int), 1, fp) != 1) error = 1;
  if (fwrite(&atom_type, sizeof(int), 1, fp) != 1) error = 1;
  if (fwrite(&storage_type, sizeof(int), 1, fp) != 1) error = 1;
  if (fwrite(&nframes, sizeof(int), 1, fp) != 1) error = 1;
  if (fwrite(&nelements, sizeof(int), 1, fp) != 1) error = 1;
  if (fwrite(&world_radius, sizeof(RNScalar), 1, fp) != 1) error = 1;
  if (fwrite(&res, sizeof(int), 1, fp) != 1) error = 1;
  if (error) {
    fprintf(stderr, "Unable to write map file: %s\n", filename);
    return 0;
  }

  // Write map distributions
  for (int i = 0; i < nframes; i++) {
    for (int j = 0; j < nelements; j++) {
      // Check storage type
      if (storage_type == POINT_MAP_STORAGE) {
        RNArray<MapPoint *> *points = (RNArray<MapPoint *> *) distributions[i][j];
        int npoints = (points) ? points->NEntries() : 0;
        if (fwrite(&npoints, sizeof(int), 1, fp) != 1) error = 1;
        for (int k = 0; k < npoints; k++) {
          if (fwrite(points->Kth(k), sizeof(MapPoint), 1, fp) != 1) error = 1;
        }
      }
      else if (storage_type == GRID_MAP_STORAGE) {
        R3Grid *grid = (R3Grid *) distributions[i][j];
        int found = (grid) ? 1 : 0;
        if (fwrite(&found, sizeof(int), 1, fp) != 1) error = 1;
        if (found) { if (!grid->Write(fp)) error = 1; }
      }
      else if (storage_type == DISTANCE_MAP_STORAGE) {
        RNScalar *histogram = (RNScalar *) distributions[i][j];
        int nbins = (histogram) ? res : 0;
        if (fwrite(&nbins, sizeof(int), 1, fp) != 1) error = 1;
        if (fwrite(histogram, sizeof(RNScalar), nbins, fp) != (unsigned int) nbins) error = 1;
      }
      else {
        RNAbort("Unknown map storage type");
      }

      // Check for error
      if (error) {
        fprintf(stderr, "Error writing map file: %s\n", filename);
        return 0;
      }
    }
  }

  // Close map file 
  fclose(fp);

  // Return success
  return 1;
}



int Map::
ReadFile(const char *filename)
{
  // Open map file
  FILE *fp = fopen(filename, "rb");
  if (!fp) {
    fprintf(stderr, "Unable to open map file: %s\n", filename);
    return 0;
  }

  // Initialize error status
  int error = 0;

  // Read map header info
  if (fread(&frame_type, sizeof(int), 1, fp) != 1) error = 1;
  if (fread(&atom_type, sizeof(int), 1, fp) != 1) error = 1;
  if (fread(&storage_type, sizeof(int), 1, fp) != 1) error = 1;
  if (fread(&nframes, sizeof(int), 1, fp) != 1) error = 1;
  if (fread(&nelements, sizeof(int), 1, fp) != 1) error = 1;
  if (fread(&world_radius, sizeof(RNScalar), 1, fp) != 1) error = 1;
  if (fread(&res, sizeof(int), 1, fp) != 1) error = 1;
  if (error) {
    fprintf(stderr, "Unable to read map file: %s\n", filename);
    return 0;
  }

  // Check frame type
  if ((frame_type < 0) && (frame_type >= NUM_MAP_FRAME_TYPES)) {
    fprintf(stderr, "Invalid frame type (%d) in map file: %s\n", frame_type, filename);
    return 0;
  }

  // Check atom type
  if ((atom_type < 0) && (atom_type >= NUM_MAP_ATOM_TYPES)) {
    fprintf(stderr, "Invalid atom type (%d) in map file: %s\n", atom_type, filename);
    return 0;
  }

  // Check map type
  if ((storage_type < 0) && (storage_type >= NUM_MAP_STORAGE_TYPES)) {
    fprintf(stderr, "Invalid storage type (%d) in map file: %s\n", storage_type, filename);
    return 0;
  }

  // Check nframes
  if (nframes == 0) {
    fprintf(stderr, "Invalid nframes (%d) in map file: %s\n", nframes, filename);
    return 0;
  }

  // Check nelements
  if (nelements == 0) {
    fprintf(stderr, "Invalid nelements (%d) in map file: %s\n", nelements, filename);
    return 0;
  }

  // Check world radius
  if (world_radius == 0) {
    fprintf(stderr, "Invalid world radius (%g) in map file: %s\n", world_radius, filename);
    return 0;
  }

  // Check resolution
  if (res == 0) {
    fprintf(stderr, "Invalid resolution (%d) in map file: %s\n", res, filename);
    return 0;
  }


  // Allocate distributions pointer arrays
  distributions = new void **[ nframes ];
  if (!distributions) {
    fprintf(stderr, "Unable to allocated distributions for map\n");
    return 0;
  }

  // Read map distributions
  for (int i = 0; i < nframes; i++) {
    // Allocate distributions pointers
    distributions[i] = new void * [ nelements ];
    if (!distributions[i]) {
      fprintf(stderr, "Unable to allocated pointers for map\n");
      return 0;
    }

    // Read distributions
    for (int j = 0; j < nelements; j++) {
      // Initailize distributions
      distributions[i][j] = NULL;

      // Read distributions according to storage type
      if (storage_type == POINT_MAP_STORAGE) {
        int npoints = 0;
        if (fread(&npoints, sizeof(int), 1, fp) != 1) error = 1;
        if (npoints > 0) {
          RNArray<MapPoint *> *points = new RNArray<MapPoint *>();
          distributions[i][j] = points;
          for (int k = 0; k < npoints; k++) {
            MapPoint *point = new MapPoint(R3zero_point, 0);
            if (fread(point, sizeof(MapPoint), 1, fp) != 1) error = 1;
            points->Insert(point);
          }
        }
      }
      else if (storage_type == GRID_MAP_STORAGE) {
        int found = 0;
        if (fread(&found, sizeof(int), 1, fp) != 1) error = 1;
        if (found) {
          R3Grid *grid = new R3Grid();
          distributions[i][j] = grid;
          if (!grid->Read(fp)) error = 1;
        }
      }
      else if (storage_type == DISTANCE_MAP_STORAGE) {
        int nbins = 0;
        if (fread(&nbins, sizeof(int), 1, fp) != 1) error = 1;
        if (nbins > 0) {
          assert(nbins == res);
          RNScalar *histogram = new RNScalar [ nbins ];
          distributions[i][j] = histogram;
          if (fread(histogram, sizeof(RNScalar), nbins, fp) != (unsigned int) nbins) error = 1;
        }
      }
      else {
        RNAbort("Unknown map storage type");
      }

      // Check for error
      if (error) {
        fprintf(stderr, "Error reading map file: %s\n", filename);
        return 0;
      }
    }
  }

  // Close map file 
  fclose(fp);

  // Return success
  return 1;
}



RNScalar Map::
L1Norm(int i, int j) const
{
  // Initialize result
  RNScalar sum = 0;

  // Check distribution
  if (!distributions[i][j]) return sum;

  // Check storage type
  if (storage_type == POINT_MAP_STORAGE) {
    RNArray<MapPoint *> *points = (RNArray<MapPoint *> *) distributions[i][j];
    for (int k = 0; k < points->NEntries(); k++) {
      MapPoint *point = points->Kth(k);
      sum += point->value;
    }
  }
  else if (storage_type == GRID_MAP_STORAGE) {
    R3Grid *grid = (R3Grid *) distributions[i][j];
    sum = grid->L1Norm();
  }
  else if (storage_type == DISTANCE_MAP_STORAGE) {
    RNScalar *histogram = (RNScalar *) distributions[i][j];
    for (int k = 0; k < res; k++) {
      sum += histogram[k];
    }
  }
  else {
    RNAbort("Unknown map storage type");
  }

  // Return L1 norm
  return sum;
}



void Map::
Scale(int i, int j, RNScalar factor)
{
  // Check distribution
  if (!distributions[i][j]) return;

  // Check storage type
  if (storage_type == POINT_MAP_STORAGE) {
    RNArray<MapPoint *> *points = (RNArray<MapPoint *> *) distributions[i][j];
    for (int k = 0; k < points->NEntries(); k++) {
      MapPoint *point = points->Kth(k);
      point->value *= factor;
    }
  }
  else if (storage_type == GRID_MAP_STORAGE) {
    R3Grid *grid = (R3Grid *) distributions[i][j];
    grid->Multiply(factor);
  }
  else if (storage_type == DISTANCE_MAP_STORAGE) {
    RNScalar *histogram = (RNScalar *) distributions[i][j];
    for (int k = 0; k < res; k++) {
      histogram[k] *= factor;
    }
  }
  else {
    RNAbort("Unknown map storage type");
  }
}



void Map::
PrintStatistics(void) const
{
  // Allocate memory for sums
  RNScalar *frame_counts = new RNScalar [ nframes ];
  RNScalar *element_counts = new RNScalar [ nelements ];
  assert(frame_counts && element_counts);
  for (int i = 0; i < nframes; i++) frame_counts[i] = 0;
  for (int i = 0; i < nelements; i++) element_counts[i] = 0;

  // Print atom counts for each map
  RNScalar total_count = 0;
  printf("  Map Counts:\n");
  for (int i = 0; i < nframes; i++) {
    for (int j = 0; j < nelements; j++) {
      // Check storage type
      if (storage_type == POINT_MAP_STORAGE) {
        RNArray<MapPoint *> *points = (RNArray<MapPoint *> *) distributions[i][j];
        if (!points) continue;
        RNScalar count = points->NEntries();
        printf("    Map Count %s %s = %g\n", FrameName(i), PDBelements[j].Name(), count);
        frame_counts[i] += count;
        element_counts[j] += count;
        total_count += count;
      }
      else if (storage_type == GRID_MAP_STORAGE) {
        R3Grid *grid = (R3Grid *) distributions[i][j];
        if (!grid) continue;
        RNScalar count = grid->L1Norm();
        printf("    Map Count %s %s = %g\n", FrameName(i), PDBelements[j].Name(), count);
        frame_counts[i] += count;
        element_counts[j] += count;
        total_count += count;
      }
      else if (storage_type == DISTANCE_MAP_STORAGE) {
        RNScalar *histogram = (RNScalar *) distributions[i][j];
        if (!histogram) continue;
        RNScalar count = 0;
        for (int k = 0; k < res; k++) count += histogram[k];
        printf("    Map Count %s %s = %g\n", FrameName(i), PDBelements[j].Name(), count);
        frame_counts[i] += count;
        element_counts[j] += count;
        total_count += count;
      }
      else {
        RNAbort("Unknown map storage type");
      }
    }
  }
  printf("    Map Count Average = %g\n", total_count / (nframes * nelements));
  printf("    Map Count Total = %g\n", total_count);
  printf("\n");

  // Print atom counts for each frame
  printf("  Frame Counts:\n");
  for (int i = 0; i < nframes; i++) {
    printf("    Frame Count %s = %g\n", FrameName(i), frame_counts[i]);
  }
  printf("    Frame Count Average = %g\n", total_count / nframes);
  printf("    Frame Count Total = %g\n", total_count);
  printf("\n");

  // Print atom counts for each element
  printf("  Element Counts:\n");
  for (int i = 0; i < nelements; i++) {
    printf("    Element Count %s = %g\n", PDBelements[i].Name(), element_counts[i]);
  }
  printf("    Element Count Average = %g\n", total_count / nelements);
  printf("    Element Count Total = %g\n", total_count);
  printf("\n");

  // Print L1Norms for each map
  total_count = 0;
  for (int i = 0; i < nframes; i++) frame_counts[i] = 0;
  for (int i = 0; i < nelements; i++) element_counts[i] = 0;
  printf("  Map L1Norms:\n");
  for (int i = 0; i < nframes; i++) {
    for (int j = 0; j < nelements; j++) {
      RNScalar count = L1Norm(i, j);
      if (count == 0) continue;
      printf("    Map L1Norm %s %s = %g\n", FrameName(i), PDBelements[j].Name(), count);
      frame_counts[i] += count;
      element_counts[j] += count;
      total_count += count;
    }
  }
  printf("    Map L1Norm Average = %g\n", total_count / (nframes * nelements));
  printf("    Map L1Norm Total = %g\n", total_count);
  printf("\n");

  // Print L1Norms for each frame
  printf("  Frame L1Norms:\n");
  for (int i = 0; i < nframes; i++) {
    printf("    Frame L1Norm %s = %g\n", FrameName(i), frame_counts[i]);
  }
  printf("    Frame L1Norm Average = %g\n", total_count / nframes);
  printf("    Frame L1Norm Total = %g\n", total_count);
  printf("\n");

  // Print atom counts for each element
  printf("  Element L1Norm:\n");
  for (int i = 0; i < nelements; i++) {
    printf("    Element L1Norm %s = %g\n", PDBelements[i].Name(), element_counts[i]);
  }
  printf("    Element L1Norm Average = %g\n", total_count / nelements);
  printf("    Element L1Norm Total = %g\n", total_count);
  printf("\n");

  // Delete memory for counts
  delete [] frame_counts;
  delete [] element_counts;
}



int Map::
NAtomTypes(void) const
{
  // Return number of atom types for triads
  switch (atom_type) {
  case ONE_MAP_ATOM: return 1; 
  case ELEMENT_MAP_ATOM: return 3; 
  case AMINOACID_MAP_ATOM: return PDB_NUM_AMINOACIDS; 
  case XSITE_MAP_ATOM: return 11; 
  case ENGHHUBER_MAP_ATOM: return PDB_NUM_ATOM_GROUPS-1;
  case UNIQUE_MAP_ATOM: return PDB_NUM_ATOMS-1;
  default: 
    RNAbort("Unrecognized map type: %d", frame_type);
  }
  return 0;
}



int Map::
AtomType(PDBAtom *atom) const
{
  // Check atom
  if (atom->IsHetAtom()) return 0;

  // Check map type
  switch (atom_type) {
  case ONE_MAP_ATOM:
    return 0;

  case ELEMENT_MAP_ATOM: {
    // Return element type for atom (0=C, 1=N, 2=O)
    PDBElement *element = atom->Element();
    if (!element) return -1;
    int element_id = element->ID();
    if (element_id > PDB_O_ELEMENT) return -1;
    return element_id - 1;
  } break;

  case AMINOACID_MAP_ATOM: {
    // Return aminoacid id
    PDBResidue *residue = atom->Residue();
    if (!residue) return -1;
    PDBAminoAcid *aminoacid = residue->AminoAcid();
    if (!aminoacid) return -1;
    return aminoacid->ID();
    } break;

  case XSITE_MAP_ATOM:
    // Return atom group from Xsite
    switch (PDBaminoacid_atoms[atom->aminoacid_atom_type].group) {
    case PDB_NH1_ATOM_GROUP: 
    case PDB_NH1S_ATOM_GROUP: 
      return 0;
    case PDB_NC2_ATOM_GROUP:                                  
    case PDB_NH3_ATOM_GROUP:                                  
      return 1;
    case PDB_NH2_ATOM_GROUP:                                  
    case PDB_N_ATOM_GROUP:                                    
      return 2;
    case PDB_O_ATOM_GROUP:    
    case PDB_OS_ATOM_GROUP:   
      return 3;
    case PDB_OC_ATOM_GROUP:   
      return 4;
    case PDB_OH1_ATOM_GROUP:  
      return 5;
    case PDB_C_ATOM_GROUP:
    case PDB_CH0_ATOM_GROUP:  
      return 6;
    case PDB_CR1E_ATOM_GROUP: 
    case PDB_CR1W_ATOM_GROUP:                                 
    case PDB_CRHH_ATOM_GROUP:                                 
    case PDB_CR1H_ATOM_GROUP:                                 
      return 7;
    case PDB_CH1S_ATOM_GROUP: 
    case PDB_CH1E_ATOM_GROUP: 
    case PDB_CH2E_ATOM_GROUP: 
    case PDB_CH3E_ATOM_GROUP: 
    case PDB_CH2G_ATOM_GROUP: 
    case PDB_CH2P_ATOM_GROUP: 
    case PDB_CY2_ATOM_GROUP:                                  
    case PDB_CF_ATOM_GROUP:                                   
    case PDB_CY_ATOM_GROUP:                                   
    case PDB_CW_ATOM_GROUP:                                   
    case PDB_C5_ATOM_GROUP:                                   
    case PDB_C5W_ATOM_GROUP:                                  
      return 8;
    case PDB_SH1E_ATOM_GROUP:                                 
    case PDB_SM_ATOM_GROUP:   
      return 9;
    case PDB_HOH_ATOM_GROUP: 
      return 10;                                 
    case PDB_CL_ATOM_GROUP:                                   
    case PDB_F_ATOM_GROUP:                                    
    case PDB_S2_ATOM_GROUP:                                   
    case PDB_N4_ATOM_GROUP:                                   
    case PDB_N5_ATOM_GROUP:                                   
    case PDB_N6_ATOM_GROUP:                                   
    case PDB_O4_ATOM_GROUP:                                   
    case PDB_O6_ATOM_GROUP:                                   
    case PDB_C5N_ATOM_GROUP:                                  
    case PDB_P_ATOM_GROUP: 
    default: 
      return -1;
    }
    break;

  case ENGHHUBER_MAP_ATOM:
    // Return atom group from Engh & Huber
    return PDBaminoacid_atoms[atom->aminoacid_atom_type].group - 1;
    break;

  case UNIQUE_MAP_ATOM:
    // Return atom id
    return atom->aminoacid_atom_type - 1;
    break;

  default:
    RNAbort("Invalid map type: %d\n", frame_type);
    break;
  }
  
  // Return NULL atom type
  return -1;
}



char *Map::
FrameName(int frame_index) const
{
  static char frame_name[256];
  if (frame_type == ATOM_MAP_FRAME) {
    if (atom_type == ONE_MAP_ATOM) strcpy(frame_name, "ALL");
    else if (atom_type == ELEMENT_MAP_ATOM) strcpy(frame_name, PDBelements[frame_index+1].Name());
    else if (atom_type == AMINOACID_MAP_ATOM) strcpy(frame_name, PDBaminoacids[frame_index].Code());
    else if (atom_type == XSITE_MAP_ATOM) {
      switch (frame_index) {
      case 0: strcpy(frame_name, "NH1"); break;
      case 1: strcpy(frame_name, "NH3"); break;
      case 2: strcpy(frame_name, "N"); break;
      case 3: strcpy(frame_name, "O"); break;
      case 4: strcpy(frame_name, "OC"); break;
      case 5: strcpy(frame_name, "OH"); break;
      case 6: strcpy(frame_name, "C"); break;
      case 7: strcpy(frame_name, "CR"); break;
      case 8: strcpy(frame_name, "CH"); break;
      case 9: strcpy(frame_name, "S"); break;
      case 10: strcpy(frame_name, "HOH"); break;
      default: RNAbort("Invalid frame index"); break;
      }
    }
    else sprintf(frame_name, "%d", frame_index);
  } 
  else if (frame_type == TRIPLE_MAP_FRAME) {
    int natomtypes = NAtomTypes();
    int i1 = frame_index / (natomtypes * natomtypes);
    int i2 = (frame_index % (natomtypes * natomtypes)) / natomtypes;
    int i3 = frame_index % natomtypes;
    sprintf(frame_name, "%d-%d-%d", i1, i2, i3);
  }
  else if (frame_type == RESIDUE_MAP_FRAME) {
    strcpy(frame_name, PDBaminoacids[frame_index].Code());
  }
  else if (frame_type == SITE_MAP_FRAME) {
    strcpy(frame_name, "SITE");
  }
  return frame_name;
}



