// Program to generate D2 distribution



// Include files

#include "R3Shapes/R3Shapes.h"



// Program arguments

static char *modelname  = NULL;
static char *histname = NULL;
static RNBoolean print_verbose = FALSE;
static RNScalar max_distance = 0;
static int npoints = 8192;
static int nbins = 64;



int
ParseArgs (int argc, char **argv)
{
  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-npoints")) { argv++; argc--; npoints = atoi(*argv); }
      else if (!strcmp(*argv, "-nbins")) { argv++; argc--; nbins = atoi(*argv); }
      else if (!strcmp(*argv, "-max")) { argv++; argc--; max_distance = atof(*argv); }
      else if (!strcmp(*argv, "-v")) { print_verbose = TRUE; }
      else RNAbort("Invalid program argument: %s", *argv);
      argv++; argc--;
    }
    else {
      if (!modelname) modelname = *argv;
      else if (!histname) histname = *argv;
      else RNAbort("Invalid program argument: %s", *argv);
      argv++; argc--;
    }
  }

  // Check arguments
  if (!modelname) {
    // Print usage statement
    fprintf(stderr, 
      "Usage: off2D2<options> input_model output_histogram\n\n"
      "options:\n"
      " -npoints #   (default: 8192)\n"      
      " -nbins #     (default: 64)\n\n"
      "\n");
    return FALSE;
  }

  // Return success
  return TRUE;
}



static R3Mesh *
ReadMesh(char *mesh_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate mesh
  R3Mesh *mesh = new R3Mesh();
  assert(mesh);

  // Read mesh from file
  if (!mesh->ReadFile(mesh_name)) {
    delete mesh;
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read mesh ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Return success
  return mesh;
}



int
GeneratePoints(R3Mesh *mesh, int& npoints, int max_npoints, R3Point *points)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Count total area of triangles
  RNArea total_area = 0.0;
  RNArea *areas = new RNArea [mesh->NFaces()];
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    areas[i] = mesh->FaceArea(face);
    total_area += areas[i];
  }
    
  // Generate points
  int count = 0;
  RNSeedRandomScalar();
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);

    // Get vertex positions
    R3MeshVertex *v0 = mesh->VertexOnFace(face, 0);
    R3MeshVertex *v1 = mesh->VertexOnFace(face, 1);
    R3MeshVertex *v2 = mesh->VertexOnFace(face, 2);
    const R3Point& p0 = mesh->VertexPosition(v0);
    const R3Point& p1 = mesh->VertexPosition(v1);
    const R3Point& p2 = mesh->VertexPosition(v2);

    // Determine number of points for face 
    RNScalar ideal_face_npoints = npoints * areas[i] / total_area;
    int face_npoints = (int) ideal_face_npoints;
    RNScalar remainder = ideal_face_npoints - face_npoints;
    if (remainder > RNRandomScalar()) face_npoints++;

    // Generate random points in face
    for (int j = 0; j < face_npoints; j++) {
      if (count == max_npoints) break;
      RNScalar r1 = sqrt(RNRandomScalar());
      RNScalar r2 = RNRandomScalar();
      points[count++] = p0 * (1.0 - r1) + p1 * r1 * (1.0 - r2) + p2 * r1 * r2;
    }
  }

  // Update number of points
  assert(count <= max_npoints);
  assert(npoints <= max_npoints);
  npoints = count;

  // Print message
  if (print_verbose) {
    fprintf(stdout, "Generated points ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", npoints);
    fflush(stdout);
  }

  // Return number of points
  return count;
}



int
GenerateSamples(int npoints, R3Point points[], int nsamples, RNLength samples[])
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Generate samples
  int count = 0;
  for (int i = 0; i < npoints; i++) {
    for (int j = i+1; j < npoints; j++) {
      RNLength sample = R3Distance(points[i], points[j]);
      samples[count++] = sample;
    }
  }

  // Print message
  if (print_verbose) {
    fprintf(stdout, "Generated samples ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Samples = %d\n", count);
    fflush(stdout);
  }

  // Return number of samples
  assert(count == nsamples);
  return nsamples;
}



int
GenerateHistogram(int nsamples, RNLength *samples, int nbins, RNScalar *bins, RNScalar max_distance)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Determine max_distance
  if (max_distance == 0) {
    RNScalar sum = 0;
    for (int i = 0; i < nsamples; i++) sum += samples[i];
    max_distance = 3.0 * sum / nsamples;
  }

  // Determine scaling from samples to bins
  RNScalar xscale = (RNScalar) (nbins-1) / max_distance;

  // Initialize bins
  for (int i = 0; i < nbins; i++) {
    bins[i] = 0;
  }

  // Fill in bins
  for (int i = 0; i < nsamples; i++) {
    RNScalar x = xscale * samples[i];
    if (x >= nbins - 1) x = nbins - 1 - RN_EPSILON;
    else if (x < 0) x = 0;
    int ix = (int) x;
    assert(ix < nbins);
    RNScalar dx = x - ix;
    bins[ix] += 1 - dx;
    bins[ix+1] += dx;
  }

  // Normalize Y scaling (area under curve is 1000000);
  RNScalar yscale = 1000000.0 / nsamples;
  for (int i = 0; i < nbins; i++) 
    bins[i] = yscale * bins[i];

  // Print message
  if (print_verbose) {
    fprintf(stdout, "Generated histogram ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bins = %d\n", nbins);
    fflush(stdout);
  }

  // Return number of bins
  return nbins;
}



RNBoolean 
WriteHistogram(int nbins, RNScalar *bins, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *file = stdout;
  if (filename) {
    file = fopen(filename, "w");
    if (!file) {
      fprintf(stderr, "Unable to open output file %s\n", filename);
      return FALSE;
    }
  }

  // Write number of bins
  if (fprintf(file, "%d\n", nbins) < 0) {      
    fprintf(stderr, "Unable to write to output file %s\n", filename);
    return FALSE;
  }

  // Write bins
  for (int i = 0; i < nbins; i++) {
    if (fprintf(file, "%10.3f\n", bins[i]) < 0) {      
      fprintf(stderr, "Unable to write bin to output file %s\n", filename);
      return FALSE;
    }
  }

  // Close file
  if (filename) fclose(file);

  // Print message
  if (print_verbose) {
    fprintf(stdout, "Wrote histogram to %s ...\n", (filename) ? filename : "stdout");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return TRUE;
}



int
main(int argc, char **argv)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Parse args
  if(!ParseArgs(argc, argv)) exit(-1);

  // Init RING
  if (!R3InitShapes()) exit(-1);
  
  // Read the model
  R3Mesh *mesh = ReadMesh(modelname);
  if (!mesh) exit(-1);

  // Generate points
  int max_npoints = 2 * npoints;
  R3Point *points = new R3Point[max_npoints];
  if (!GeneratePoints(mesh, npoints, max_npoints, points)) exit(-1);

  // Generate samples
  int nsamples = npoints * (npoints - 1) / 2;
  RNLength *samples = new RNLength[nsamples];
  if (!GenerateSamples(npoints, points, nsamples, samples)) exit(-1);

  // Generate histogram
  RNScalar *bins = new RNScalar [nbins];
  if (!GenerateHistogram(nsamples, samples, nbins, bins, max_distance)) exit(-1);

  // Write histogram
  if (!WriteHistogram(nbins, bins, histname)) exit(-1);

  // Print message
  if (print_verbose) {
    fprintf(stdout, "Finished in %6.3f seconds.\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 0;
}








