// Include file for the map class



// Storage types

typedef enum {
  POINT_MAP_STORAGE,
  GRID_MAP_STORAGE,
  DISTANCE_MAP_STORAGE,
  NUM_MAP_STORAGE_TYPES
} MapStorageType;



// Frame types

typedef enum {
  ATOM_MAP_FRAME,
  TRIPLE_MAP_FRAME,
  RESIDUE_MAP_FRAME,
  SITE_MAP_FRAME,
  NUM_MAP_FRAME_TYPES
} MapFrameType;



// Atom types

typedef enum {
  ONE_MAP_ATOM,
  ELEMENT_MAP_ATOM,
  AMINOACID_MAP_ATOM,
  XSITE_MAP_ATOM,
  ENGHHUBER_MAP_ATOM,
  UNIQUE_MAP_ATOM,
  NUM_MAP_ATOM_TYPES
} MapAtomType;



// Class definition

class Map {
public:
  // Constructors/destructors
  Map(void);
  Map(int frame_type, int atom_type = RESIDUE_MAP_FRAME, int storage_type = POINT_MAP_STORAGE, 
      int nelements = 6, RNScalar world_radius = 10, int res = 32);
  ~Map(void);

  // Property functions
  int FrameType(void) const;
  int AtomType(void) const;
  int StorageType(void) const;
  int NFrames(void) const;
  int NElements(void) const;
  int NDistributions(void) const;
  RNScalar WorldRadius(void) const;
  RNScalar WorldSpacing(void) const;
  int Resolution(void) const;

  // Rasterization functions
  void AddPoint(int frame_index, int element_index, const R3Point& position, RNScalar value);
  void SplatIntoGrid(int frame_index, int element_index, R3Grid *grid, const R3Affine& transform);

  // Manipulation functions
  void Blend(void);
  void Normalize(void);
  void Blur(RNLength world_sigma);

  // I/O functions
  int ReadFile(const char *filename);
  int WriteFile(const char *filename);

  // Debug functions
  int NAtomTypes(void) const;
  int AtomType(PDBAtom *atom) const;
  char *FrameName(int frame_index) const;
  void PrintStatistics(void) const;

public:
  // Internal utility functions
  RNScalar L1Norm(int frame_index, int element_index) const;
  void Scale(int frame_index, int element_index, RNScalar factor);

public:
  int frame_type;
  int atom_type;
  int storage_type;
  int nframes;
  int nelements;
  RNScalar world_radius;
  void ***distributions;
  int res;
};



// Inline functions

inline int Map::
FrameType(void) const
{
  // Return frame type
  return frame_type;
}



inline int Map::
AtomType(void) const
{
  // Return atom type
  return atom_type;
}



inline int Map::
StorageType(void) const
{
  // Return storage type
  return storage_type;
}



inline int Map::
NFrames(void) const
{
  // Return number of frames
  return nframes;
}



inline int Map::
NElements(void) const
{
  // Return number of elements
  return nelements;
}



inline int Map::
NDistributions(void) const
{
  // Return number of distributions
  return nframes * nelements;
}



inline int Map::
Resolution(void) const
{
  // Return sampling resolution
  return res;
}



inline RNScalar Map::
WorldRadius(void) const
{
  // Return world radius
  return world_radius;
}



inline RNScalar Map::
WorldSpacing(void) const
{
  // Return world radius
  return 2.0 * world_radius / (res - 1);
}



