// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *pdb_name = NULL;
static char *vol_name = NULL;
static int grid_resolution[3] = { 64, 64, 64 };
static int rasterization_type = 0;  // 0=point, 1=solid, 2=surface
static RNBoolean count_only = 0;
static int nelements = 1;
static int print_verbose = 0;



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static float *
CreateCounts(PDBFile *file, int nelements)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get model from PDB file
  if (file->NModels() == 0) return NULL;
  PDBModel *model = file->Model(0);

  // Allocate array of counts
  float *counts = new float [ nelements ];
  if (!counts) {
    fprintf(stderr, "Unable to allocate counts\n");
    return NULL;
  }

  // Compute count for each element type separately
  RNScalar total_count = 0;
  for (int i = 0; i < nelements; i++) {
    // Rasterize atoms of element type
    for (int j = 0; j < model->NAtoms(); j++) {
      PDBAtom *atom = model->Atom(j);
      PDBElement *element = atom->Element();
      if (!element) continue;
      int element_index = element - PDBelements;
      assert((element_index >= 0) && (element_index < PDBnelements));
      if (element_index >= nelements) element_index = 0;
      if (i != element_index) continue;
      counts[i]++;
      total_count++;
    }

  }

  // Print statistics
  if (print_verbose) {
    printf("Created counts ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Counts = %d\n", nelements);
    printf("  Count Total = %g\n", total_count);
    for (int i = 0; i < nelements; i++) 
      printf("    Count %s = %g\n", PDBelements[i].Name(), counts[i]);
    fflush(stdout);
  }

  // Return counts
  return counts;
}



static float *
CreateVolumes(PDBFile *file, int nelements)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get model from PDB file
  if (file->NModels() == 0) return NULL;
  PDBModel *model = file->Model(0);

  // Allocate array of volumes
  float *volumes = new float [ nelements ];
  if (!volumes) {
    fprintf(stderr, "Unable to allocate volumes\n");
    return NULL;
  }

  // Allocate temporary grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    return NULL;
  }
 
  // Set grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Compute volume for each element type separately
  RNScalar total_volume = 0;
  for (int i = 0; i < nelements; i++) {
    // Clear grid
    grid->Clear();

    // Rasterize atoms of element type
    for (int j = 0; j < model->NAtoms(); j++) {
      PDBAtom *atom = model->Atom(j);
      PDBElement *element = atom->Element();
      if (!element) continue;
      int element_index = element - PDBelements;
      assert((element_index >= 0) && (element_index < PDBnelements));
      if (element_index >= nelements) element_index = 0;
      if (i != element_index) continue;
      if (rasterization_type == 0) grid->RasterizeWorldPoint(atom->Position(), 1.0);
      else grid->RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
    }

    // Compute surface, if needed
    if (rasterization_type == 2) {
      grid->Threshold(0.5, 0, 1);
      grid->DetectEdges(); 
      grid->Threshold(0.5, 0, 1);
    }

    // Compute grid volume
    volumes[i] = grid->Volume();
    total_volume += volumes[i];
  }

  // Delete grid
  delete grid;

  // Print statistics
  if (print_verbose) {
    printf("Created volumes ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Volumes = %d\n", nelements);
    printf("  Volume Total = %g\n", total_volume);
    for (int i = 0; i < nelements; i++) 
      printf("    Volume %s = %g\n", PDBelements[i].Name(), volumes[i]);
    fflush(stdout);
  }

  // Return volumes
  return volumes;
}



static int 
WriteDescriptor(float *values, int nvalues, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(filename, "wb");
  if (!fp) {
    fprintf(stderr, "Unable to open file %s\n", filename);
    return 0;
  }

  // Write number of values
  if (fwrite(&nvalues, sizeof(int), 1, fp) != 1) {
    fprintf(stderr, "Unable to write number of values to %s\n", filename);
    return 0;
  }

  // Write values
  if (fwrite(values, sizeof(float), nvalues, fp) != (unsigned int) nvalues) {
    fprintf(stderr, "Unable to write values to %s\n", filename);
    return 0;
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", sizeof(int) + nvalues * sizeof(float));
    fflush(stdout);
  }

  // Return number of values written
  return nvalues;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: pdb2vol pdbfile volfile [-resolution x y z] [-elements] [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-count")) { 
        count_only = 1; 
      }
      else if (!strcmp(*argv, "-elements")) { 
        nelements = 6;
      }
      else if (!strcmp(*argv, "-normalize")) { 
        argc--; argv++; // ignore this argument
      }
      else if (!strcmp(*argv, "-radius")) { 
        argc--; argv++; // ignore this argument
      }
      else if (!strcmp(*argv, "-sigma")) { 
        argc--; argv++; // ignore this argument
      }
      else if (!strcmp(*argv, "-rasterization")) { 
        argc--; argv++; rasterization_type = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; grid_resolution[0] = atoi(*argv); 
        argc--; argv++; grid_resolution[1] = atoi(*argv); 
        argc--; argv++; grid_resolution[2] = atoi(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else if (!vol_name) vol_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check grid filename
  if (!vol_name) {
    fprintf(stderr, "You did not specify a vol file.\n");
    return 0;
  }

  // Check grid resolution
  if ((grid_resolution[0] <= 0) || (grid_resolution[1] <= 0) || (grid_resolution[2] <= 0)) {
    fprintf(stderr, "Invalid grid resolution: %d %d %d\n", grid_resolution[0], grid_resolution[1], grid_resolution[2]);
    return 0;
  }

  // Check number of elements
  if ((nelements <= 0) || (nelements > 6)) {
    fprintf(stderr, "Invalid number of elements: %d\n", nelements);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read PDB file
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Create descriptor
  float *volumes;
  if (count_only) volumes = CreateCounts(file, nelements);
  else volumes = CreateVolumes(file, nelements);
  if (!volumes) exit(-1);

  // Write descriptor
  int status = WriteDescriptor(volumes, nelements, vol_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}
