// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *pdb_name = NULL;
static char *grid_name = NULL;
static char *mask_name = NULL;
static char *ligand_name = NULL;
static RNScalar sigma = 0;
static RNScalar min_cavity_radius = 0;
static RNScalar min_cavity_volume = 0;
static RNScalar max_total_volume = 0;
static RNScalar max_total_volume_as_fraction_of_ligand = 0;
static RNScalar min_protein_offset = -RN_INFINITY;
static RNScalar max_protein_offset = RN_INFINITY;
static RNScalar ideal_protein_offset = 1.5;
static RNScalar interior_protein_sigma = 0;
static RNScalar exterior_protein_sigma = 0;
static RNScalar min_ligand_offset = -RN_INFINITY;
static RNScalar max_ligand_offset = RN_INFINITY;
static RNScalar ideal_ligand_offset = 1.5;
static RNScalar interior_ligand_sigma = 0;
static RNScalar exterior_ligand_sigma = 0;
static RNBoolean largest_cavity_only = FALSE;
static RNBoolean print_verbose = FALSE;



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static R3Grid *
ReadGrid(char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    RNFail("Unable to allocated grid");
    return NULL;
  }

  // Read grid 
  if (!grid->ReadFile(grid_name)) {
    RNFail("Unable to read grid file %s", grid_name);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    const R3Box bbox = grid->WorldBox();
    printf("  World Box = ( %g %g %g ) ( %g %g %g )\n", bbox[0][0], bbox[0][1], bbox[0][2], bbox[1][0], bbox[1][1], bbox[1][2]);
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    fflush(stdout);
  }

  // Return success
  return grid;
}



static RNScalar 
LigandVolume(PDBFile *file, PDBResidue *ligand)
{
  // Check ligand
  if (!ligand) return 0;

  // Create grid with ligand atoms rasterized
  R3Grid ligand_grid(32, 32, 32);
  ligand_grid.SetWorldToGridTransformation(ligand->BBox());
  for (int i = 0; i < ligand->NAtoms(); i++) {
    PDBAtom *atom = ligand->Atom(i);
    ligand_grid.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
  }

  // Return volume
  return ligand_grid.Volume();
}



static PDBResidue *
FindLigand(PDBFile *file, const char *ligand_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Find ligand
  PDBResidue *ligand = file->FindResidue(ligand_name);
  if (!ligand) {
    fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, pdb_name); 
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Found ligand ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Volume = %g\n", LigandVolume(file, ligand));
    printf("  # Atoms = %d\n", ligand->NAtoms());
    fflush(stdout);
  }

  // Return ligand
  return ligand;
}



static void
Blur(R3Grid *grid, RNScalar sigma)
{
  // Check sigma
  if (sigma == 0) return;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Blur the grid
  grid->Blur(sigma * grid->WorldToGridScaleFactor());

  // Print statistics
  if (print_verbose) {
    printf("Blurred grid .. \n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Sigma = %g\n", sigma);
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }
}



static void
MaskToOffset(R3Grid *mask, PDBFile *file, PDBResidue *ligand,
  RNScalar min_offset, RNScalar max_offset, 
  RNScalar ideal_offset, RNScalar interior_sigma, RNScalar exterior_sigma)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get model from PDB file
  if (file->NModels() == 0) return;
  PDBModel *model = file->Model(0);

  // Get atoms from ligand/model
  RNArray<PDBAtom *>& atoms = (ligand) ? ligand->atoms : model->atoms;

  // Create offset mask
  R3Grid offset_mask(mask->XResolution(), mask->YResolution(), mask->ZResolution());
  offset_mask.SetWorldToGridTransformation(mask->WorldToGridTransformation());
  for (int i = 0; i < atoms.NEntries(); i++) {
    PDBAtom *atom = atoms.Kth(i);
    if (!ligand && atom->IsHetAtom()) continue;
    R3Point grid_position = mask->GridPosition(atom->Position());
    int x = (int) (grid_position[0] + 0.5);
    if ((x < 0) || (x >= mask->XResolution())) continue;
    int y = (int) (grid_position[1] + 0.5);
    if ((y < 0) || (y >= mask->YResolution())) continue;
    int z = (int) (grid_position[2] + 0.5);
    if ((z < 0) || (z >= mask->ZResolution())) continue;
    offset_mask.SetGridValue(x, y, z, i + 0.5);
  }

  // Compute mask as Gaussian function of distance to closest atom minus offset
  offset_mask.Voronoi();
  for (int k = 0; k < offset_mask.ZResolution(); k++) {
    for (int j = 0; j < offset_mask.YResolution(); j++) {
      for (int i = 0; i < offset_mask.XResolution(); i++) {
        R3Point world_position = offset_mask.WorldPosition(i, j, k);
        int closest_atom_id = (int) offset_mask.GridValue(i, j, k);
        PDBAtom *closest_atom = atoms.Kth(closest_atom_id);
        RNLength closest_atom_distance = R3Distance(world_position, closest_atom->Position()) - closest_atom->Radius();
        if ((min_offset > -RN_INFINITY) && (closest_atom_distance < min_offset)) {
          // Too close
          offset_mask.SetGridValue(i, j, k, 0.0);
        }
        else if ((max_offset < RN_INFINITY) && (closest_atom_distance > max_offset)) {
          // Too far
          offset_mask.SetGridValue(i, j, k, 0.0);
        }
        else {
          // Within range - apply gaussian falloff from ideal offset
          RNLength distance = closest_atom_distance - ideal_offset;
          RNLength distance_squared = distance * distance;
          if ((distance < 0) && (interior_sigma > 0)) {
            // Use interior sigma
            RNScalar denom = -2.0 * interior_sigma * interior_sigma;
            RNScalar value = exp( distance_squared / denom );
            offset_mask.SetGridValue(i, j, k, value);
          }
          else if ((distance >= 0) && (exterior_sigma > 0)) {
            // Use exterior sigma
            RNScalar denom = -2.0 * exterior_sigma * exterior_sigma;
            RNScalar value = exp( distance_squared / denom );
            offset_mask.SetGridValue(i, j, k, value);
          }
          else {
            // Apply step function 
            offset_mask.SetGridValue(i, j, k, 1.0);
          }
        }
      }
    }
  }

  // Apply mask
  mask->Multiply(offset_mask);

  // Print statistics
  if (print_verbose) {
    printf("Masked to offset from %s .. \n", (ligand) ? "ligand" : "protein");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Min offset = %g\n", min_offset);
    printf("  Max offset = %g\n", max_offset);
    printf("  Ideal offset = %g\n", ideal_offset);
    printf("  Interior sigma = %g\n", interior_sigma);
    printf("  Exterior sigma = %g\n", exterior_sigma);
    printf("  Cardinality = %d\n", mask->Cardinality());
    printf("  Volume = %g\n", mask->Volume());
    RNInterval grid_range = mask->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", mask->L1Norm());
    printf("  L2Norm = %g\n", mask->L2Norm());
    fflush(stdout);
  }
}



static void
MaskToCavityRadius(R3Grid *mask, RNScalar min_cavity_radius)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Create mask
  R3Grid breadth_mask(*mask);

  // Erode and then dilate
  breadth_mask.Erode(min_cavity_radius);  
  breadth_mask.Dilate(min_cavity_radius);  

  // Apply mask
  mask->Mask(breadth_mask);

  // Print statistics
  if (print_verbose) {
    printf("Masked to cavity radius .. \n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Min cavity radius = %g\n", min_cavity_radius);
    printf("  Cardinality = %d\n", mask->Cardinality());
    printf("  Volume = %g\n", mask->Volume());
    RNInterval grid_range = mask->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", mask->L1Norm());
    printf("  L2Norm = %g\n", mask->L2Norm());
    fflush(stdout);
  }
}



static void
MaskToCavityVolume(R3Grid *mask, RNScalar min_cavity_volume)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Create mask
  R3Grid cavity_mask(*mask);

  // Keep only cavities large enough to hold a certain volume
  RNScalar world_grid_cell_volume = cavity_mask.GridToWorldScaleFactor();
  world_grid_cell_volume = world_grid_cell_volume * world_grid_cell_volume * world_grid_cell_volume;
  int min_cavity_size = (int) (min_cavity_volume / world_grid_cell_volume);
  int max_cavities = cavity_mask.NEntries();
  int *cavity_size = new int [ max_cavities ];
  int *cavity_grid = new int [ cavity_mask.NEntries() ];
  int ncavities = cavity_mask.ConnectedComponents(1.0E-20, max_cavities, NULL, cavity_size, cavity_grid);
  RNScalar *mask_valuesp = (RNScalar *) cavity_mask.GridValues();
  for (int i = 0; i < cavity_mask.NEntries(); i++) {
    int cavity = cavity_grid[i];
    assert(cavity < ncavities);
    if (cavity < 0) { *mask_valuesp = 0; }
    else if (cavity_size[cavity] < min_cavity_size) { *mask_valuesp = 0; }
    else *mask_valuesp = 1;
    mask_valuesp++;
  }
  delete [] cavity_size;
  delete [] cavity_grid;

  // Apply mask
  mask->Mask(cavity_mask);

  // Print statistics
  if (print_verbose) {
    printf("Masked to cavity volume .. \n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Min cavity volume = %g\n", min_cavity_volume);
    printf("  # Cavities = %d\n", ncavities);
    printf("  Cardinality = %d\n", mask->Cardinality());
    printf("  Volume = %g\n", mask->Volume());
    RNInterval grid_range = mask->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", mask->L1Norm());
    printf("  L2Norm = %g\n", mask->L2Norm());
    fflush(stdout);
  }
}



static void 
MaskToLargestCavity(R3Grid *mask)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate arrays to store cavity information
  int *cavity_sizes = new int [ mask->NEntries() ];
  int *cavity_ids = new int [ mask->NEntries() ];
  assert(cavity_sizes && cavity_ids);

  // Find connected cavities 
  int ncavities = mask->ConnectedComponents(1.0E-20, mask->NEntries(), NULL, cavity_sizes, cavity_ids);
  assert(ncavities <= mask->NEntries());

  // Find maximum connected cavity
  int biggest_cavity = -1;
  int biggest_cavity_size = -1;
  for (int j = 0; j < ncavities; j++) {
    if (cavity_sizes[j] > biggest_cavity_size) {
      biggest_cavity_size = cavity_sizes[j];
      biggest_cavity = j;
    }
  }

  // Assign '0' to values not in biggest cavity at threshold
  if (biggest_cavity >= 0) {
    int *cavity_idsp = cavity_ids;
    RNScalar *grid_valuesp = (RNScalar *) mask->GridValues();
    for (int j = 0; j < mask->NEntries(); j++) {
      if (*cavity_idsp != biggest_cavity) *grid_valuesp = 0;
      cavity_idsp++;
      grid_valuesp++;
    }
  }

  // Delete arrays to store connected cavity information
  delete [] cavity_sizes;
  delete [] cavity_ids;

  // Print statistics
  if (print_verbose) {
    printf("Masked to largest cavity only .. \n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Cavities = %d\n", ncavities);
    printf("  Biggest cavity size = %d\n", biggest_cavity_size);
    printf("  Cardinality = %d\n", mask->Cardinality());
    printf("  Volume = %g\n", mask->Volume());
    RNInterval grid_range = mask->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", mask->L1Norm());
    printf("  L2Norm = %g\n", mask->L2Norm());
    fflush(stdout);
  }
}



static void 
MaskToTotalVolume(R3Grid *mask, RNScalar max_total_volume, RNScalar min_cavity_radius, RNScalar min_cavity_volume, RNBoolean largest_cavity_only)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Determine ranges
  RNInterval threshold_range = mask->Range();
  RNInterval volume_range(0, mask->Volume()); 

  // Check ranges
  if (threshold_range.IsEmpty()) return;
  if (volume_range.IsEmpty()) return;
  if (max_total_volume <= volume_range.Min()) return;
  if (max_total_volume >= volume_range.Max()) return;

  // Create copy of mask so that can modify/test iteratively
  R3Grid mask_copy(*mask);

  // Temporarily turn off print_verbose (because call other functions from here)
  RNBoolean saved_print_verbose = print_verbose;
  print_verbose = FALSE;

  // Binary search for threshold that provides target volume
  for (int i = 0; i < 100; i++) {
    // Compute new guess for threshold
    // RNScalar t = (max_total_volume - volume_range.Min()) / (volume_range.Max() - volume_range.Min());
    // RNScalar threshold = threshold_range.Max() - t * (threshold_range.Max() - threshold_range.Min());
    RNScalar threshold = threshold_range.Mid();

    // Make fresh copy of mask
    const RNScalar *mask_valuesp = mask->GridValues();
    RNScalar *mask_copy_valuesp = (RNScalar *) mask_copy.GridValues();
    for (int j = 0; j < mask->NEntries(); j++) {
      *mask_copy_valuesp++ = *mask_valuesp++;
    }

    // Threshold copy of mask
    mask_copy.Threshold(threshold, 0, R3_GRID_KEEP_VALUE);

    // Mask by cavity radius
    if (min_cavity_radius > 0) {
      MaskToCavityRadius(&mask_copy, min_cavity_radius);
    }

    // Mask by cavity volume
    if (min_cavity_volume > 0) {
      MaskToCavityVolume(&mask_copy, min_cavity_volume);
    }

    // Mask to largest cavity only
    if (largest_cavity_only) {
      MaskToLargestCavity(&mask_copy);
    }

    // Determine volume of non-zero entries
    RNScalar volume = mask_copy.Volume();

    // Adjust ranges according to volume
    if (RNIsGreater(volume, max_total_volume, 1)) {
      threshold_range.SetMin(threshold);
      volume_range.SetMax(volume);
    }
    else if (RNIsLess(volume, max_total_volume, 1)) {
      threshold_range.SetMax(threshold);
      volume_range.SetMin(volume);
    }
    else {
      // Found target volume!
      break;
    }
  }

  // Restore previous print_verbose status
  print_verbose = saved_print_verbose;

  // Copy thresholded mask
  *mask = mask_copy;      

  // Print statistics
  if (print_verbose) {
    printf("Masked to total volume .. \n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Max total volume = %g\n", max_total_volume);
    printf("  Minimum cavity volume = %g\n", min_cavity_volume);
    printf("  Largest cavity only = %d\n", largest_cavity_only);
    printf("  Grid treshold = %g\n", threshold_range.Mid());
    printf("  Cardinality = %d\n", mask->Cardinality());
    printf("  Volume = %g\n", mask->Volume());
    RNInterval grid_range = mask->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", mask->L1Norm());
    printf("  L2Norm = %g\n", mask->L2Norm());
    fflush(stdout);
  }
}



static R3Grid * 
CreateMask(PDBFile *file, R3Grid *grid, PDBResidue *ligand)
{
  // Allocate mask
  R3Grid *mask = new R3Grid(*grid);
  if (!mask) {
    fprintf(stderr, "Unable to allocate mask\n");
    exit(-1);
  }

  // Blur 
  if (sigma > 0) {
    Blur(mask, sigma);
  }

  // Mask to offset from protein
  if ((min_protein_offset > -RN_INFINITY) || (max_protein_offset < RN_INFINITY) || ((interior_protein_sigma > 0) && (exterior_protein_sigma > 0))) {
    MaskToOffset(mask, file, NULL, min_protein_offset, max_protein_offset, ideal_protein_offset, interior_protein_sigma, exterior_protein_sigma);
  }

  // Mask to offset from ligand
  if (ligand && (min_ligand_offset > -RN_INFINITY) || (max_ligand_offset < RN_INFINITY) || ((interior_ligand_sigma > 0) && (exterior_ligand_sigma > 0))) {
    MaskToOffset(mask, file, ligand, min_ligand_offset, max_ligand_offset, ideal_ligand_offset, interior_ligand_sigma, exterior_ligand_sigma);
  }

  // Determine max_total_volume from ligand
  if ((max_total_volume_as_fraction_of_ligand > 0) && (max_total_volume_as_fraction_of_ligand < RN_INFINITY)) {
    max_total_volume = max_total_volume_as_fraction_of_ligand * LigandVolume(file, ligand);
  }

  // MaskToTotalVolume executes MaskToCavityRadius, MaskToCavityVolume, and MaskToLargestCavity
  if ((max_total_volume > 0) && (max_total_volume < RN_INFINITY)) {
    // Mask by total volume
    MaskToTotalVolume(mask, max_total_volume, min_cavity_radius, min_cavity_volume, largest_cavity_only);
  }
  else {
    // Mask by cavity radius
    if (min_cavity_radius > 0) {
      MaskToCavityRadius(mask, min_cavity_radius);
    }

    // Mask by cavity volume
    if (min_cavity_volume > 0) {
      MaskToCavityVolume(mask, min_cavity_volume);
    }

    // Mask to largest cavity only
    if (largest_cavity_only) {
      MaskToLargestCavity(mask);
    }
  }

  // Return mask
  return mask;
}



static int 
WriteMask(R3Grid *mask, const char *mask_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write mask
  int status = mask->WriteFile(mask_name);

  // Print statistics
  if (print_verbose) {
    printf("Wrote mask ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", mask->XResolution(), mask->YResolution(), mask->ZResolution());
    printf("  Spacing = %g\n", mask->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", mask->Cardinality());
    printf("  Volume = %g\n", mask->Volume());
    RNInterval grid_range = mask->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", mask->L1Norm());
    printf("  L2Norm = %g\n", mask->L2Norm());
    fflush(stdout);
  }

  // Return status
  return status;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 4) {
    printf("Usage: pdb2msk pdbfile gridfile maskfile [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { print_verbose = TRUE; }
      else if (!strcmp(*argv, "-largest_cavity_only")) { largest_cavity_only = TRUE; }
      else if (!strcmp(*argv, "-ligand")) { argc--; argv++; ligand_name = *argv; }
      else if (!strcmp(*argv, "-protein_interior")) { if (max_protein_offset > 0) max_protein_offset = 0; }
      else if (!strcmp(*argv, "-protein_exterior")) { if (min_protein_offset < 0) min_protein_offset = 0; }
      else if (!strcmp(*argv, "-ligand_interior")) { if (max_ligand_offset > 0) max_ligand_offset = 0; }
      else if (!strcmp(*argv, "-ligand_exterior")) { if (min_ligand_offset < 0) min_ligand_offset = 0; }
      else if (!strcmp(*argv, "-sigma")) { argc--; argv++; sigma = atof(*argv); }
      else if (!strcmp(*argv, "-min_protein_offset")) { argc--; argv++; min_protein_offset = atof(*argv); }
      else if (!strcmp(*argv, "-max_protein_offset")) { argc--; argv++; max_protein_offset = atof(*argv); }
      else if (!strcmp(*argv, "-min_ligand_offset")) { argc--; argv++; min_ligand_offset = atof(*argv); }
      else if (!strcmp(*argv, "-max_ligand_offset")) { argc--; argv++; max_ligand_offset = atof(*argv); }
      else if (!strcmp(*argv, "-protein_offset")) { 
        argc--; argv++; min_protein_offset = atof(*argv); 
        argc--; argv++; ideal_protein_offset = atof(*argv); 
        argc--; argv++; max_protein_offset = atof(*argv); 
        argc--; argv++; interior_protein_sigma = atof(*argv); 
        argc--; argv++; exterior_protein_sigma = atof(*argv); 
      }
      else if (!strcmp(*argv, "-ligand_offset")) { 
        argc--; argv++; min_ligand_offset = atof(*argv); 
        argc--; argv++; ideal_ligand_offset = atof(*argv); 
        argc--; argv++; max_ligand_offset = atof(*argv); 
        argc--; argv++; interior_ligand_sigma = atof(*argv); 
        argc--; argv++; exterior_ligand_sigma = atof(*argv); 
      }
      else if (!strcmp(*argv, "-min_cavity_radius")) { 
        argc--; argv++; min_cavity_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-min_cavity_volume")) { 
        argc--; argv++; min_cavity_volume = atof(*argv); 
      }
      else if (!strcmp(*argv, "-max_total_volume")) { 
        argc--; argv++; max_total_volume = atof(*argv); 
      }
      else if (!strcmp(*argv, "-max_total_volume_as_fraction_of_ligand")) { 
        argc--; argv++; max_total_volume_as_fraction_of_ligand = atof(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else if (!grid_name) grid_name = *argv;
      else if (!mask_name) mask_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check mask filename
  if (!mask_name) {
    fprintf(stderr, "You did not specify a mask file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read PDB file
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Read grid file
  R3Grid *grid = ReadGrid(grid_name);
  if (!grid) exit(-1);

  // Find ligand
  PDBResidue *ligand = NULL;
  if (ligand_name) {
    ligand = FindLigand(file, ligand_name);
    if (!ligand) exit(-1);
  }

  // Create mask
  R3Grid *mask = CreateMask(file, grid, ligand);
  if (!mask) exit(-1);

  // Write mask
  int status = WriteMask(mask, mask_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}

