// Source file for the pdb viewer program



// Include files 

#include "R3Graphics/R3Graphics.h"
#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"
#include "KDTree.h"
#include <GL/glut.h>
#include "GL/glui.h"



// Type definitions

enum {
  NULL_COMMAND,
  QUIT_COMMAND
};

typedef enum {
  NULL_DISPLAY_SCHEME,
  POINT_DISPLAY_SCHEME,
  BALL_DISPLAY_SCHEME,
  // STICK_DISPLAY_SCHEME,
  NUM_DISPLAY_SCHEMES
} DisplayScheme;

typedef enum {
  FILE_COLOR_SCHEME,
  MODEL_COLOR_SCHEME,
  CHAIN_COLOR_SCHEME,
  RESIDUE_COLOR_SCHEME,
  ATOM_COLOR_SCHEME,
  NUM_COLOR_SCHEMES
} ColorScheme;

typedef enum {
  NO_SURFNET_SCHEME,
  THRESHOLD_SURFNET_SCHEME,
  SPHERE_OVERLAP_SURFNET_SCHEME,
  COMPONENT_SIZE_SURFNET_SCHEME,
  BIGGEST_ONLY_SURFNET_SCHEME,
  NUM_SURFNET_SCHEMES
} SurfnetScheme;

typedef enum {
  RANDOM_GRID,
  POINT_GRID,
  INTERIOR_GRID,
  EXTERIOR_GRID,
  DISTANCE_GRID,
  INPUT_GRID,
  SURFNET_GRID,
  SYMMETRY_GRID,
  PROXIMITY_GRID,
  CENTERLINE_GRID,
  GAUSS_GRID,
  CHARGE_GRID,
  CONSERVATION_GRID,
  HYDROPHOBICITY_GRID,
  COMBINED_GRID,
  NUM_GRIDS
} GridType;

struct GridProperties {
  GridProperties(char *name, const RNRgb& color, float threshold, float exponent, RNBoolean show)
  : name(name), color(color), threshold(threshold), exponent(exponent), show(show) {};
  char *name;
  RNRgb color;
  float threshold;
  float exponent;
  RNBoolean show;
};

static GridProperties grid_properties[NUM_GRIDS] = {
  GridProperties("Random        ", RNRgb(0.5, 0.5, 1),  0.5, 1, 0),
  GridProperties("Point         ", RNRgb(0.5, 1, 1),    0.5, 1, 0),
  GridProperties("Interior      ", RNRgb(0, 0.5, 1),    0.5, 1, 0),
  GridProperties("Exterior      ", RNRgb(1, 1, 0),      0.5, 1, 0),
  GridProperties("Distance      ", RNRgb(1, 0.5, 1),    1.8, 1, 0),
  GridProperties("Input         ", RNRgb(0, 0, 1),      0.5, 1, 0),
  GridProperties("Surfnet       ", RNRgb(0, 1, 0),      0.5, 1, 0),
  GridProperties("Symmetry      ", RNRgb(0, 1, 1),      0.5, 1, 0),
  GridProperties("Proximity     ", RNRgb(0.5, 0.0, 1),  0.5, 1, 0),
  GridProperties("Centerline    ", RNRgb(0.5, 0.5, 0),  0.5, 1, 0),
  GridProperties("Gauss         ", RNRgb(1, 0, 1),      0.5, 1, 0),
  GridProperties("Charge        ", RNRgb(1, 0.5, 0),    0.5, 1, 0),
  GridProperties("Conservation  ", RNRgb(0, 1, 1),      0.5, 1, 0),
  GridProperties("Hydrophobicity", RNRgb(1, 1, 0),      0.5, 1, 0),
  GridProperties("Combination   ", RNRgb(1, 0, 0),      0.100001, 1, 1)
};



// Program variables

static char *pdb_name = NULL;
static char *input_grid_name = NULL;
static char *ligand_name = NULL;
static char *surfnet_name = NULL;
static char *asa_name = NULL;
static char *consurf_name = NULL;
static char *hssp_name = NULL;
static char *grow_name = NULL;
static char *dart_name = NULL;
static char *output_grid_name = NULL;
static char *image_name = NULL;

static RNBoolean rndom = FALSE;
static RNBoolean exterior = FALSE;
static RNBoolean symmetry = FALSE;
static RNBoolean proximity = FALSE;
static RNBoolean centerline = FALSE;
static RNBoolean charge = FALSE;
static RNBoolean hydrophobicity = FALSE;
static RNBoolean gauss = FALSE;
static int max_resolution[3] = { 256, 256, 256 };
static RNScalar grid_spacing = 0.5; // angstroms
static RNScalar surfnet_sigma = 0;
static RNScalar gauss_sigma = 5;
static RNScalar conservation_sigma = 0;
static RNScalar charge_sigma = 0;
static RNScalar hydrophobicity_sigma = 0;
static RNBoolean ignore_zeros = FALSE;
static RNBoolean power_of_two = FALSE;
static int surfnet_scheme = NO_SURFNET_SCHEME;
static RNScalar initial_grid_threshold = PDB_UNKNOWN;
static int coordinate_system_type = 1; // sidechain
static RNScalar world_radius = PDB_UNKNOWN;
static R3Point world_origin(PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN);
static RNScalar threshold_coverage = 0;
static RNScalar threshold_volume = 0;
static RNBoolean largest_component_only = 0;

static RNBoolean biomolecule = FALSE;
static RNScalar background_color[3] = { 1, 1, 1 };
static R3Point initial_camera_origin = R3Point(0.0, 0.0, 0.0);
static R3Vector initial_camera_towards = R3Vector(0.0, 0.0, -1.0);
static R3Vector initial_camera_up = R3Vector(0.0, 1.0, 0.0);
static RNBoolean initial_camera = FALSE;
static int print_verbose = 0;
static int print_debug = 0;
static int batch = 0;



// GLUI variables 

static GLUI *glui = NULL;
static int GLUIwindow = 0;
static int GLUIwindow_height = 1024;
static int GLUIwindow_width = 1024;
static int GLUImouse[2] = { 0, 0 };
static int GLUIbutton[3] = { 0, 0, 0 };
static int GLUImodifiers = 0;



// Application variables

static PDBFile *file = NULL;
static PDBFile *surfnet_file = NULL;
static PDBResidue *ligand = NULL;
static R3Grid *grids[NUM_GRIDS] = { NULL };
static R3Viewer *viewer = NULL;
static const int max_darts = 100;
static RNScalar dart_values[max_darts];
static R3Point dart_positions[max_darts];
static RNScalar dart_residue_distances[max_darts];
static RNScalar dart_atom_distances[max_darts];
static RNBoolean dart_hits[max_darts];
static int ndarts = 0;
const RNScalar DONT_CARE = -56789;



// Display variables

static int show_names = 0;
static int show_ligand_coverage = 0;
static int protein_display_scheme = BALL_DISPLAY_SCHEME;
static int hetatom_display_scheme = BALL_DISPLAY_SCHEME;
static int ligand_display_scheme = BALL_DISPLAY_SCHEME;
static int protein_color_scheme = RESIDUE_COLOR_SCHEME;
static int hetatom_color_scheme = RESIDUE_COLOR_SCHEME;
static int ligand_color_scheme = RESIDUE_COLOR_SCHEME;
static int show_bonds = 0;
static int show_backbone_only = 0;
static int show_contacts_only = 0;
static int show_occupancy = 0;
static int show_charge = 0;
static int show_conservation = 0;
static int show_hydrophobicity = 0;
static int show_accessible_surface_area = 0;
static int show_surfnet = 0;
static int show_darts = 0;
static int show_atom_frames = 0;
static int show_residue_frames = 0;
static int show_grid_points = 0;
static int show_grid_text = 0;
static int show_grid_isosurface = 0;
static int show_grid_slices[3] = { 0, 0, 0 };
static int show_grid_box = 0;
static int grid_slice_coords[3] = { 0, 0, 0 };
static int split_image = 0;



////////////////////////////////////////////////////////////////////////
// USER INTERFACE STUFF
////////////////////////////////////////////////////////////////////////

static R3Viewer *
CreateViewer(PDBFile *file, PDBResidue *ligand)
{
  // Get file bounding box
  R3Box bbox = file->BBox();
  assert(!bbox.IsEmpty());
  RNLength r = world_radius;
  if (r == PDB_UNKNOWN) {
    r = bbox.DiagonalRadius();
    assert((r > 0.0) && RNIsFinite(r));
  }

  // Setup default camera view looking down the Z axis
  if (!initial_camera) initial_camera_origin = bbox.Centroid() - initial_camera_towards * (2 * r);;
  R3Camera camera(initial_camera_origin, initial_camera_towards, initial_camera_up, 0.4, 0.4, 0.1 * r, 1000.0 * r);
  R2Viewport viewport(0, 0, GLUIwindow_width, GLUIwindow_height);
  R3Viewer *viewer = new R3Viewer(camera, viewport);
  if (world_origin[0] == PDB_UNKNOWN) world_origin = file->BBox().Centroid();

  // Zoom in on ligand, if there is one
  if (!initial_camera && ligand) {
    // Reset camera
    R3Box ligand_bbox = ligand->BBox();
    R3Point ligand_center = ligand_bbox.Centroid();
    RNLength ligand_radius = world_radius;
    if (ligand_radius == PDB_UNKNOWN) {
      ligand_radius = ligand_bbox.DiagonalRadius();
      assert((ligand_radius > 0.0) && RNIsFinite(ligand_radius));
    }
#if 1
    R3Vector towards = initial_camera_towards;
    R3Vector up = initial_camera_up;
#else
    R3Point file_center = file->BBox().Centroid();
    R3Vector towards = file_center - ligand_center;
    RNLength towards_length = towards.Length();
    if (RNIsZero(towards_length)) towards = R3negz_vector;
    else towards /= towards_length;
    R3Vector up = R3posy_vector;
    RNScalar dot = up.Dot(towards);
    if (RNIsEqual(fabs(dot), 1.0)) up = R3posx_vector;
#endif
    R3Point eye = ligand_center - towards * (3 * ligand_radius);;
    viewer->ResetCamera(eye, towards, up);
    if (world_origin[0] == PDB_UNKNOWN) world_origin = ligand->BBox().Centroid();
  }

  // Return viewer
  return viewer;
}




void GLUIStop(void)
{
  // Destroy window 
  glutDestroyWindow(GLUIwindow);

  // Exit
  exit(0);
}



void GLUIRedraw(void)
{
  // Set viewing transformation
  viewer->Camera().Load();

  // Clear window 
  glClearColor(background_color[0], background_color[1], background_color[2], 1.0);
  if (!image_name || !split_image) glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Set lights
  static GLfloat light0_position[] = { 3.0, 4.0, 5.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
  static GLfloat light1_position[] = { -3.0, -2.0, -3.0, 0.0 };
  glLightfv(GL_LIGHT1, GL_POSITION, light1_position);

  // Draw PDB file
  if (file) {
    // Show protein
    if (protein_display_scheme != NULL_DISPLAY_SCHEME) {
      glEnable(GL_LIGHTING);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      RNRgb color(0.5, 0.25, 0.5);
      for (int m = 0; m < file->NModels(); m++) {
        PDBModel *model = file->Model(m);
        if (protein_color_scheme == MODEL_COLOR_SCHEME) color = model->Color();
        for (int i = 0; i < model->NChains(); i++) {
          PDBChain *chain = model->Chain(i);
          if (protein_color_scheme == CHAIN_COLOR_SCHEME) color = chain->Color();
          for (int j = 0; j < chain->NResidues(); j++) {
            PDBResidue *residue = chain->Residue(j);
            if (residue == ligand) continue;
            if (protein_color_scheme == RESIDUE_COLOR_SCHEME) color = residue->Color();
            // if (show_contacts_only && ligand && (!residue->IsBonded(ligand))) continue;
            for (int k = 0; k < residue->NAtoms(); k++) {
              PDBAtom *atom = residue->Atom(k);
              if (show_backbone_only && !atom->IsBackbone()) continue;
              if (atom->IsHetAtom()) continue;
              if (show_contacts_only && ligand && (!atom->IsBonded(ligand))) continue;
              if (show_charge) {
                if (atom->Charge() == PDB_UNKNOWN) color = RNblue_rgb;
                else { RNScalar c = 0.5 + atom->Charge(); color[0] = c; color[1] = c; color[2] = c; }
              }
              else if (show_conservation) {
                if (atom->conservation == PDB_UNKNOWN) color = RNblue_rgb;
                else { color[0] = 1 - atom->conservation; color[1] = 1 - atom->conservation; color[2] = 1 - atom->conservation; }
              }
              else if (show_hydrophobicity) {
                if (atom->IsHetAtom()) color = RNblue_rgb;
                else { RNScalar h = 1 - (atom->Hydrophobicity() + 1)/3; color[0] = h; color[1] = h; color[2] = h; }
              }
              else if (show_accessible_surface_area) {
                if (atom->accessible_surface_area == PDB_UNKNOWN) color = RNblue_rgb;
                else { RNScalar asa = atom->accessible_surface_area / 50; color[0] = asa; color[1] = asa; color[2] = asa; }
              }
              else if (protein_color_scheme == ATOM_COLOR_SCHEME) color = atom->Color();
              if (show_occupancy && (atom->Occupancy() < 1.0)) glEnable(GL_BLEND);
              static GLfloat material[4];
              material[0] = color[0]; material[1] = color[1]; material[2] = color[2]; material[3] = atom->Occupancy();
              glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
              RNLength radius = (protein_display_scheme == BALL_DISPLAY_SCHEME) ? atom->Radius() : 0.25;
              R3Sphere(atom->Position(), radius).Draw();
              if (show_occupancy && (atom->Occupancy() < 1.0)) glDisable(GL_BLEND);
            }
          }
        }
      }
    }

    // Show hetatoms
    if (hetatom_display_scheme != NULL_DISPLAY_SCHEME) {
      glEnable(GL_LIGHTING);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      RNRgb color(0.5, 0.25, 0.5);
      for (int m = 0; m < file->NModels(); m++) {
        PDBModel *model = file->Model(m);
        if (hetatom_color_scheme == MODEL_COLOR_SCHEME) color = model->Color();
        for (int i = 0; i < model->NChains(); i++) {
          PDBChain *chain = model->Chain(i);
          if (hetatom_color_scheme == CHAIN_COLOR_SCHEME) color = chain->Color();
          for (int j = 0; j < chain->NResidues(); j++) {
            PDBResidue *residue = chain->Residue(j);
            if (residue == ligand) continue;
            if (hetatom_color_scheme == RESIDUE_COLOR_SCHEME) color = residue->Color();
            if (show_contacts_only && ligand && (!residue->IsBonded(ligand))) continue;
            for (int k = 0; k < residue->NAtoms(); k++) {
              PDBAtom *atom = residue->Atom(k);
              if (!atom->IsHetAtom()) continue;
              if (show_charge) {
                if (atom->Charge() == PDB_UNKNOWN) color = RNblue_rgb;
                else { RNScalar c = 0.5 + atom->Charge(); color[0] = c; color[1] = c; color[2] = c; }
              }
              else if (show_conservation) {
                if (atom->conservation == PDB_UNKNOWN) color = RNblue_rgb;
                else { color[0] = 1 - atom->conservation; color[1] = 1 - atom->conservation; color[2] = 1 - atom->conservation; }
              }
              else if (show_hydrophobicity) {
                if (atom->IsHetAtom()) color = RNblue_rgb;
                else { RNScalar h = 1 - (atom->Hydrophobicity() + 1)/3; color[0] = h; color[1] = h; color[2] = h; }
              }
              else if (show_accessible_surface_area) {
                if (atom->accessible_surface_area == PDB_UNKNOWN) color = RNblue_rgb;
                else { RNScalar asa = atom->accessible_surface_area / 50; color[0] = asa; color[1] = asa; color[2] = asa; }
              }
              else if (hetatom_color_scheme == ATOM_COLOR_SCHEME) color = atom->Color();
              if (show_occupancy && (atom->Occupancy() < 1.0)) glEnable(GL_BLEND);
              static GLfloat material[4];
              material[0] = color[0]; material[1] = color[1]; material[2] = color[2]; material[3] = atom->Occupancy();
              glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
              RNLength radius = (hetatom_display_scheme == BALL_DISPLAY_SCHEME) ? atom->Radius() : 0.25;
              R3Sphere(atom->Position(), radius).Draw();
              if (show_occupancy && (atom->Occupancy() < 1.0)) glDisable(GL_BLEND);
            }
          }
        }
      }
    }

    // Show ligand
    if (ligand && (ligand_display_scheme != NULL_DISPLAY_SCHEME)) {
      // Show atoms
      if (ligand_display_scheme != NULL_DISPLAY_SCHEME) {
        int count = 0;
        PDBChain *chain = ligand->Chain();
        glEnable(GL_LIGHTING);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        static RNRgb default_color(0, 1, 0);
        RNRgb color = default_color;
        if (ligand_color_scheme == CHAIN_COLOR_SCHEME) color = chain->Color();
        else if (ligand_color_scheme == RESIDUE_COLOR_SCHEME) color = ligand->Color();
        for (int i = 0; i < ligand->NAtoms(); i++) {
          PDBAtom *atom = ligand->Atom(i);
          if (show_occupancy && (atom->Occupancy() < 1.0)) glEnable(GL_BLEND);
          if (show_ligand_coverage) {
            RNScalar value = grids[COMBINED_GRID]->WorldValue(atom->Position());
            if (value >= grid_properties[COMBINED_GRID].threshold) { color = 0.5 * default_color; count++; }
            else color = default_color;
          }
          else if (ligand_color_scheme == ATOM_COLOR_SCHEME) color = atom->Color();
          static GLfloat material[4];
          material[0] = color[0]; material[1] = color[1]; material[2] = color[2]; material[3] = atom->Occupancy();
          glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
          RNLength radius = (ligand_display_scheme == BALL_DISPLAY_SCHEME) ? atom->Radius() : 0.5;
          R3Sphere(atom->Position(), radius).Draw();
          if (show_occupancy && (atom->Occupancy() < 1.0)) glDisable(GL_BLEND);
        }

        if (show_ligand_coverage) {
          glDisable(GL_LIGHTING);
          char buffer[64];
          sprintf(buffer, "%d%%", 100 * count / ligand->NAtoms());
          R3Point p = viewer->WorldRay(GLUIwindow_width - 100, GLUIwindow_height - 100).Point(2 * viewer->Camera().Near());
          glColor3f(1 - background_color[0], 1 - background_color[1], 1 - background_color[2]);
          R3DrawText(p, buffer);
        }
      }
    }    

    // Show atom frames
    if (show_atom_frames && (file->NModels() > 0)) {
      // Create KD tree of all protein atoms
      PDBModel *model = file->Model(0);
      static KDTree<PDBAtom *> *kdtree = NULL;
      PDBAtom *tmp = file->Model(0)->Atom(0);
      int position_offset = (unsigned char *) &(tmp->position) - (unsigned char *) tmp;
      if (!kdtree) kdtree = new KDTree<PDBAtom *>(model->atoms, position_offset);
      if (!kdtree) { fprintf(stderr, "Unable to create KD tree for %s\n", file->Name()); return; }

      // Draw coordinate frame for all atoms
      glDisable(GL_LIGHTING);
      glLineWidth(7);
      RNRgb color(0.5, 0.25, 0.5);
      for (int m = 0; m < file->NModels(); m++) {
        PDBModel *model = file->Model(m);
        if (protein_color_scheme == MODEL_COLOR_SCHEME) color = model->Color();
        for (int i = 0; i < model->NChains(); i++) {
          PDBChain *chain = model->Chain(i);
          if (protein_color_scheme == CHAIN_COLOR_SCHEME) color = chain->Color();
          for (int j = 0; j < chain->NResidues(); j++) {
            PDBResidue *residue = chain->Residue(j);
            if (show_contacts_only && ligand && (!residue->IsBonded(ligand))) continue;
            if (protein_color_scheme == RESIDUE_COLOR_SCHEME) color = residue->Color();
            for (int k = 0; k < residue->NAtoms(); k++) {
              PDBAtom *atom1 = residue->Atom(k);
              if (show_backbone_only && !atom1->IsBackbone()) continue;
              if (atom1->IsHetAtom()) continue;

              // Find all nearby atoms 
              const R3Point& p1 = atom1->Position();
              const RNScalar nearby_radius = 3.5;
              RNArray<PDBAtom *> nearby_atoms;
              kdtree->FindAll(atom1, RN_EPSILON, 2 * nearby_radius, nearby_atoms);
              if (nearby_atoms.IsEmpty()) continue;

              // Compute first vector of atom coordinate system 
              R3Vector vector1 = R3zero_vector;
              for (int a = 0; a < nearby_atoms.NEntries(); a++) {
                PDBAtom *atom2 = nearby_atoms[a];
                if (atom2->IsHetAtom()) continue;
                if (!atom2->AminoAcid()) continue;
                R3Point p2 = atom2->Position();
                R3Vector v = p2 - p1;
                RNLength d = v.Length();
                if (RNIsZero(d)) continue;
                vector1 += v / d;
              }

              // Compute second vector of atom coordinate system 
              if (vector1.IsZero()) continue;
              R3Vector vector2 = R3zero_vector;
              R3Plane plane1(atom1->Position(), vector1);
              for (int a = 0; a < nearby_atoms.NEntries(); a++) {
                PDBAtom *atom2 = nearby_atoms[a];
                if (atom2->IsHetAtom()) continue;
                if (!atom2->AminoAcid()) continue;
                R3Point p2 = atom2->Position();
                p2.Project(plane1);
                R3Vector v = p2 - p1;
                RNLength d = v.Length();
                if (RNIsZero(d)) continue;
                vector2 += v / d;
              }

              // Build coordinate system
              if (vector2.IsZero()) continue;
              R3Triad triad(vector1, vector2);
              R3CoordSystem cs(p1, triad);

              // Draw coordinate system
              RNLoadRgb(color);
              R3BeginLine();
              R3LoadPoint(cs.Origin());
              R3LoadPoint(cs.Origin() + 4 * cs.Axes().Axis(0));
              R3EndLine();
              RNLoadRgb(0.6 * color);
              R3BeginLine();
              R3LoadPoint(cs.Origin());
              R3LoadPoint(cs.Origin() + 4 * cs.Axes().Axis(1));
              R3EndLine();
              RNLoadRgb(0.2 * color);
              R3BeginLine();
              R3LoadPoint(cs.Origin());
              R3LoadPoint(cs.Origin() + 4 * cs.Axes().Axis(2));
              R3EndLine();
            }
          }
        }
      }
      glLineWidth(1);
    }

    // Show residue frames
    if (show_residue_frames) {
      glDisable(GL_LIGHTING);
      glLineWidth(7);
      RNRgb color(0.5, 0.25, 0.5);
      for (int m = 0; m < file->NModels(); m++) {
        PDBModel *model = file->Model(m);
        if (protein_color_scheme == MODEL_COLOR_SCHEME) color = model->Color();
        for (int i = 0; i < model->NChains(); i++) {
          PDBChain *chain = model->Chain(i);
          if (protein_color_scheme == CHAIN_COLOR_SCHEME) color = chain->Color();
          for (int j = 0; j < chain->NResidues(); j++) {
            PDBResidue *residue = chain->Residue(j);
            PDBAminoAcid *aminoacid = residue->AminoAcid();
            if (!aminoacid) continue;
            if (residue->HasHetAtoms()) continue;
            if (show_contacts_only && ligand && (!residue->IsBonded(ligand))) continue;
            if (protein_color_scheme == RESIDUE_COLOR_SCHEME) color = residue->Color();

            // Construct residue coordinate system 
            R3CoordSystem cs;
            if (!residue->ConstructCoordinateSystem(cs, coordinate_system_type)) {
              fprintf(stderr, "Unable to construct coordinate system for residue %s %d\n", residue->Name(), residue->Sequence());
              continue;
            }

            // Draw coordinate system
            RNLoadRgb(color);
            R3BeginLine();
            R3LoadPoint(cs.Origin());
            R3LoadPoint(cs.Origin() + 4 * cs.Axes().Axis(0));
            R3EndLine();
            // RNLoadRgb(0.6 * color);
            R3BeginLine();
            R3LoadPoint(cs.Origin());
            R3LoadPoint(cs.Origin() + 4 * cs.Axes().Axis(1));
            R3EndLine();
            // RNLoadRgb(0.2 * color);
            R3BeginLine();
            R3LoadPoint(cs.Origin());
            R3LoadPoint(cs.Origin() + 4 * cs.Axes().Axis(2));
            R3EndLine();
          }
        }
      }
      glLineWidth(1);
    }

    // Show names
    if (show_names) {
      glDisable(GL_LIGHTING);
      for (int m = 0; m < file->NModels(); m++) {
        PDBModel *model = file->Model(m);
        if ((protein_display_scheme != NULL_DISPLAY_SCHEME) && (protein_color_scheme == MODEL_COLOR_SCHEME)) {
          RNLoadRgb(model->Color());
          R3DrawText(model->BBox().Centroid(), model->Name());
        }
        for (int i = 0; i < model->NChains(); i++) {
          PDBChain *chain = model->Chain(i);
          if ((protein_display_scheme != NULL_DISPLAY_SCHEME) && (protein_color_scheme == CHAIN_COLOR_SCHEME)) {
            RNLoadRgb(chain->Color());
            R3DrawText(chain->BBox().Centroid(), chain->Name());
          }
          for (int j = 0; j < chain->NResidues(); j++) {
            PDBResidue *residue = chain->Residue(j);
            if (show_contacts_only && ligand && (!residue->IsBonded(ligand))) continue;
            if (((residue == ligand) && (ligand_display_scheme != NULL_DISPLAY_SCHEME) && (ligand_color_scheme == RESIDUE_COLOR_SCHEME)) ||
                ((!residue->AminoAcid()) && (hetatom_display_scheme != NULL_DISPLAY_SCHEME) && (hetatom_color_scheme == RESIDUE_COLOR_SCHEME)) ||
                ((residue->AminoAcid()) && ((protein_display_scheme != NULL_DISPLAY_SCHEME) || show_residue_frames) && (protein_color_scheme == RESIDUE_COLOR_SCHEME))) {
              RNLoadRgb(residue->Color());
              R3DrawText(residue->BBox().Centroid(), residue->Name());
            }
            for (int k = 0; k < residue->NAtoms(); k++) {
              PDBAtom *atom = residue->Atom(k);
              if (((residue == ligand) && (ligand_display_scheme != NULL_DISPLAY_SCHEME) && (ligand_color_scheme == ATOM_COLOR_SCHEME)) ||
                  ((!residue->AminoAcid()) && (hetatom_display_scheme != NULL_DISPLAY_SCHEME) && (hetatom_color_scheme == ATOM_COLOR_SCHEME)) ||
                  ((residue->AminoAcid()) && (protein_display_scheme != NULL_DISPLAY_SCHEME) && (protein_color_scheme == ATOM_COLOR_SCHEME))) {
                RNLoadRgb(atom->Color());
                R3DrawText(atom->Position(), atom->Name());
              }
            }
          }
        }
      }
    }
    
    // Show bonds
    if (show_bonds) {
      glDisable(GL_LIGHTING);
      glLineWidth(3);
      const RNArray<PDBBond *>& bonds = (ligand) ? ligand->bonds : file->Model(0)->bonds;
      for (int i = 0; i < bonds.NEntries(); i++) {
        PDBBond *bond = bonds[i];
        RNLoadRgb(bond->Color());
        R3BeginLine();
        R3LoadPoint(bond->Atom(0)->Position());
        R3LoadPoint(bond->Atom(1)->Position());
        R3EndLine();
      }
      glLineWidth(1);
    }

    // Show sites
    if (surfnet_file && show_surfnet) {
      glEnable(GL_LIGHTING);
      static GLfloat material[] = { 0, 0.5, 0, 1 };
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
      for (int i = 0; i < surfnet_file->NModels(); i++) {
        PDBModel *surfnet_model = surfnet_file->Model(i);
        for (int j = 0; j < surfnet_model->NAtoms(); j++) {
          PDBAtom *atom = surfnet_model->Atom(j);
          R3Sphere(atom->Position(), atom->Occupancy()).Draw();
        }
      }
    }

    // Show darts
    if (show_darts) {
      glDisable(GL_LIGHTING);
      glLineWidth(5);
      int n = (ndarts > 10) ? 10 : ndarts;
      for (int i = 0; i < n; i++) {
        // Set color to indicate hit/miss
        if (dart_hits[i]) glColor3f(0, 1, 0);
        else glColor3f(1, 0, 1);

        // Draw a little cross at dart position
        R3BeginLine();
        R3LoadPoint(dart_positions[i] - R3posx_vector);
        R3LoadPoint(dart_positions[i] + R3posx_vector);
        R3EndLine();
        R3BeginLine();
        R3LoadPoint(dart_positions[i] - R3posy_vector);
        R3LoadPoint(dart_positions[i] + R3posy_vector);
        R3EndLine();
        R3BeginLine();
        R3LoadPoint(dart_positions[i] - R3posz_vector);
        R3LoadPoint(dart_positions[i] + R3posz_vector);
        R3EndLine();

        // Draw text with dart number
        char buffer[64];
        sprintf(buffer, "%d", i+1);
        R3DrawText(dart_positions[i] + 0.5 * R3ones_vector, buffer);
      }
      glLineWidth(1);
    }
  }

  // Draw grids
  for (int g = 0; g < NUM_GRIDS; g++) {
    R3Grid *grid = grids[g];
    if (!grid) continue;
    if (!grid_properties[g].show) continue;

    // Push grid transformation
    const R3Affine& grid_to_world = grid->GridToWorldTransformation();
    grid_to_world.Push();

    // Draw grid
    if (show_grid_points) {
      glPointSize(5.0);
      glDisable(GL_LIGHTING);
      glDepthMask(FALSE);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glEnable(GL_BLEND);
      glBegin(GL_POINTS);
      RNScalar maximum = grid->Maximum();
      for (int i = 0; i < grid->XResolution(); i++) {
        for (int j = 0; j < grid->YResolution(); j++) {
          for (int k = 0; k < grid->ZResolution(); k++) {
            float value = grid->GridValue(i, j, k);
            if (value <= grid_properties[g].threshold) continue;
            RNLoadRgb((maximum - value) * grid_properties[g].color / maximum);
            glVertex3i(i, j, k);
          }
        }
      }
      glEnd();
      glDisable(GL_BLEND);
      glBlendFunc(GL_ONE, GL_ZERO);
      glDepthMask(TRUE);
      glPointSize(1.0);
    }

    // Draw grid text
    if (show_grid_text) {
      char buffer[64];
      glDisable(GL_LIGHTING);
      RNLoadRgb(grid_properties[g].color);
      for (int i = 0; i < grid->XResolution(); i++) {
        for (int j = 0; j < grid->YResolution(); j++) {
          for (int k = 0; k < grid->ZResolution(); k++) {
            float value = grid->GridValue(i, j, k);
            if (value <= grid_properties[g].threshold) continue;
            sprintf(buffer, "%.2g", value);
            R3DrawText((float) i, (float) j, (float) k, buffer);
          }
        }
      }
    }

    // Draw grid isosurface
    if (show_grid_isosurface) {
      glDisable(GL_LIGHTING);
      RNLoadRgb(grid_properties[g].color);
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      grid->DrawIsoSurface(grid_properties[g].threshold);
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    // Draw grid slices
    if (show_grid_slices[RN_X] || show_grid_slices[RN_Y] || show_grid_slices[RN_Z]) {
      glDisable(GL_LIGHTING);
      RNLoadRgb(grid_properties[g].color);
      if (show_grid_slices[RN_X]) grid->DrawSlice(RN_X, grid_slice_coords[RN_X]);
      if (show_grid_slices[RN_Y]) grid->DrawSlice(RN_Y, grid_slice_coords[RN_Y]);
      if (show_grid_slices[RN_Z]) grid->DrawSlice(RN_Z, grid_slice_coords[RN_Z]);
    }

    // Draw grid bounding box
    if (show_grid_box) {
      glDisable(GL_LIGHTING);
      RNLoadRgb(grid_properties[g].color);
      R3Box(0, 0, 0, grid->XResolution()-1, grid->YResolution()-1, grid->ZResolution()-1).Outline();
#if 0
      if (show_grid_slices[0]) R3Box(grid_slice_coords[0], 0, 0, grid_slice_coords[0], grid->YResolution()-1, grid->ZResolution()-1).Outline();
      if (show_grid_slices[1]) R3Box(0, grid_slice_coords[1], 0, grid->XResolution()-1, grid_slice_coords[1], grid->ZResolution()-1).Outline();
      if (show_grid_slices[2]) R3Box(0, 0, grid_slice_coords[2], grid->XResolution()-1, grid->YResolution()-1, grid_slice_coords[2]).Outline();
#endif
    }

    // Pop grid transformation
    grid_to_world.Pop();
  }

  // Capture image
  if (image_name) {
    static int first = 1;
    if (!split_image) {
      R2Image image(GLUIwindow_width, GLUIwindow_height, 3);
      image.Capture();
      image.Write(image_name);
      image_name = NULL;
    }
    else if (first) {
      // Save everything
      first = 0;
      int saved_protein_display_scheme = protein_display_scheme;
      int saved_initial_camera = initial_camera;
      R3Viewer *saved_viewer = viewer;
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

      // Draw images with initial viewer
      // viewer = new R3Viewer(*viewer);
      protein_display_scheme = BALL_DISPLAY_SCHEME;
      viewer->ResizeViewport(10, 10, GLUIwindow_width/2 - 20, GLUIwindow_height/2 - 20);
      glViewport(10, 10, GLUIwindow_width/2 - 20, GLUIwindow_height/2 - 20);
      GLUIRedraw();
      glutSwapBuffers();
      protein_display_scheme = NULL_DISPLAY_SCHEME;
      viewer->ResizeViewport(GLUIwindow_width/2 + 10, 10, GLUIwindow_width/2 - 20, GLUIwindow_height/2 - 20);
      glViewport(GLUIwindow_width/2 + 10, 10, GLUIwindow_width/2 - 20, GLUIwindow_height/2 - 20);
      GLUIRedraw();
      glutSwapBuffers();

      // Draw images zoomed in
      initial_camera = FALSE;
      viewer = CreateViewer(file, ligand);
      protein_display_scheme = BALL_DISPLAY_SCHEME;
      viewer->ResizeViewport(10, GLUIwindow_height/2 + 10, GLUIwindow_width/2 - 20, GLUIwindow_height/2 - 20);
      glViewport(10, GLUIwindow_height/2 + 10, GLUIwindow_width/2 - 20, GLUIwindow_height/2 - 20);
      GLUIRedraw();
      glutSwapBuffers();
      protein_display_scheme = NULL_DISPLAY_SCHEME;
      viewer->ResizeViewport(GLUIwindow_width/2 + 10, GLUIwindow_height/2 + 10, GLUIwindow_width/2 - 20, GLUIwindow_height/2 - 20);
      glViewport(GLUIwindow_width/2 + 10, GLUIwindow_height/2 + 10, GLUIwindow_width/2 - 20, GLUIwindow_height/2 - 20);
      GLUIRedraw();
      glutSwapBuffers();

      // Capture image
      glViewport(0, 0, GLUIwindow_width, GLUIwindow_height);
      R2Image image(GLUIwindow_width, GLUIwindow_height, 3);
      image.Capture();
      image.Write(image_name);

      // Reset evertyhing
      protein_display_scheme = saved_protein_display_scheme;
      initial_camera = saved_initial_camera;
      viewer = saved_viewer;
      image_name = NULL;
    }
  }

  // Stop
  if (batch && !image_name) GLUIStop();

  // Swap buffers 
  glutSwapBuffers();
}    



PDBAtom *
GLUIPickAtom(PDBFile *file, int x, int y)
{
  // Check file
  if (!file) return FALSE;

  // Check position
  if ((x < 0) || (GLUIwindow_width <= x) || (y < 0) || (GLUIwindow_height <= y)) { 
    printf("Pick outside viewport\n"); 
    return FALSE;
  }

  // Get model
  if (file->NModels() == 0) return FALSE;
  PDBModel *model = file->Model(0);

  // Allocate select buffer
  const int SELECT_BUFFER_SIZE = 1024;
  GLuint select_buffer[SELECT_BUFFER_SIZE];
  GLint select_buffer_hits;

  // Initialize select buffer
  glSelectBuffer(SELECT_BUFFER_SIZE, select_buffer);
  glRenderMode(GL_SELECT);
  glInitNames();
  glPushName(0);

  // Draw atoms with pick names into selection buffer
  GLint viewport[4];
  glViewport(0, 0, GLUIwindow_width, GLUIwindow_height);
  glGetIntegerv(GL_VIEWPORT, viewport);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPickMatrix((GLdouble) x, (GLdouble) y, 1, 1, viewport);
  viewer->Camera().Load(TRUE);
  glMatrixMode(GL_MODELVIEW);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    PDBResidue *residue = atom->Residue();
    RNLength radius = atom->Radius();
    if (show_contacts_only && ligand && (!residue->IsBonded(ligand))) continue;
    if (show_backbone_only && !atom->IsBackbone()) continue;
    if (residue == ligand) {
      if (ligand_display_scheme == NULL_DISPLAY_SCHEME) continue;
      else if (ligand_display_scheme == POINT_DISPLAY_SCHEME) radius = 0.25;
    }
    else if (atom->IsHetAtom()) {
      if (hetatom_display_scheme == NULL_DISPLAY_SCHEME) continue;
      else if (hetatom_display_scheme == POINT_DISPLAY_SCHEME) radius = 0.25;
    }
    else {
      if (protein_display_scheme == NULL_DISPLAY_SCHEME) continue;
      else if (protein_display_scheme == POINT_DISPLAY_SCHEME) radius = 0.25;
    }
    glLoadName(i + 1);
    R3Sphere(atom->Position(), radius).Draw();
  }
  glFlush();
  select_buffer_hits = glRenderMode(GL_RENDER);

  // Process select buffer to find front-most hit
  GLuint hit = 0;
  GLuint hit_z = 0xFFFFFFFF;
  GLuint *bufp = select_buffer;
  GLuint numnames, z1, z2;
  for (int i = 0; i < select_buffer_hits; i++) {
    numnames = *bufp++;
    z1 = *bufp++;
    z2 = *bufp++;
    while (numnames--) {
      if (z1 < hit_z) {
        hit = *bufp;
        hit_z = z1;
      }
      bufp++;
    }
  }

  // Return closest atom
  if (hit > 0) {
    // We hit something -- parse hit
    int atom_index = hit - 1;
    if ((atom_index < 0) || (atom_index >= model->NAtoms())) {
      fprintf(stderr, "Invalid hit index in GLUIPickAtom: %d\n", atom_index);
      return FALSE;
    }

    // Get selected atom
    return model->Atom(atom_index);
  }

  // No atom was selected
  return NULL;
}



void GLUIResize(int a, int b)
{
  // Resize window
  int x, y, w, h;
  GLUI_Master.get_viewport_area( &x, &y, &w, &h );
  glViewport( x, y, w, h );

  // Resize viewer viewport
  viewer->ResizeViewport(x, y, w, h);

  // Remember window size 
  GLUIwindow_width = w;
  GLUIwindow_height = h;

  // Redraw
  glutPostRedisplay();
}



void GLUIMotion(int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Compute mouse movement
  int dx = x - GLUImouse[0];
  int dy = y - GLUImouse[1];
  
  // World in hand navigation 
  if (GLUIbutton[0]) viewer->RotateWorld(1.0, world_origin, x, y, dx, dy);
  else if (GLUIbutton[1]) viewer->ScaleWorld(1.0, world_origin, x, y, dx, dy);
  else if (GLUIbutton[2]) viewer->TranslateWorld(1.0, world_origin, x, y, dx, dy);
  if (GLUIbutton[0] || GLUIbutton[1] || GLUIbutton[2]) glutPostRedisplay();

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;
}



void GLUIMouse(int button, int state, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Process button event
  if ((button == GLUT_LEFT_BUTTON) && (state == GLUT_DOWN)) {
    // Check for double click
    static RNTime click_time;
    const RNScalar max_double_click_elapsed = 0.5;
    RNBoolean double_click = (click_time.Elapsed() < max_double_click_elapsed);
    click_time.Read();

    // Pick atom
    if (double_click && file && file->NModels()) {
      PDBAtom *selected_atom = GLUIPickAtom(file, x, y);
      if (selected_atom) {
        // Get ancestors in hierarchy
        PDBResidue *selected_residue = selected_atom->Residue();
        PDBChain *selected_chain = selected_residue->Chain();
        PDBModel *selected_model = selected_chain->Model();

        // Print message
        printf("%s-%s-%s-%s-%d-%c %s %d\n", 
               file->Name(), 
               selected_model->Name(), 
               (strcmp(selected_chain->Name(), " ")) ? selected_chain->Name() : "_",
               selected_residue->Name(), selected_residue->Sequence(), 
               (selected_residue->InsertionCode() != ' ') ? selected_residue->InsertionCode() : '_',
               selected_atom->Name(), selected_atom->Serial());
      }
    }
  }

  // Remember button state 
  int b = (button == GLUT_LEFT_BUTTON) ? 0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
  GLUIbutton[b] = (state == GLUT_DOWN) ? 1 : 0;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Redraw
  glutPostRedisplay();
}



void GLUISpecial(int key, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Set useful variables
  static RNBoolean first = TRUE;
  static RNScalar threshold_step = 1;
  static RNInterval threshold_range(0,0);
  if (first) {
    RNInterval grid_range = grids[COMBINED_GRID]->Range();
    threshold_step = grids[COMBINED_GRID]->StandardDeviation() / 10;
    threshold_range.Reset(grid_range.Min() + 1.0E-20, grid_range.Max() - 1.0E-20);
    first = FALSE;
  }

  // Process keyboard button event 
  switch (key) {
  case GLUT_KEY_DOWN:
    grid_properties[COMBINED_GRID].threshold -= threshold_step;
    if (grid_properties[COMBINED_GRID].threshold < threshold_range.Min()) grid_properties[COMBINED_GRID].threshold = threshold_range.Min();
    break;

  case GLUT_KEY_UP:
    grid_properties[COMBINED_GRID].threshold += threshold_step;
    if (grid_properties[COMBINED_GRID].threshold > threshold_range.Max()) grid_properties[COMBINED_GRID].threshold = threshold_range.Max();
    break;

  case GLUT_KEY_LEFT:
    threshold_step /= 2;
    break;

  case GLUT_KEY_RIGHT:
    threshold_step *= 2;
    break;
  }

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Synchronize GLUI interface
  GLUI_Master.sync_live_all();

  // Redraw
  glutPostRedisplay();
}



void GLUIKeyboard(unsigned char key, int x, int y)
{
  // Process keyboard button event 
  switch (key) {
  case '~': {
    glReadBuffer(GL_FRONT);
    R2Image image(GLUIwindow_width, GLUIwindow_height, 3);
    image.Capture();
    char imagename[256];
    static int image_count = 1;
    sprintf(imagename, "i%d.jpg", image_count++);
    image.Write(imagename);
    glReadBuffer(GL_BACK);
    break; }

  case '^':
    show_charge = !show_charge;
    break;

  case '&':
    show_conservation = !show_conservation;
    break;

  case '*':
    show_hydrophobicity = !show_hydrophobicity;
    break;

  case '!':
    show_accessible_surface_area = !show_accessible_surface_area;
    break;

  case '@':
    show_ligand_coverage = !show_ligand_coverage;
    break;

  case '#':
    coordinate_system_type = 1 - coordinate_system_type;
    break;

  case '$':
    show_backbone_only = !show_backbone_only;
    break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    grid_properties[key - '0'].show = !grid_properties[key - '0'].show;
    break;

  case 'B':
    show_bonds = !show_bonds;
    break;

  case 'b':
    show_contacts_only = !show_contacts_only;
    break;

  case 'D':
  case 'd':
    show_darts = !show_darts;
    break;

  case 'E':
  case 'e':
    show_grid_box = !show_grid_box;
    break;

  case 'F':
    show_atom_frames = !show_atom_frames;
    break;

  case 'f':
    show_residue_frames = !show_residue_frames;
    break;

  case 'G':
  case 'g':
    show_grid_points = !show_grid_points;
    break;

  case 'H':
    hetatom_color_scheme = (hetatom_color_scheme + 1) % NUM_COLOR_SCHEMES;
    break;

  case 'h':
    hetatom_display_scheme = (hetatom_display_scheme + 1) % NUM_DISPLAY_SCHEMES;
    break;

  case 'I':
  case 'i':
    show_grid_isosurface = !show_grid_isosurface;
    break;

  case 'L':
    ligand_color_scheme = (ligand_color_scheme + 1) % NUM_COLOR_SCHEMES;
    break;

  case 'l':
    ligand_display_scheme = (ligand_display_scheme + 1) % NUM_DISPLAY_SCHEMES;
    break;

  case 'N':
  case 'n':
    show_names = !show_names;
    break;

  case 'O':
  case 'o':
    show_occupancy = !show_occupancy;
    break;

  case 'P':
    protein_color_scheme = (protein_color_scheme + 1) % NUM_COLOR_SCHEMES;
    break;

  case 'p':
    protein_display_scheme = (protein_display_scheme + 1) % NUM_DISPLAY_SCHEMES;
    break;

  case 'S':
  case 's':
    show_surfnet = !show_surfnet;
    break;

  case 'T':
  case 't':
    show_grid_text = !show_grid_text;
    break;

  case 'W':
  case 'w':
    show_grid_slices[0] = !show_grid_slices[0];
    show_grid_slices[1] = show_grid_slices[0];
    show_grid_slices[2] = show_grid_slices[0];
    break;

  case 'X':
  case 'x':
    show_grid_slices[0] = !show_grid_slices[0];
    break;

  case 'Y':
  case 'y':
    show_grid_slices[1] = !show_grid_slices[1];
    break;

  case 'Z':
  case 'z':
    show_grid_slices[2] = !show_grid_slices[2];
    break;

  case ' ': {
    // Print camera
    const R3Camera& camera = viewer->Camera();
    printf("#camera  %g %g %g  %g %g %g  %g %g %g  %g \n",
           camera.Origin().X(), camera.Origin().Y(), camera.Origin().Z(),
           camera.Towards().X(), camera.Towards().Y(), camera.Towards().Z(),
           camera.Up().X(), camera.Up().Y(), camera.Up().Z(),
           camera.YFOV());
    break; }
      
  case 27: // ESCAPE
    GLUIStop();
    break;
  }

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = GLUIwindow_height - y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Synchronize GLUI interface
  GLUI_Master.sync_live_all();

  // Redraw
  glutPostRedisplay();  
}




void GLUICallback(int control)
{
  switch (control) {
  case QUIT_COMMAND:
    GLUIStop();
    break;
  }
}



#if 0

void GLUIIdle(void)
{
  // Set current window
  if ( glutGetWindow() != GLUIwindow ) 
    glutSetWindow(GLUIwindow);  

  // Redraw
  glutPostRedisplay();
}

#endif



void GLUISetup(void)
{
  GLUI_Panel *panel;
  GLUI_Spinner *spinner;

  // Set the main window's title
  glutSetWindowTitle(pdb_name);

  // Initialize GLUI
  glui = GLUI_Master.create_glui("MBE Controls", 0, GLUIwindow_width + 120, 100);
  glui->set_main_gfx_window(GLUIwindow);

  // Add protein display variables
  panel = glui->add_panel("Protein Color Scheme");
  GLUI_RadioGroup *group = glui->add_radiogroup_to_panel(panel, &protein_color_scheme );
  glui->add_radiobutton_to_group(group, "File");
  glui->add_radiobutton_to_group(group, "Model");
  glui->add_radiobutton_to_group(group, "Chain");
  glui->add_radiobutton_to_group(group, "Amino acid");
  glui->add_radiobutton_to_group(group, "Element");

  // Add ligand display variables
  if (ligand) {
    panel = glui->add_panel("Ligand Color Scheme");
    GLUI_RadioGroup *group = glui->add_radiogroup_to_panel(panel, &ligand_color_scheme );
    glui->add_radiobutton_to_group(group, "File");
    glui->add_radiobutton_to_group(group, "Model");
    glui->add_radiobutton_to_group(group, "Chain");
    glui->add_radiobutton_to_group(group, "Amino acid");
    glui->add_radiobutton_to_group(group, "Element");
  }

  panel = glui->add_panel("Binding Site Display");
  glui->add_checkbox_to_panel(panel, "Surfnet spheres", &show_surfnet );

  // Add grid selection variables
  panel = glui->add_rollout("Grids");
  for (int i = 0; i < NUM_GRIDS; i++) {
    if (!grids[i]) continue;
    GLUI_Panel *tmp = glui->add_panel_to_panel(panel, "", GLUI_PANEL_NONE);
    glui->add_checkbox_to_panel(tmp, "", &(grid_properties[i].show));
    glui->add_column_to_panel(tmp, FALSE);
    GLUI_Rollout *rollout = glui->add_rollout_to_panel(tmp, grid_properties[i].name, FALSE);
    RNInterval range = grids[i]->Range();
    spinner = glui->add_spinner_to_panel(rollout, "Threshold:", GLUI_SPINNER_FLOAT, &(grid_properties[i].threshold));
    spinner->set_float_limits(range.Min(), range.Max(), GLUI_LIMIT_CLAMP);
    spinner = glui->add_spinner_to_panel(rollout, "Exponent:", GLUI_SPINNER_FLOAT, &(grid_properties[i].exponent));
    spinner->set_float_limits(-10, 10, GLUI_LIMIT_CLAMP);
    spinner->set_speed(10);
  }

  // Add grid display variables
  panel = glui->add_panel("Grid Display");
  glui->add_checkbox_to_panel(panel, "Isosurface", &show_grid_isosurface );
  glui->add_checkbox_to_panel(panel, "Points", &show_grid_points );
  glui->add_checkbox_to_panel(panel, "Labels", &show_grid_text );
  glui->add_checkbox_to_panel(panel, "Bounding box", &show_grid_text );

  // Add slice display variables
  panel = glui->add_panel("Slice Display");
  glui->add_checkbox_to_panel(panel, "X Slice", &show_grid_slices[0] );
  spinner = glui->add_spinner_to_panel(panel, "X Slice", GLUI_SPINNER_INT, &grid_slice_coords[0]);
  spinner->set_int_limits(0, grids[COMBINED_GRID]->XResolution()-1, GLUI_LIMIT_CLAMP);
  glui->add_checkbox_to_panel(panel, "Y Slice", &show_grid_slices[1] );
  spinner = glui->add_spinner_to_panel(panel, "Y Slice", GLUI_SPINNER_INT, &grid_slice_coords[1]);
  spinner->set_int_limits(0, grids[COMBINED_GRID]->YResolution()-1, GLUI_LIMIT_CLAMP);
  glui->add_checkbox_to_panel(panel, "Z Slice", &show_grid_slices[2] );
  spinner = glui->add_spinner_to_panel(panel, "Z Slice", GLUI_SPINNER_INT, &grid_slice_coords[2]);
  spinner->set_int_limits(0, grids[COMBINED_GRID]->ZResolution()-1, GLUI_LIMIT_CLAMP);

  // Add button
  glui->add_button("Quit", QUIT_COMMAND, GLUICallback);
}



void GLUIInit(int *argc, char **argv)
{
  // Open window 
  glutInit(argc, argv);
  glutInitWindowPosition(100, 100);
  glutInitWindowSize(GLUIwindow_width, GLUIwindow_height);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH); // | GLUT_STENCIL
  GLUIwindow = glutCreateWindow("OpenGL Viewer");

  // Initialize background color 
  glClearColor(200.0/255.0, 200.0/255.0, 200.0/255.0, 1.0);

  // Initialize lights
  static GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  static GLfloat light0_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glEnable(GL_LIGHT0);
  static GLfloat light1_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
  glEnable(GL_LIGHT1);
  glEnable(GL_NORMALIZE);
  glEnable(GL_LIGHTING);

  // Initialize graphics modes 
  glEnable(GL_DEPTH_TEST);

  // Initialize GLUT callback functions 
  glutDisplayFunc(GLUIRedraw);
  GLUI_Master.set_glutReshapeFunc(GLUIResize);
  GLUI_Master.set_glutKeyboardFunc(GLUIKeyboard);
  GLUI_Master.set_glutSpecialFunc(GLUISpecial);
  GLUI_Master.set_glutMouseFunc(GLUIMouse);
  glutMotionFunc(GLUIMotion);

  // Initialize font
#if (RN_OS == RN_WINDOWSNT)
  int font = glGenLists(256);
  wglUseFontBitmaps(wglGetCurrentDC(), 0, 256, font); 
  glListBase(font);
#endif
}



void GLUIMainLoop(void)
{
  // Set default values
  if (grid_properties[COMBINED_GRID].threshold == -98765) {
    RNScalar max_threshold = grids[COMBINED_GRID]->Maximum() - 1.0E-20;
    grid_properties[COMBINED_GRID].threshold = grids[COMBINED_GRID]->Mean() + 3 * grids[COMBINED_GRID]->StandardDeviation();
    if (grid_properties[COMBINED_GRID].threshold > max_threshold) grid_properties[COMBINED_GRID].threshold = max_threshold;
  }

  // Setup GLUI controls
  GLUISetup();

  // Run main loop -- never returns 
  glutMainLoop();
}




////////////////////////////////////////////////////////////////////////
// FILE I/O STUFF
////////////////////////////////////////////////////////////////////////


static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) {
    RNFail("Unable to read PDB file: %s", filename);
    return NULL;
  }

  // Get biomolecule
  if (biomolecule && (!file->IsBiomolecule())) {
    PDBFile *file2 = file->CopyBiomolecule();
    delete file;
    file = file2;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static PDBResidue *
FindLigand(PDBFile *file, const char *ligand_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Find ligand
  PDBResidue *ligand = file->FindResidue(ligand_name);
  if (!ligand) {
    fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, pdb_name); 
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Found ligand ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", ligand->NAtoms());
    fflush(stdout);
  }

  // Return ligand
  return ligand;
}



static PDBFile *
ReadSurfnetFile(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return NULL;

  // Print statistics
  if (print_verbose) {
    printf("Read surfnet file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Spheres = %d\n", (file->NModels() > 0) ? file->Model(0)->NAtoms() : 0);
    fflush(stdout);
  }

  // Return success
  return file;
}



static int
ReadConsurfFiles(PDBFile *file, char *hssp_basename)
{
  // Check number of models
  if (file->NModels() < 1) {
    fprintf(stderr, "File must have at least one model to read conservation scores: %s.\n", hssp_basename);
    return 0;
  }

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read conservation files
  int nresidues = file->ReadConsurfFiles(hssp_basename);

  // Print statistics
  if (print_verbose && (nresidues > 0)) {
    printf("Read consurf files ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Residues = %d\n", nresidues);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ReadHsspFiles(PDBFile *file, char *hssp_basename)
{
  // Check number of models
  if (file->NModels() < 1) {
    fprintf(stderr, "File must have at least one model to read conservation scores: %s.\n", hssp_basename);
    return 0;
  }

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read conservation files
  int nresidues = file->ReadHsspFiles(hssp_basename);

  // Print statistics
  if (print_verbose && (nresidues > 0)) {
    printf("Read hssp files ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf(" # Residues = %d\n", nresidues);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ReadASAFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read ASA file
  int natoms = file->ReadASAFile(filename);
  if (natoms == 0) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read accessible surface area file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Surface Atoms = %d\n", natoms);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ReadGrowFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read grow file
  int nbonds = file->ReadGrowFile(filename);
  if (nbonds == 0) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read grow file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bonds = %d\n", nbonds);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ReadDartFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open dart file
  FILE *fp = fopen(filename, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open dart file: %s\n", filename);
    return 0;
  }

  // Read header line
  int hit_count, first_hit;
  if (fscanf(fp, "%d%d%d", &ndarts, &hit_count, &first_hit) != 3) {
    fprintf(stderr, "Unable to read dart file: %s\n", filename);
    return 0;
  }

  // Read darts
  for (int i = 0; i < ndarts; i++) {
    char buffer[64];
    int index, dummy, dart_atom_id, dart_residue_id;
    if (fscanf(fp, "%s%d%lf%lf%lf%lf%d%d%lf%lf%d%d", buffer, &index, &dart_values[i], &(dart_positions[i][0]), &(dart_positions[i][1]), &(dart_positions[i][2]),
                    &dart_residue_id, &dart_atom_id, &dart_residue_distances[i], &dart_atom_distances[i], &dart_hits[i], &dummy) != 12) {
      fprintf(stderr, "Error reading dart file %s at line %d\n", filename, i+2);
      return 0;
    }
  }

  // Close dart file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Read dart file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Darts = %d\n", ndarts);
    printf("  # Hits = %d\n", hit_count);
    printf("  First Hit = %d\n", first_hit);
    fflush(stdout);
  }

  // Return number of darts
  return ndarts;
}



static int 
WriteGrid(R3Grid *grid, const char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write grid
  int status = grids[COMBINED_GRID]->WriteFile(grid_name);

  // Print statistics
  if (print_verbose) {
    printf("Wrote grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", status * sizeof(RNScalar));
    fflush(stdout);
  }

  // Return status
  return status;
}



////////////////////////////////////////////////////////////////////////
// GRID CREATION STUFF
////////////////////////////////////////////////////////////////////////

RNScalar 
GridVolume(R3Grid *grid, RNScalar threshold)
{
  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Count number of grid samples at or above threshold
  int grid_count = 0;
  if (!largest_component_only) {
    for (int i = 0; i < grid->NEntries(); i++) 
      if (grid->GridValue(i) >= threshold) grid_count++;
  }
  else {
    const int max_components = 1024;
    int component_sizes[max_components];
    int ncomponents = grid->ConnectedComponents(threshold, max_components, NULL, component_sizes);
    for (int i = 0; i < ncomponents; i++) {
      if (component_sizes[i] > grid_count) grid_count = component_sizes[i];
    }
  }

  // Return volume of grid samples at or above threshold
  return grid_count * volume_scale_factor;
}



RNScalar 
GridThreshold(R3Grid *grid, RNScalar target_volume)
{
  // Determine ranges of thresholds and volumes
  RNInterval threshold_range = grid->Range();
  threshold_range.SetMin(threshold_range.Min() + 1.0E-6);
  threshold_range.SetMax(threshold_range.Max() - 1.0E-6);
  RNScalar min_volume = GridVolume(grid, threshold_range.Max());
  RNScalar max_volume = GridVolume(grid, threshold_range.Min());
  RNInterval volume_range(min_volume, max_volume); 

  // Check if target volume is within range
  if (target_volume >= volume_range.Max()) return threshold_range.Min();
  if (target_volume <= volume_range.Min()) return threshold_range.Max();

  // Binary search for threshold that provides target volume
  for (int i = 0; i < 1000; i++) {
    RNScalar t = (target_volume - volume_range.Min()) / (volume_range.Max() - volume_range.Min());
    RNScalar threshold = threshold_range.Max() - t * (threshold_range.Max() - threshold_range.Min());
    RNScalar volume = GridVolume(grid, threshold);
    if (RNIsGreater(volume, target_volume, 1)) {
      threshold_range.SetMin(threshold);
      volume_range.SetMax(volume);
    }
    else if (RNIsLess(volume, target_volume, 1)) {
      threshold_range.SetMax(threshold);
      volume_range.SetMin(volume);
    }
    else return threshold;
  }

  // Return middle of remaining range
  return threshold_range.Mid();
}



void
BlurGrid(R3Grid *grid, RNLength world_sigma) 
{
  // Build filter
  RNScalar sigma = world_sigma * grid->WorldToGridScaleFactor();
  int filter_radius = (int) (3 * sigma);
  RNScalar *filter = new RNScalar [ filter_radius + 1 ];
  assert(filter);

  // Make buffer for temporary copy of row
  int res = grid->XResolution();
  if (res < grid->YResolution()) res = grid->YResolution();
  if (res < grid->ZResolution()) res = grid->ZResolution();
  RNScalar *buffer = new RNScalar [ res ];
  assert(buffer);

  // Fill filter with Gaussian 
  const RNScalar sqrt_two_pi = sqrt(RN_TWO_PI);
  double a = sqrt_two_pi * sigma;
  double fac = 1.0 / (a * a * a);
  double denom = 2.0 * sigma * sigma;
  for (int i = 0; i <= filter_radius; i++) {
    filter[i] = fac * exp(-i * i / denom);
  }

  // Convolve grid with filter in X direction
  for (int k = 0; k < grid->ZResolution(); k++) {
    for (int j = 0; j < grid->YResolution(); j++) { 
      for (int i = 0; i < grid->XResolution(); i++) 
        buffer[i] = grid->GridValue(i, j, k); 
      for (int i = 0; i < grid->XResolution(); i++) { 
        RNScalar sum = 0.0;
        RNScalar weight = 0;
        RNScalar value = buffer[i];
        if (value != DONT_CARE) {
          sum += filter[0] * value;
          weight += filter[0];
        }
        int nsamples = i;
        if (nsamples > filter_radius) nsamples = filter_radius;
        for (int m = 1; m <= nsamples; m++) {
          RNScalar value = buffer[i - m];
          if (value != DONT_CARE) {
            sum += filter[m] * value;
            weight += filter[m];
          }
        }
        nsamples = grid->XResolution() - 1 - i;
        if (nsamples > filter_radius) nsamples = filter_radius;
        for (int m = 1; m <= nsamples; m++) {
          RNScalar value = buffer[i + m];
          if (value != DONT_CARE) {
            sum += filter[m] * value;
            weight += filter[m];
          }
        }
        if (weight > 0) grid->SetGridValue(i, j, k, sum / weight);
      }
    }
  }

  // Convolve grid with filter in Y direction
  for (int k = 0; k < grid->ZResolution(); k++) {
    for (int j = 0; j < grid->XResolution(); j++) { 
      for (int i = 0; i < grid->YResolution(); i++) 
        buffer[i] = grid->GridValue(j, i, k); 
      for (int i = 0; i < grid->YResolution(); i++) { 
        RNScalar sum = 0.0;
        RNScalar weight = 0;
        RNScalar value = buffer[i];
        if (value != DONT_CARE) {
          sum += filter[0] * value;
          weight += filter[0];
        }
        int nsamples = i;
        if (nsamples > filter_radius) nsamples = filter_radius;
        for (int m = 1; m <= nsamples; m++) {
          RNScalar value = buffer[i - m];
          if (value != DONT_CARE) {
            sum += filter[m] * value;
            weight += filter[m];
          }
        }
        nsamples = grid->YResolution() - 1 - i;
        if (nsamples > filter_radius) nsamples = filter_radius;
        for (int m = 1; m <= nsamples; m++) {
          RNScalar value = buffer[i + m];
          if (value != DONT_CARE) {
            sum += filter[m] * value;
            weight += filter[m];
          }
        }
        if (weight > 0) grid->SetGridValue(j, i, k, sum / weight);
      }
    }
  }

  // Convolve grid with filter in Z direction
  for (int k = 0; k < grid->YResolution(); k++) {
    for (int j = 0; j < grid->XResolution(); j++) { 
      for (int i = 0; i < grid->ZResolution(); i++) 
        buffer[i] = grid->GridValue(j, k, i); 
      for (int i = 0; i < grid->ZResolution(); i++) { 
        RNScalar sum = 0.0;
        RNScalar weight = 0;
        RNScalar value = buffer[i];
        if (value != DONT_CARE) {
          sum += filter[0] * value;
          weight += filter[0];
        }
        int nsamples = i;
        if (nsamples > filter_radius) nsamples = filter_radius;
        for (int m = 1; m <= nsamples; m++) {
          RNScalar value = buffer[i - m];
          if (value != DONT_CARE) {
            sum += filter[m] * value;
            weight += filter[m];
          }
        }
        nsamples = grid->ZResolution() - 1 - i;
        if (nsamples > filter_radius) nsamples = filter_radius;
        for (int m = 1; m <= nsamples; m++) {
          RNScalar value = buffer[i + m];
          if (value != DONT_CARE) {
            sum += filter[m] * value;
            weight += filter[m];
          }
        }
        if (weight > 0) grid->SetGridValue(j, k, i, sum / weight);
      }
    }
  }

  // Deallocate memory
  delete [] filter;
  delete [] buffer;
}



R3Grid *
CreateRandomGrid(PDBModel *model, const R3Box& world_box, int xres, int yres, int zres)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(world_box);

  // Fill grid with random values
  RNSeedRandomScalar();
  for (int i = 0; i < grid->NEntries(); i++) {
    grid->SetGridValue(i, RNRandomScalar());
  }

  // Print statistics
  if (print_verbose) {
    RNInterval grid_range = grid->Range();
    printf("Created random grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreatePointGrid(PDBModel *model, const R3Box& world_box, int xres, int yres, int zres)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(world_box);

  // Rasterize atoms into grid
  int natoms = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (atom->IsHetAtom()) continue;
    if (atom->accessible_surface_area == 0) continue;
    R3Point grid_position = grid->GridPosition(atom->Position());
    int x = (int) (grid_position[0] + 0.5);
    int y = (int) (grid_position[1] + 0.5);
    int z = (int) (grid_position[2] + 0.5);
    grid->SetGridValue(x, y, z, i + 0.5);
    natoms++;
  }

  // Print statistics
  if (print_verbose) {
    printf("Created point grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  # Atoms = %d\n", natoms);
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateInteriorGrid(PDBModel *model, const R3Box& world_box, int xres, int yres, int zres)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(world_box);

  // Rasterize spheres into grid
  int natoms = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (atom->IsHetAtom()) continue;
    grid->RasterizeWorldSphere(atom->Position(), atom->Radius(), 1);
    natoms++;
  }

  // Threshold
  grid->Threshold(0.5, 0, 1);

  // Print statistics
  if (print_verbose) {
    printf("Created interior grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  # Atoms = %d\n", natoms);
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateExteriorGrid(PDBModel *model, R3Grid *interior_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*interior_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Invert 0<->1
  grid->Threshold(0.5, 1, 0);

  // Print statistics
  if (print_verbose) {
    printf("Created exterior grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateDistanceGrid(PDBModel *model, R3Grid *interior_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate distance grid
  R3Grid *grid = new R3Grid(*interior_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Compute squared distance transform
  grid->SquaredDistanceTransform();

  // Convert squared distances in grid units to distances in angstroms
  grid->Sqrt();
  grid->Multiply(grid->GridToWorldScaleFactor());

  // Print statistics
  if (print_verbose) {
    printf("Created distance grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateProximityGrid(PDBModel *model, R3Grid *distance_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*distance_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Invert distance transform
  grid->Threshold(1, 1, R3_GRID_KEEP_VALUE);
  grid->Invert();

  // Normalize the values
  RNInterval range = grid->Range();
  grid->Subtract(range.Min());
  grid->Divide((range.Max() - range.Min()) / 0.9);
  grid->Add(0.1);

  // Print statistics
  if (print_verbose) {
    printf("Created proximity grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateSurfnetGrid(PDBModel *model, PDBModel *surfnet_model, const R3Box& world_box, int xres, int yres, int zres)
{
  // Just checking
  assert(surfnet_model);

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(world_box);

  // Rasterize each surfnet sphere into grid
  int nclefts = 0;
  int nspheres = 0;
  for (int i = 0; i < surfnet_model->NResidues(); i++) {
    PDBResidue *cleft = surfnet_model->Residue(i);
    for (int j = 0; j < cleft->NAtoms(); j++) {
      PDBAtom *atom = cleft->Atom(j);
      grid->RasterizeWorldSphere(atom->Position(), atom->Occupancy(), 1.0);
      nspheres++;
    }
    nclefts++;
  }

  // Print statistics
  if (print_verbose) {
    printf("Created surfnet grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  # Clefts = %d\n", nclefts);
    printf("  # Spheres = %d\n", nspheres);
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateSurfnetGrid(PDBModel *model, R3Grid *distance_grid)
{
  // Just checking
  assert(distance_grid);

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(distance_grid->XResolution(), distance_grid->YResolution(), distance_grid->ZResolution());
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(distance_grid->WorldToGridTransformation());

  // Rasterize sphere at midpoint between every pair of atoms
  int npairs = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    // Get atom1
    PDBAtom *atom1 = model->Atom(i);
    if (atom1->IsHetAtom()) continue;
    if (atom1->accessible_surface_area == 0) continue;
    RNScalar radius1 = atom1->Radius();
    for (int j = i+1; j < model->NAtoms(); j++) {
      // Get atom2
      PDBAtom *atom2 = model->Atom(j);
      if (atom2->IsHetAtom()) continue;
      if (atom2->accessible_surface_area == 0) continue;
      RNScalar radius2 = atom2->Radius();

      // Compute sphere between the atoms
      R3Point center = 0.5 * (atom1->position + atom2->position);
      RNLength span_length = R3Distance(atom1->position, atom2->position) - radius1 - radius2;
      if (span_length  > 10) continue;
      RNScalar distance = distance_grid->WorldValue(center);
      if (distance < 1.5) continue;
      if (distance > 4) distance = 4;
      grid->RasterizeWorldSphere(center, distance, 1.0);
      npairs++;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created surfnet grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  # Pairs = %d\n", npairs);
    fflush(stdout);
  }

  // Return grid
  return grid;
}



void
ApplySurfnetScheme(R3Grid *grid, int scheme)
{
  // Just checking
  if (!grid) return;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Re-weight surfnet grid
  if (scheme == THRESHOLD_SURFNET_SCHEME) {
    // Make all clefts equal weight
    grid->Threshold(0.5, 0, 1);
  }
  else if (scheme == SPHERE_OVERLAP_SURFNET_SCHEME) {
    // Normalize to [0...1]
    grid->Divide(grid->Maximum());
  }
  else if ((scheme == BIGGEST_ONLY_SURFNET_SCHEME) ||
           (scheme == COMPONENT_SIZE_SURFNET_SCHEME)) {
    // Allocate arrays to store connected component information
    int *component_sizes = new int [ grid->NEntries() ];
    int *component_ids = new int [ grid->NEntries() ];
    assert(component_sizes && component_ids);

    // Find connected components 
    int ncomponents = grid->ConnectedComponents(0.5, grid->NEntries(), NULL, component_sizes, component_ids);
    assert(ncomponents <= grid->NEntries());

    // Check number of components
    if (ncomponents > 0) {
      if (scheme == BIGGEST_ONLY_SURFNET_SCHEME) {
        // Find maximum connected component
        int biggest_component = -1;
        int biggest_component_size = -1;
        for (int i = 0; i < ncomponents; i++) {
          if (component_sizes[i] > biggest_component_size) {
            biggest_component_size = component_sizes[i];
            biggest_component = i;
          }
        }

        // Assign '1' to values in biggest component, and '0' everywhere else
        assert(biggest_component >= 0);
        int *component_idsp = component_ids;
        RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
        for (int i = 0; i < grid->NEntries(); i++) {
          if (*component_idsp == biggest_component) *grid_valuesp = 1;
          else *grid_valuesp = 0;
          component_idsp++;
          grid_valuesp++;
        }
      }
      else if (scheme == COMPONENT_SIZE_SURFNET_SCHEME) {
        // Assign 'component size' to all grid entries
        RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
        for (int i = 0; i < grid->NEntries(); i++) {
          int component = component_ids[i];
          assert(((component >= 0) || (*grid_valuesp == 0)) && (component < ncomponents));
          if (component >= 0) *grid_valuesp = component_sizes[component];
          else *grid_valuesp = 0;
          grid_valuesp++;
        }

        // Normalize to [0...1]
        grid->Divide(grid->Maximum());
      }
    }

    // Deallocate memory
    delete [] component_sizes;
    delete [] component_ids;
  }

  // Blur the grid within the non-zero region
  if (surfnet_sigma > 0) {
    R3Grid tmp(*grid);
    BlurGrid(&tmp, surfnet_sigma);
    grid->Multiply(tmp);
  }

  // Print statistics
  if (print_verbose) {
    printf("Weighted surfnet grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  Surfnet scheme = %d\n", scheme);
    fflush(stdout);
  }
}



R3Grid *
CreateSymmetryGrid(PDBModel *model, R3Grid *exterior_grid)
{
  // Just checking
  assert(exterior_grid);

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(exterior_grid->XResolution(), exterior_grid->YResolution(), exterior_grid->ZResolution());
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(exterior_grid->WorldToGridTransformation());

  // Consider span between all pairs of atoms on protein surface
  int npairs = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    // Get atom1
    PDBAtom *atom1 = model->Atom(i);
    if (atom1->IsHetAtom()) continue;
    for (int j = i+1; j < model->NAtoms(); j++) {
      // Get atom2
      PDBAtom *atom2 = model->Atom(j);
      if (atom2->IsHetAtom()) continue;

      // Get span and check its length
      R3Span span(atom1->position, atom2->position);
      RNScalar min_distance = 5;
      RNScalar max_distance = 20;
      RNLength distance = span.Length();
      if (distance >= max_distance) continue;
      if (distance <= min_distance) continue;

      // Compute visibility of atom1 and atom2
      RNBoolean visible = FALSE;
      RNScalar t_step = exterior_grid->GridToWorldScaleFactor();
      RNScalar t_start = atom1->Radius() + t_step;
      RNScalar t_stop = span.Length() - atom2->Radius() - t_step; 
      for (RNScalar t = t_start; t < t_stop; t += t_step) {
        R3Point world_position = span.Point(t);
        R3Point grid_position = exterior_grid->GridPosition(world_position);
        int x = (int) (grid_position[0] + 0.5);
        int y = (int) (grid_position[1] + 0.5);
        int z = (int) (grid_position[2] + 0.5);
        if (exterior_grid->GridValue(x, y, z) == 0) { visible = FALSE; break; }
        else visible = TRUE;
      }

      // Check visibility
      if (!visible) continue;

      // Add splat at span midpoint
      RNScalar weight = 1;
      grid->RasterizeWorldPoint(span.Midpoint(), weight);

      // Increment counter
      npairs++;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created symmetry grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  # Pairs = %d\n", npairs);
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateCenterlineGrid(PDBModel *model, R3Grid *cleft_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*cleft_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Invert (1<->0) and compute distance transform
  grid->Threshold(0.5, 1, 0);
  grid->SquaredDistanceTransform();
  grid->Sqrt();
  grid->Multiply(grid->GridToWorldScaleFactor());

  // Normalize the values
  RNInterval range = grid->Range();
  grid->Divide(range.Max() / 0.9);
  grid->Add(0.1);

  // Print statistics
  if (print_verbose) {
    printf("Created centerline grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateGaussGrid(PDBModel *model, R3Grid *point_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*point_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Threshold and blur with gaussian
  grid->Threshold(1.0E-20, 0, 1);
  BlurGrid(grid, gauss_sigma);

  // Normalize the values
  RNInterval range = grid->Range();
  grid->Subtract(range.Min());
  grid->Divide((range.Max() - range.Min()) / 0.9);
  grid->Add(0.1);

  // Print statistics
  if (print_verbose) {
    printf("Created gauss grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateChargeGrid(PDBModel *model, R3Grid *point_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*point_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Convert atom IDs to charge values
  RNScalar min_charge = 0;
  RNScalar max_charge = 1;
  RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
  for (int i = 0; i < grid->NEntries(); i++) {
    RNScalar atom_id = *grid_valuesp;
    if (atom_id > 0) {
      PDBAtom *atom = model->Atom((int) atom_id);
      RNScalar charge = atom->Charge();
      if (charge == PDB_UNKNOWN) charge = 0.0;
      else charge = (fabs(charge) - min_charge)  / (max_charge - min_charge);
      *grid_valuesp = charge;
    }
    grid_valuesp++;
  }

  // Blur with gaussian
  if (ignore_zeros) grid->Threshold(1.0E-20, DONT_CARE, R3_GRID_KEEP_VALUE);
  BlurGrid(grid, charge_sigma);
  if (ignore_zeros) grid->Threshold(DONT_CARE + 1.0E-20, 0, R3_GRID_KEEP_VALUE);

  // Normalize the values
  RNInterval range = grid->Range();
  grid->Subtract(range.Min());
  grid->Divide((range.Max() - range.Min()) / 0.9);
  grid->Add(0.1);

  // Print statistics
  if (print_verbose) {
    printf("Created charge grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateConservationGrid(PDBModel *model, R3Grid *point_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*point_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Convert atom IDs to conservation values
  RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
  for (int i = 0; i < grid->NEntries(); i++) {
    RNScalar atom_id = *grid_valuesp;
    if (atom_id > 0) {
      PDBAtom *atom = model->Atom((int) atom_id);
      RNScalar conservation = 0.5 * (2 - atom->conservation);
      if (atom->conservation == PDB_UNKNOWN) conservation = 0;
      if (conservation < 0) conservation = 0;
      else if (conservation > 1) conservation = 1;
      *grid_valuesp = conservation;
    }
    grid_valuesp++;
  }

  // Blur with gaussian
  if (ignore_zeros) grid->Threshold(1.0E-20, DONT_CARE, R3_GRID_KEEP_VALUE);
  BlurGrid(grid, conservation_sigma);
  if (ignore_zeros) grid->Threshold(DONT_CARE + 1.0E-20, 0, R3_GRID_KEEP_VALUE);

  // Normalize the values
  RNInterval range = grid->Range();
  grid->Subtract(range.Min());
  grid->Divide((range.Max() - range.Min()) / 0.9);
  grid->Add(0.1);

  // Print statistics
  if (print_verbose) {
    printf("Created conservation grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateHydrophobicityGrid(PDBModel *model, R3Grid *point_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*point_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Convert atom IDs to hydrophobicity values
  RNScalar min_hydrophobicity = -1.01;
  RNScalar max_hydrophobicity = 2.25;
  RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
  for (int i = 0; i < grid->NEntries(); i++) {
    RNScalar atom_id = *grid_valuesp;
    if (atom_id > 0) {
      PDBAtom *atom = model->Atom((int) atom_id);
      PDBAminoAcid *aminoacid = atom->AminoAcid();
      RNScalar hydrophobicity = 0.5;
      if (aminoacid) {
        hydrophobicity = 
          (aminoacid->hydrophobicity - min_hydrophobicity)  / 
          (max_hydrophobicity - min_hydrophobicity);
      }
      *grid_valuesp = hydrophobicity;
    }
    grid_valuesp++;
  }

  // Blur with gaussian
  if (ignore_zeros) grid->Threshold(1.0E-20, DONT_CARE, R3_GRID_KEEP_VALUE);
  BlurGrid(grid, hydrophobicity_sigma);
  if (ignore_zeros) grid->Threshold(DONT_CARE + 1.0E-20, 0, R3_GRID_KEEP_VALUE);

  // Normalize the values
  RNInterval range = grid->Range();
  grid->Subtract(range.Min());
  grid->Divide((range.Max() - range.Min()) / 0.9);
  grid->Add(0.1);

  // Print statistics
  if (print_verbose) {
    printf("Created hydrophobicity grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateCombinedGrid(PDBModel *model, R3Grid **grids, GridProperties *grid_properties, int ngrids, 
  const R3Box& world_box, int xres, int yres, int zres)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(world_box);

  // Fill grid with ones initially
  grid->Clear(1);

  // Combine grids
  R3Grid tmp(xres, yres, zres);
  for (int i = 0; i < ngrids; i++) {
    // Check grid
    if (!grids[i]) continue;
    if (i == POINT_GRID) continue;
    else if (i == INTERIOR_GRID) continue;
    else if (i == DISTANCE_GRID) continue;

    // Copy grid
    tmp = *grids[i];

    // Scale by exponent
    tmp.Pow(grid_properties[i].exponent);

    // Multiply grid by tmp
    grid->Multiply(tmp);
  }

  // Normalize the values
  RNInterval range = grid->Range();
  grid->Subtract(range.Min());
  grid->Divide((range.Max() - range.Min()) / 0.9);
  grid->Add(0.1);

  // Print statistics
  if (print_verbose) {
    printf("Created combined grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
ReadGrid(char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    RNFail("Unable to allocated grid");
    return 0;
  }

  // Read grid 
  if (!grid->ReadFile(grid_name)) {
    RNFail("Unable to read grid file %s", grid_name);
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Dimensions = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Maximum = %g\n", grid->Maximum());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



int
CreateGrids(PDBFile *file, R3Grid **grids, int max_grids)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get PDB model from file
  if (file->NModels() < 1) return 0;
  PDBModel *model = file->Model(0);

  // Set origin and radius 
  if (world_radius == PDB_UNKNOWN) world_radius = (ligand) ? 15 : 0.5 * model->BBox().LongestAxisLength();
  if (world_origin[0] == PDB_UNKNOWN) world_origin = (ligand) ? ligand->Centroid() : model->BBox().Centroid();

  // Determine world box (must be square for shape matching)
  R3Point bottom(world_origin[0]-world_radius, world_origin[1]-world_radius, world_origin[2]-world_radius);
  R3Point top(world_origin[0]+world_radius, world_origin[1]+world_radius, world_origin[2]+world_radius);
  R3Box world_box = R3Box(bottom, top);

  // Determine resolution for grids 
  int xres, yres, zres;
  if (power_of_two) {
    // Determine grid resolution (must be power of 2 for shape matching)
    int res = 4;
    int max_res = max_resolution[0];
    if (max_res < max_resolution[1]) max_res = max_resolution[1];
    if (max_res < max_resolution[2]) max_res = max_resolution[2];
    while (((world_box.XLength() / res) > grid_spacing) && (res < max_res)) res *= 2;
    xres = yres = zres = res;
  }
  else {
    // Determine grid resolution
    xres = (int) (world_box.XLength() / grid_spacing + 1);
    yres = (int) (world_box.YLength() / grid_spacing + 1);
    zres = (int) (world_box.ZLength() / grid_spacing + 1);
    if (xres > max_resolution[0]) xres = max_resolution[0];
    if (yres > max_resolution[1]) yres = max_resolution[1];
    if (zres > max_resolution[2]) zres = max_resolution[2];
    if (xres < 1) xres = 1;
    if (yres < 1) yres = 1;
    if (zres < 1) zres = 1;
  }

  // Determine which grids to compute
  RNBoolean compute_conservation_grid = (hssp_name || consurf_name) && (conservation_sigma > 0);
  RNBoolean compute_hydrophobicity_grid = hydrophobicity && (hydrophobicity_sigma > 0);
  RNBoolean compute_charge_grid = charge && (charge_sigma > 0);
  RNBoolean compute_gauss_grid = gauss && (gauss_sigma > 0);
  RNBoolean compute_surfnet_grid = surfnet_scheme || surfnet_file || centerline;
  RNBoolean compute_distance_grid = proximity || compute_surfnet_grid;
  RNBoolean compute_exterior_grid = exterior || symmetry;
  RNBoolean compute_interior_grid = compute_distance_grid || compute_exterior_grid;
  RNBoolean compute_point_grid = compute_gauss_grid || compute_charge_grid || compute_conservation_grid || compute_hydrophobicity_grid;

  // Initialize number of grids
  int ngrids = 0;

  // Create input grid
  if (input_grid_name) {
    grids[INPUT_GRID] = ReadGrid(input_grid_name);
    if (!grids[INPUT_GRID]) return 0;
    else ngrids++;

    // Update grid parameters
    world_box = grids[INPUT_GRID]->WorldBox();
    xres = grids[INPUT_GRID]->XResolution();
    yres = grids[INPUT_GRID]->YResolution();
    zres = grids[INPUT_GRID]->ZResolution();
  }

  // Create random grid
  if (rndom) {
    grids[RANDOM_GRID] = CreateRandomGrid(model, world_box, xres, yres, zres);
    if (!grids[RANDOM_GRID]) return 0;
    else ngrids++;
  }

  // Create point grid
  if (compute_point_grid) {
    grids[POINT_GRID] = CreatePointGrid(model, world_box, xres, yres, zres);
    if (!grids[POINT_GRID]) return 0;
    else ngrids++;
  }

  // Create interior grid
  if (compute_interior_grid) {
    grids[INTERIOR_GRID] = CreateInteriorGrid(model, world_box, xres, yres, zres);
    if (!grids[INTERIOR_GRID]) return 0;
    else ngrids++;
  }

  // Create exterior grid
  if (compute_exterior_grid) {
    grids[EXTERIOR_GRID] = CreateExteriorGrid(model, grids[INTERIOR_GRID]);
    if (!grids[EXTERIOR_GRID]) return 0;
    else ngrids++;
  }

  // Create distance grid
  if (compute_distance_grid) {
    grids[DISTANCE_GRID] = CreateDistanceGrid(model, grids[INTERIOR_GRID]);
    if (!grids[DISTANCE_GRID]) return 0;
    else ngrids++;
  }

  // Create proximity grid
  if (proximity) {
    grids[PROXIMITY_GRID] = CreateProximityGrid(model, grids[DISTANCE_GRID]);
    if (!grids[PROXIMITY_GRID]) return 0;
    else ngrids++;
  }

  // Create surfnet grid
  if (compute_surfnet_grid) {
    if (!surfnet_file) grids[SURFNET_GRID] = CreateSurfnetGrid(model, grids[DISTANCE_GRID]);
    else grids[SURFNET_GRID] = CreateSurfnetGrid(model, surfnet_file->Model(0), world_box, xres, yres, zres);
    ApplySurfnetScheme(grids[SURFNET_GRID], surfnet_scheme);
    if (!grids[SURFNET_GRID]) return 0;
    else ngrids++;
  }

  // Create symmetry grid
  if (symmetry) {
    grids[SYMMETRY_GRID] = CreateSymmetryGrid(model, grids[EXTERIOR_GRID]);
    if (!grids[SYMMETRY_GRID]) return 0;
    else ngrids++;
  }

  // Create centerline grid
  if (centerline) {
    grids[CENTERLINE_GRID] = CreateCenterlineGrid(model, grids[SURFNET_GRID]);
    if (!grids[CENTERLINE_GRID]) return 0;
    else ngrids++;
  }

  // Create gauss grid
  if (gauss) {
    grids[GAUSS_GRID] = CreateGaussGrid(model, grids[POINT_GRID]); 
    if (!grids[GAUSS_GRID]) return 0;
    else ngrids++;
  }

  // Create charge grid
  if (compute_charge_grid) {
    grids[CHARGE_GRID] = CreateChargeGrid(model, grids[POINT_GRID]); 
    if (!grids[CHARGE_GRID]) return 0;
    else ngrids++;
  }

  // Create conservation grid
  if (compute_conservation_grid) {
    grids[CONSERVATION_GRID] = CreateConservationGrid(model, grids[POINT_GRID]); 
    if (!grids[CONSERVATION_GRID]) return 0;
    else ngrids++;
  }

  // Create hydrophobicity grid
  if (compute_hydrophobicity_grid) {
    grids[HYDROPHOBICITY_GRID] = CreateHydrophobicityGrid(model, grids[POINT_GRID]); 
    if (!grids[HYDROPHOBICITY_GRID]) return 0;
    else ngrids++;
  }

  // Create combined grid
  grids[COMBINED_GRID] = CreateCombinedGrid(model, grids, grid_properties, NUM_GRIDS, world_box, xres, yres, zres);
  if (!grids[COMBINED_GRID]) return 0;
  else ngrids++;

  // Set slice coords to middle of grid
  grid_slice_coords[RN_X] = xres/2;
  grid_slice_coords[RN_Y] = yres/2;
  grid_slice_coords[RN_Z] = zres/2;

  // Print statistics
  if (print_verbose) {
    printf("Overall statistics ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Grids = %d\n", ngrids);
    fflush(stdout);
  }

  // Return number of grids created
  return ngrids;
}



int
AnalyzeGrids(PDBFile *file, PDBResidue *ligand, R3Grid **grids, int max_grids, RNScalar target_coverage, RNScalar target_volume)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Check if should analyze grids
  if ((target_coverage == 0) && (target_volume == 0)) return 1;

  // Create grid with ligands rasterized
  const RNArray<PDBAtom *>& atoms = (ligand) ? ligand->atoms : file->Model(0)->atoms;
  R3Grid ligand_grid(grids[COMBINED_GRID]->XResolution(), grids[COMBINED_GRID]->YResolution(), grids[COMBINED_GRID]->ZResolution());
  ligand_grid.SetWorldToGridTransformation(grids[COMBINED_GRID]->WorldToGridTransformation());
  for (int i = 0; i < atoms.NEntries(); i++) {
    PDBAtom *atom = atoms.Kth(i);
    if (!atom->IsHetAtom()) continue;
    ligand_grid.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
  }

  // Set thresholds for every grid to cover certain percentage of ligand
  if (target_coverage > 0) {
    // Allocate array of values found at positions of ligand atoms
    int ligand_nvalues = ligand_grid.Cardinality();
    RNScalar *ligand_values = new RNScalar [ ligand_nvalues ];
    assert(ligand_values);

    // Set thresholds for every grid
    for (int i = 0; i < NUM_GRIDS; i++) {
      if (!grids[i]) continue;

      // Fill array of ligand values 
      int count = 0;
      for (int j = 0; j < grids[i]->NEntries(); j++) {
        if (ligand_grid.GridValue(j) == 0) continue;
        assert(count < ligand_nvalues);
        ligand_values[count++] = grids[i]->GridValue(j);
      }

      // Sort the values
      qsort(ligand_values, ligand_nvalues, sizeof(RNScalar), RNCompareScalars);

      // Set the grid threshold such that the grid covers X% of the ligand
      int ligand_index = (int) ((1.0 - target_coverage) * ligand_nvalues);
      if (ligand_index >= ligand_nvalues) ligand_index = ligand_nvalues - 1;
      grid_properties[i].threshold = ligand_values[ligand_index];
    }
  }
  else {
    // Determine target volume
    if (target_volume <= 0) {
      RNScalar length_scale_factor = grids[COMBINED_GRID]->GridToWorldScaleFactor();
      RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;
      target_volume = 0.5 * ligand_grid.Cardinality() * volume_scale_factor;
    }

    // Set threshold for each grid to meet target volume
    for (int i = 0; i < NUM_GRIDS; i++) {
      if (!grids[i]) continue;
      grid_properties[i].threshold = GridThreshold(grids[i], target_volume);
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Analyzed grids ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    if ((ligand) && (target_coverage > 0)) printf("  Target coverage = %g\n", target_coverage);
    if (target_coverage == 0) printf("  Target volume = %g\n", target_volume);
    printf("  Volumes ...\n");
    for (int i = 0; i < NUM_GRIDS; i++) {
      if (!grids[i]) continue;
      printf("  %s volume = %g\n", grid_properties[i].name, GridVolume(grids[i], grid_properties[i].threshold));
    }
    printf("  Thresholds ...\n");
    for (int i = 0; i < NUM_GRIDS; i++) {
      if (!grids[i]) continue;
      printf("  %s threshold = %g\n", grid_properties[i].name, grid_properties[i].threshold);
    }
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if ((argc == 2) && (*argv[1] == '-')) {
    printf("Usage: symsam [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-debug")) print_debug = 1; 
      else if (!strcmp(*argv, "-batch")) batch = 1; 
      else if (!strcmp(*argv, "-biomolecule")) biomolecule = 1; 
      else if (!strcmp(*argv, "-input")) { argc--; argv++; input_grid_name = *argv; }
      else if (!strcmp(*argv, "-output")) { argc--; argv++; output_grid_name = *argv; }
      else if (!strcmp(*argv, "-consurf")) { argc--; argv++; consurf_name = *argv; }
      else if (!strcmp(*argv, "-hssp")) { argc--; argv++; hssp_name = *argv; }
      else if (!strcmp(*argv, "-surfnet")) { argc--; argv++; surfnet_name = *argv; }
      else if (!strcmp(*argv, "-asa")) { argc--; argv++; asa_name = *argv; }
      else if (!strcmp(*argv, "-grow")) { argc--; argv++; grow_name = *argv; }
      else if (!strcmp(*argv, "-dart")) { argc--; argv++; dart_name = *argv; }
      else if (!strcmp(*argv, "-ligand")) { argc--; argv++; ligand_name = *argv; }
      else if (!strcmp(*argv, "-image")) { argc--; argv++; image_name = *argv; }
      else if (!strcmp(*argv, "-gauss_sigma")) { argc--; argv++; gauss_sigma = atof(*argv); }
      else if (!strcmp(*argv, "-charge_sigma")) { argc--; argv++; charge_sigma = atof(*argv); }
      else if (!strcmp(*argv, "-conservation_sigma")) { argc--; argv++; conservation_sigma = atof(*argv); }
      else if (!strcmp(*argv, "-hydrophobicity_sigma")) { argc--; argv++; hydrophobicity_sigma = atof(*argv); }
      else if (!strcmp(*argv, "-surfnet_sigma")) { argc--; argv++; surfnet_sigma = atof(*argv); }
      else if (!strcmp(*argv, "-surfnet_scheme")) { argc--; argv++; surfnet_scheme = atoi(*argv); }
      else if (!strcmp(*argv, "-grid_threshold")) { argc--; argv++; initial_grid_threshold = atof(*argv); }
      else if (!strcmp(*argv, "-ignore_zeros")) { ignore_zeros = TRUE; }
      else if (!strcmp(*argv, "-power_of_two")) { power_of_two = TRUE; }
      else if (!strcmp(*argv, "-symmetry")) symmetry = 1; 
      else if (!strcmp(*argv, "-proximity")) proximity = 1; 
      else if (!strcmp(*argv, "-centerline")) centerline = 1; 
      else if (!strcmp(*argv, "-gauss")) gauss = 1; 
      else if (!strcmp(*argv, "-charge")) charge = 1; 
      else if (!strcmp(*argv, "-hydrophobicity")) hydrophobicity = 1; 
      else if (!strcmp(*argv, "-random")) rndom = 1; 
      else if (!strcmp(*argv, "-exterior")) exterior = 1; 
      else if (!strcmp(*argv, "-largest_component_only")) largest_component_only = 1; 
      else if (!strcmp(*argv, "-show_names")) show_names = 1; 
      else if (!strcmp(*argv, "-show_residue_frames")) show_residue_frames = 1; 
      else if (!strcmp(*argv, "-show_grid_isosurface")) show_grid_isosurface = 1; 
      else if (!strcmp(*argv, "-show_charge")) show_charge = 1; 
      else if (!strcmp(*argv, "-show_hydrophobicity")) show_hydrophobicity = 1; 
      else if (!strcmp(*argv, "-show_conservation")) show_conservation = 1; 
      else if (!strcmp(*argv, "-show_contacts_only")) show_contacts_only = 1; 
      else if (!strcmp(*argv, "-show_darts")) show_darts = 1; 
      else if (!strcmp(*argv, "-spacing")) { 
        argc--; argv++; grid_spacing = atof(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; max_resolution[0] = atoi(*argv); 
        argc--; argv++; max_resolution[1] = atoi(*argv); 
        argc--; argv++; max_resolution[2] = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-colors")) {
        argv++; argc--; protein_color_scheme = atoi(*argv) % NUM_COLOR_SCHEMES; 
        argv++; argc--; hetatom_color_scheme = atoi(*argv) % NUM_COLOR_SCHEMES; 
        argv++; argc--; ligand_color_scheme = atoi(*argv) % NUM_COLOR_SCHEMES; 
      }
      else if (!strcmp(*argv, "-displays")) {
        argv++; argc--; protein_display_scheme = atoi(*argv) % NUM_DISPLAY_SCHEMES; 
        argv++; argc--; hetatom_display_scheme = atoi(*argv) % NUM_DISPLAY_SCHEMES; 
        argv++; argc--; ligand_display_scheme = atoi(*argv) % NUM_DISPLAY_SCHEMES; 
      }
      else if (!strcmp(*argv, "-background")) { 
        argc--; argv++; background_color[0] = atof(*argv); 
        argc--; argv++; background_color[1] = atof(*argv); 
        argc--; argv++; background_color[2] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-radius")) { 
        argc--; argv++; world_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-center")) { 
        argc--; argv++; world_origin[0] = atof(*argv); 
        argc--; argv++; world_origin[1] = atof(*argv); 
        argc--; argv++; world_origin[2] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-threshold_coverage")) { 
        argc--; argv++; threshold_coverage = atof(*argv); 
      }
      else if (!strcmp(*argv, "-threshold_volume")) { 
        argc--; argv++; threshold_volume = atof(*argv); 
      }
      else if (!strcmp(*argv, "-camera")) { 
        RNCoord x, y, z, tx, ty, tz, ux, uy, uz;
        argv++; argc--; x = atof(*argv); 
        argv++; argc--; y = atof(*argv); 
        argv++; argc--; z = atof(*argv); 
        argv++; argc--; tx = atof(*argv); 
        argv++; argc--; ty = atof(*argv); 
        argv++; argc--; tz = atof(*argv); 
        argv++; argc--; ux = atof(*argv); 
        argv++; argc--; uy = atof(*argv); 
        argv++; argc--; uz = atof(*argv); 
        initial_camera_origin = R3Point(x, y, z);
        initial_camera_towards.Reset(tx, ty, tz);
        initial_camera_up.Reset(ux, uy, uz);
        initial_camera = TRUE;
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check grid resolution
  if ((max_resolution[0] <= 0) || (max_resolution[1] <= 0) || (max_resolution[2] <= 0)) {
    fprintf(stderr, "Invalid grid resolution: %d %d %d\n", max_resolution[0], max_resolution[1], max_resolution[2]);
    return 0;
  }

  // Check surfnet scheme
  if (surfnet_scheme == NO_SURFNET_SCHEME){
    if (surfnet_name || centerline || (surfnet_sigma > 0)) surfnet_scheme = THRESHOLD_SURFNET_SCHEME;
    if (surfnet_name && !strcmp(surfnet_name, "compute")) surfnet_name = NULL;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Initialize GLUI
  GLUIInit(&argc, argv);

  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read pdb
  file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Find ligand
  if (ligand_name) {
    ligand = FindLigand(file, ligand_name);
    if (!ligand) exit(-1);
  }

  // Read conservation files
  if (hssp_name) {
    int status = ReadHsspFiles(file, hssp_name);
    if (!status) exit(-1);
  }
  else if (consurf_name) {
    int status = ReadConsurfFiles(file, consurf_name);
    if (!status) exit(-1);
  }

  // Read accessible surface area file
  if (asa_name) {
    int status = ReadASAFile(file, asa_name);
    if (!status) exit(-1);
  }

  // Read grow file
  if (grow_name) {
    int status = ReadGrowFile(file, grow_name);
    if (!status) exit(-1);
  }

  // Create bonds to ligand
  if (ligand && !grow_name) {
    file->CreateBonds(ligand);
  }

  // Read surfnet file
  if (surfnet_name) {
    surfnet_file = ReadSurfnetFile(surfnet_name);
    if (!surfnet_file) exit(-1);
  }

  // Read dart file
  if (dart_name) {
    int status = ReadDartFile(file, dart_name);
    if (!status) exit(-1);
  }

  // Create grids
  int status1 = CreateGrids(file, grids, NUM_GRIDS);
  if (!status1) exit(-1);

  // Write grid
  if (output_grid_name && grids[COMBINED_GRID]) {
    int status = WriteGrid(grids[COMBINED_GRID], output_grid_name);
    if (!status) exit(-1);
  }

  // Analyze grids
  int status2 = AnalyzeGrids(file, ligand, grids, NUM_GRIDS, threshold_coverage, threshold_volume);
  if (!status2) exit(-1);

  // Set grid threshold
  if (initial_grid_threshold != PDB_UNKNOWN) {
    grid_properties[COMBINED_GRID].threshold = initial_grid_threshold;
  }

  // Run GUI
  if (!batch || image_name) {
    // Create viewer
    viewer = CreateViewer(file, ligand);
    if (!viewer) exit(-1);

    // Run user interface
    GLUIMainLoop();
  }

  // Return success 
  return 0;
}



