// Source file for pocketfinder [An05]



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *pdb_name = NULL;
static char *grid_name = NULL;
static int grid_resolution[3] = { 0, 0, 0 };
static int max_grid_resolution = 256;
static RNLength grid_spacing = 0;
static R3Point *world_center = NULL;
static RNLength world_radius = 0;
static int print_verbose = 0;



// Lennard-Jones parameters from AutoDock version 1.0 (H C N O S)
// http://www.scripps.edu/mb/olson/gmm/autodock/ad305/Using_AutoDock_305.20.html#32242

static RNScalar lennard_jones_c6[6][6] = {
  { 46.73839, 226.9102, 155.9833, 124.0492, PDB_UNKNOWN, 290.0756 }, // H
  { 226.9102, 1127.684, 783.3452, 633.7542, PDB_UNKNOWN, 1476.364 }, // C
  { 155.9833, 783.3452, 546.7653, 445.9175, PDB_UNKNOWN, 1036.932 }, // N
  { 124.0492, 633.7542, 445.9175, 368.6774, PDB_UNKNOWN, 854.6872 }, // O
  { PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN }, // P
  { 290.0756, 1476.364, 1036.932, 854.6872, PDB_UNKNOWN, 1982.756 } // S
};
 
static RNScalar lennard_jones_c12[6][6] = {
  { 1908.578, 88604.24, 39093.66, 38919.64, PDB_UNKNOWN, 126821.3 },
  { 88604.24, 1272653., 610155.1, 588883.8, PDB_UNKNOWN, 1569268. },
  { 39093.66, 610155.1, 266862.2, 249961.4, PDB_UNKNOWN, 721128.6 },
  { 38919.64, 588883.8, 249961.4, 230584.4, PDB_UNKNOWN, 675844.1 },
  { PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN, PDB_UNKNOWN }, // P
  { 126821.3, 1569268., 721128.6, 675844.1, PDB_UNKNOWN, 1813147. }
};
 
 

static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static R3Grid *
CreateGrid(PDBFile *file)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get array of atoms
  if (file->NModels() == 0) return NULL;
  PDBModel *model = file->Model(0);

  // Compute world box
  R3Box world_box;
  if (world_center || (world_radius != 0)) {
    // Compute world centroid
    R3Point world_centroid;
    if (world_center) world_centroid = *world_center;
    else world_centroid = PDBCentroid(model->atoms);

    // Compute world radius
    if (world_radius == 0) {
       world_radius = 3.0 * PDBAverageDistance(model->atoms, world_centroid);
    }

    // Compute world box
    R3Point world_corner1(world_centroid - R3ones_vector * world_radius);
    R3Point world_corner2(world_centroid + R3ones_vector * world_radius);
    world_box = R3Box(world_corner1, world_corner2);
  }
  else {
    // Compute world box
    world_box = model->BBox();
  }

  // Compute grid resolution
  if (grid_spacing == 0) grid_spacing = 0.5;
  int xres = grid_resolution[0];
  if (xres == 0) {
    xres = (int) (world_box.XLength() / grid_spacing + 1);
    if (xres > max_grid_resolution) xres = max_grid_resolution;
  }
  int yres = grid_resolution[1];
  if (yres == 0) {
    yres = (int) (world_box.YLength() / grid_spacing + 1);
    if (yres > max_grid_resolution) yres = max_grid_resolution;
  }
  int zres = grid_resolution[2];
  if (zres == 0) {
    zres = (int) (world_box.ZLength() / grid_spacing + 1);
    if (zres > max_grid_resolution) zres = max_grid_resolution;
  }

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set transform 
  grid->SetWorldToGridTransformation(world_box);

  // Compute van der Waals force field using Lennard-Jones potential
  for (int a = 0; a < model->NAtoms(); a++) {
    // Get atom
    PDBAtom *atom = model->Atom(a);
    if (atom->IsHetAtom()) continue;

    // Get atom element index
    PDBElement *element = atom->Element();
    if (!element) continue;
    int element_index = element - &PDBelements[0];
    assert(element_index >= 0);
    if (element_index > 5) continue;

    // Get atom's constants for Lennard-Jones potential (assume other atom is carbon)
    RNScalar c6 = lennard_jones_c6[PDB_C_ELEMENT][element_index];
    RNScalar c12 = lennard_jones_c12[PDB_C_ELEMENT][element_index];
    if ((c6 == PDB_UNKNOWN) || (c12 == PDB_UNKNOWN)) continue;

    // Get grid extent for atom
    const RNLength world_extent = 10;
    const RNLength world_extent2 = world_extent * world_extent;
    const R3Point& world_position = atom->Position();
    RNLength grid_extent = world_extent * grid->WorldToGridScaleFactor();
    R3Point grid_position = grid->GridPosition(world_position);
    int grid_xlo = (int) (grid_position.X() - grid_extent);
    int grid_ylo = (int) (grid_position.Y() - grid_extent);
    int grid_zlo = (int) (grid_position.Z() - grid_extent);
    int grid_xhi = (int)  ceil(grid_position.X() + grid_extent);
    int grid_yhi = (int)  ceil(grid_position.Y() + grid_extent);
    int grid_zhi = (int)  ceil(grid_position.Z() + grid_extent);
    grid_xlo = (grid_xlo < 0) ? 0 : ((grid_xlo >= grid->XResolution()) ? grid->XResolution()-1 : grid_xlo);
    grid_xhi = (grid_xhi < 0) ? 0 : ((grid_xhi >= grid->XResolution()) ? grid->XResolution()-1 : grid_xhi);
    grid_ylo = (grid_ylo < 0) ? 0 : ((grid_ylo >= grid->YResolution()) ? grid->YResolution()-1 : grid_ylo);
    grid_yhi = (grid_yhi < 0) ? 0 : ((grid_yhi >= grid->YResolution()) ? grid->YResolution()-1 : grid_yhi);
    grid_zlo = (grid_zlo < 0) ? 0 : ((grid_zlo >= grid->ZResolution()) ? grid->ZResolution()-1 : grid_zlo);
    grid_zhi = (grid_zhi < 0) ? 0 : ((grid_zhi >= grid->ZResolution()) ? grid->ZResolution()-1 : grid_zhi);
    RNLength grid_to_world_scale_factor2 = grid->GridToWorldScaleFactor() * grid->GridToWorldScaleFactor();

    // Add Lennard-Jones potential for atom
    for (int k = grid_zlo; k <= grid_zhi; k++) {
      RNScalar grid_zdistance = grid_position.Z() - k;
      RNScalar grid_zdistance2 = grid_zdistance * grid_zdistance;
      for (int j = grid_ylo; j <= grid_yhi; j++) {
        RNScalar grid_ydistance = grid_position.Y() - j;
        RNScalar grid_ydistance2 = grid_ydistance * grid_ydistance;
        for (int i = grid_xlo; i <= grid_xhi; i++) {
          RNScalar grid_xdistance = grid_position.X() - i;
          RNScalar grid_xdistance2 = grid_xdistance * grid_xdistance;
          RNLength grid_distance2 = grid_xdistance2 + grid_ydistance2 + grid_zdistance2;
          RNLength world_distance2 = grid_distance2 * grid_to_world_scale_factor2;
          if (world_distance2 > world_extent2) continue;
          RNLength r6 = world_distance2 * world_distance2 * world_distance2;
          RNLength r12 = r6 * r6;
          RNScalar value = (c12/r12) - (c6/r6);
          value = -value; // Make positive values better
          if (value < 0) continue;
          grid->AddGridValue(i, j, k, value);
        }
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



static int 
WriteGrid(R3Grid *grid, const char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write grid
  int status = grid->WriteFile(grid_name);

  // Print statistics
  if (print_verbose) {
    printf("Wrote grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", status * sizeof(RNScalar));
    fflush(stdout);
  }

  // Return status
  return status;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: pdb2grd pdbfile gridfile [-resolution x y z] [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-radius")) { 
        argc--; argv++; world_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-center")) { 
        world_center = new R3Point();
        argc--; argv++; (*world_center)[0] = atof(*argv); 
        argc--; argv++; (*world_center)[1] = atof(*argv); 
        argc--; argv++; (*world_center)[2] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; grid_resolution[0] = atoi(*argv); 
        argc--; argv++; grid_resolution[1] = atoi(*argv); 
        argc--; argv++; grid_resolution[2] = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-max_resolution")) { 
        argc--; argv++; max_grid_resolution = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-spacing")) { 
        argc--; argv++; grid_spacing = atof(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else if (!grid_name) grid_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check grid resolution
  if ((grid_resolution[0] < 0) || (grid_resolution[1] < 0) || (grid_resolution[2] < 0)) {
    fprintf(stderr, "Invalid grid resolution: %d %d %d\n", grid_resolution[0], grid_resolution[1], grid_resolution[2]);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read PDB file
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Create grid 
  R3Grid *grid = CreateGrid(file);
  if (!grid) exit(-1);

  // Write grid
  int status = WriteGrid(grid, grid_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}
