// Source file for the conversion from MDL SD Format to PDB format



// Include files 

#include "PDB/PDB.h"



// Program variables

static char *sdf_name = NULL;
static char *pdb_name = NULL;
static int print_verbose = 0;



static char *
SDFLine(FILE *fp, char *buffer, int max_length, int& line_length, int& line_counter, char *filename, RNBoolean print_error)
{
  // Get next line
  if (!fgets(buffer, max_length, fp)) {
    if (print_error) fprintf(stderr, "Unable to read line %d from %s\n", line_counter, filename);
    return NULL;
  }

  // Update line length and counter
  line_length = strlen(buffer);
  line_counter++;

  // Return success
  return buffer;
}



static char *
SDFToken(char *buffer, int start, int stop, int length)
{
  static char token[128];
  assert(stop-start+1 < 32);
  char *tokenp = token;
  if (stop > length) stop = length;
  for (int i = start; i <= stop; i++) 
    if (!isspace(buffer[i])) 
      *(tokenp++) = buffer[i];
  *tokenp = '\0';
  return token;
}



static RNArray<PDBFile *> *
ReadSDF(char *sdf_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open sdf file
  FILE *fp = fopen(sdf_name, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open sdf file: %s\n", sdf_name);
    return NULL;
  }

  // Allocate array of PDB files
  RNArray<PDBFile *> *files = new RNArray<PDBFile *>();
  if (!files) {
    fprintf(stderr, "Unable to allocate array of PDB files\n");
    return NULL;
  }

  // Read molecules from file
  char buffer[1024];
  int atom_counter = 0;
  int line_counter = 0;
  int line_length = 0;
  while (TRUE) {
    // Read header
    if (!SDFLine(fp, buffer, 1023, line_length, line_counter, sdf_name, FALSE)) break;
    if (!SDFLine(fp, buffer, 1023, line_length, line_counter, sdf_name, TRUE)) return NULL;
    if (!SDFLine(fp, buffer, 1023, line_length, line_counter, sdf_name, TRUE)) return NULL;

    // Read counts
    if (!SDFLine(fp, buffer, 1023, line_length, line_counter, sdf_name, TRUE)) return NULL;
    int natoms = atoi(SDFToken(buffer, 0, 2, line_length));

    // Allocate PDB file
    PDBFile *file = new PDBFile("XXX");
    if (!file) {
      fprintf(stderr, "Unable to allocate PDB file for molecule %d in %s\n", files->NEntries()+1, sdf_name);
      return NULL;
    }

    // Create a model
    PDBModel *model = new PDBModel(file, "1");
    if (!model) {
      fprintf(stderr, "Unable to allocate PDB model for molecule %d in %s\n", files->NEntries()+1, +1, sdf_name);
      return NULL;
    }

    // Create a chain
    PDBChain *chain = new PDBChain(model, " ");
    if (!chain) {
      fprintf(stderr, "Unable to allocate PDB chain for molecule %d in %s\n", files->NEntries()+1, +1, sdf_name);
      return NULL;
    }

    // Create a residue
    PDBResidue *residue = new PDBResidue(model, chain, "XXX", 1, ' ');
    if (!residue) {
      fprintf(stderr, "Unable to allocate PDB residue for molecule %d in %s\n", files->NEntries()+1, +1, sdf_name);
      return NULL;
    }

    // Read the atoms
    for (int i = 0; i < natoms; i++) {
      // Read atom line
      if (!SDFLine(fp, buffer, 1023, line_length, line_counter, sdf_name, TRUE)) return NULL;

      // Parse atom line
      RNScalar x = atof(SDFToken(buffer, 0, 9, line_length));
      RNScalar y = atof(SDFToken(buffer, 10, 19, line_length));
      RNScalar z  = atof(SDFToken(buffer, 20, 29, line_length));
      char element_name[4]; strncpy(element_name, SDFToken(buffer, 31, 31, line_length), 4); 
      PDBElement *element = PDBFindElement(element_name);
      if (element_name[0] == 'H') continue;

      // Make PDB atom
      char atom_name[16]; sprintf(atom_name, "%c%d", element_name[0], i);
      PDBAtom *atom = new PDBAtom(model, chain, residue, element, i, atom_name, ' ', x, y, z, 1, 0, 0, 1);
      if (!atom) {
        fprintf(stderr, "Unable to allocate atom\n");
        return NULL;
      }

      // Increment counter
      atom_counter++;
    }

    // Insert pdb file into array
    if (residue->NAtoms() > 0) files->Insert(file);

    // Read trailing stuff until $$$$
    while (SDFLine(fp, buffer, 1023, line_length, line_counter, sdf_name, TRUE)) {
      if (!strncmp(buffer, "$$$$", 4)) break;
    }
  }

  // Close sdf file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Read SDF file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Molecules = %d\n", files->NEntries());
    printf("  # Atoms = %d\n", atom_counter);
    fflush(stdout);
  }

  // Return array of pdb files
  return files;
}



static int 
WritePDB(RNArray<PDBFile *> *files, const char *pdb_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get base name
  char base_name[256];
  strcpy(base_name, pdb_name);
  // char *endp = strrchr(base_name, '.');
  // if (endp) *endp = '\0';

  // Write PDB files
  for (int i = 0; i < files->NEntries(); i++) {
    PDBFile *file = files->Kth(i);
    char filename[256];
    sprintf(filename, "%s.%d.pdb", base_name, i+1);
    if (!file->WriteFile(filename)) {
      fprintf(stderr, "Unable to write %s\n", filename);
      return 0;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Wrote PDB files ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Files = %d\n", files->NEntries());
    printf("  Base name = %s\n", base_name);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 2) {
    printf("Usage: sdf2pdb sdffile [pdbfile] [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    else {
      if (!sdf_name) sdf_name = *argv;
      else if (!pdb_name) pdb_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    argv++; argc--;
  }

  // Check sdf filename
  if (!sdf_name) {
    fprintf(stderr, "You did not specify a sdf file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read sdf file
  RNArray<PDBFile *> *files = ReadSDF(sdf_name);
  if (!files) exit(-1);

  // Write pdb file
  int status = WritePDB(files, pdb_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}
