// Source file for surfnet



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *pdb_name = NULL;
static char *grid_name = NULL;
static int grid_resolution[3] = { 0, 0, 0 };
static int max_grid_resolution = 256;
static RNScalar grid_spacing = 0;
static R3Point *world_center = NULL;
static RNScalar world_radius = 0;
static int print_verbose = 0;



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



R3Grid *
CreateDistanceGrid(PDBFile *file, const R3Box& world_box, int xres, int yres, int zres)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get model from PDB file
  if (file->NModels() == 0) return NULL;
  PDBModel *model = file->Model(0);

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set transform 
  grid->SetWorldToGridTransformation(world_box);

  // Rasterize atom IDs
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (atom->IsHetAtom()) continue;
    R3Point grid_position = grid->GridPosition(atom->Position());
    int x = (int) (grid_position[0] + 0.5);
    if ((x < 0) || (x >= grid->XResolution())) continue;
    int y = (int) (grid_position[1] + 0.5);
    if ((y < 0) || (y >= grid->YResolution())) continue;
    int z = (int) (grid_position[2] + 0.5);
    if ((z < 0) || (z >= grid->ZResolution())) continue;
    grid->SetGridValue(x, y, z, i + 1.5);
  }

  // Compute atom voronoi regions
  grid->Voronoi();

  // Compute exact distances to closest atoms for every grid cell
  for (int k = 0; k < grid->ZResolution(); k++) {
    for (int j = 0; j < grid->YResolution(); j++) {
      for (int i = 0; i < grid->XResolution(); i++) {
        R3Point world_position = grid->WorldPosition(i, j, k);
        RNScalar grid_value = grid->GridValue(i, j, k);
        int closest_atom_id = (int) (grid_value - 1.0);
        assert((closest_atom_id >= 0) && (closest_atom_id < model->NAtoms()));
        PDBAtom *closest_atom = model->Atom(closest_atom_id);
        RNLength closest_atom_distance = R3Distance(world_position, closest_atom->Position()) - closest_atom->Radius();
        grid->SetGridValue(i, j, k, closest_atom_distance);
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created distance grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateSurfnetGrid(PDBFile *file, R3Grid *distance_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get model from PDB file
  if (file->NModels() == 0) return NULL;
  PDBModel *model = file->Model(0);

  // Allocate grid
  R3Grid *grid = new R3Grid(distance_grid->XResolution(), distance_grid->YResolution(), distance_grid->ZResolution());
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set transform 
  grid->SetWorldToGridTransformation(distance_grid->WorldToGridTransformation());

  // Rasterize sphere at midpoint between every pair of atoms
  int natoms = 0;
  int npairs = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    // Get atom1
    PDBAtom *atom1 = model->Atom(i);
    if (atom1->IsHetAtom()) continue;
    if (atom1->accessible_surface_area == 0) continue;
    RNScalar radius1 = atom1->Radius();
    for (int j = i+1; j < model->NAtoms(); j++) {
      // Get atom2
      PDBAtom *atom2 = model->Atom(j);
      if (atom2->IsHetAtom()) continue;
      if (atom2->accessible_surface_area == 0) continue;
      RNScalar radius2 = atom2->Radius();

      // Compute sphere between the atoms
      R3Point center = 0.5 * (atom1->position + atom2->position);
      RNLength span_length = R3Distance(atom1->position, atom2->position) - radius1 - radius2;
      if (span_length  > 10) continue;
      RNScalar distance = distance_grid->WorldValue(center);
      if (distance < 1.5) continue;
      if (distance > 4) distance = 4;
      grid->RasterizeWorldSphere(center, distance, 1.0);
      npairs++;
    }
    natoms++;
  }

  // Print statistics
  if (print_verbose) {
    printf("Created surfnet grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    printf("  L2Norm = %g\n", grid->L2Norm());
    printf("  # Atoms = %d\n", natoms);
    printf("  # Pairs = %d\n", npairs);
    fflush(stdout);
  }

  // Return grid
  return grid;
}



static R3Grid *
CreateGrid(PDBFile *file)
{
  // Get array of atoms
  if (file->NModels() == 0) return NULL;
  PDBModel *model = file->Model(0);

  // Compute world box
  R3Box world_box;
  if (world_center || (world_radius != 0)) {
    // Compute world centroid
    R3Point world_centroid;
    if (world_center) world_centroid = *world_center;
    else world_centroid = PDBCentroid(model->atoms);

    // Compute world radius
    if (world_radius == 0) {
       world_radius = 3.0 * PDBAverageDistance(model->atoms, world_centroid);
    }

    // Compute world box
    R3Point world_corner1(world_centroid - R3ones_vector * world_radius);
    R3Point world_corner2(world_centroid + R3ones_vector * world_radius);
    world_box = R3Box(world_corner1, world_corner2);
  }
  else {
    // Compute world box
    world_box = model->BBox();
  }

  // Compute grid resolution
  if (grid_spacing == 0) grid_spacing = 0.5;
  int xres = grid_resolution[0];
  if (xres == 0) {
    xres = (int) (world_box.XLength() / grid_spacing + 1);
    if (xres > max_grid_resolution) xres = max_grid_resolution;
  }
  int yres = grid_resolution[1];
  if (yres == 0) {
    yres = (int) (world_box.YLength() / grid_spacing + 1);
    if (yres > max_grid_resolution) yres = max_grid_resolution;
  }
  int zres = grid_resolution[2];
  if (zres == 0) {
    zres = (int) (world_box.ZLength() / grid_spacing + 1);
    if (zres > max_grid_resolution) zres = max_grid_resolution;
  }

  // Create distance grid
  R3Grid *distance_grid = CreateDistanceGrid(file, world_box, xres, yres, zres);
  if (!distance_grid) return NULL;

  // Create surfnetgrid
  R3Grid *grid = CreateSurfnetGrid(file, distance_grid);
  if (!grid) return NULL;

  // Delete distance grid
  delete distance_grid;

  // Return grid
  return grid;
}



static int 
WriteGrid(R3Grid *grid, const char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write grid
  int status = grid->WriteFile(grid_name);

  // Print statistics
  if (print_verbose) {
    printf("Wrote grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", status * sizeof(RNScalar));
    fflush(stdout);
  }

  // Return status
  return status;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: pdb2grd pdbfile gridfile [-resolution x y z] [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-radius")) { 
        argc--; argv++; world_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-center")) { 
        world_center = new R3Point();
        argc--; argv++; (*world_center)[0] = atof(*argv); 
        argc--; argv++; (*world_center)[1] = atof(*argv); 
        argc--; argv++; (*world_center)[2] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; grid_resolution[0] = atoi(*argv); 
        argc--; argv++; grid_resolution[1] = atoi(*argv); 
        argc--; argv++; grid_resolution[2] = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-max_resolution")) { 
        argc--; argv++; max_grid_resolution = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-spacing")) { 
        argc--; argv++; grid_spacing = atof(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else if (!grid_name) grid_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check grid resolution
  if ((grid_resolution[0] < 0) || (grid_resolution[1] < 0) || (grid_resolution[2] < 0)) {
    fprintf(stderr, "Invalid grid resolution: %d %d %d\n", grid_resolution[0], grid_resolution[1], grid_resolution[2]);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read PDB file
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Create grid 
  R3Grid *grid = CreateGrid(file);
  if (!grid) exit(-1);

  // Write grid
  int status = WriteGrid(grid, grid_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}
